
/************************************************************************/
/*                                                                      */
/*  Practica de Compiladors II                                          */
/*  Enginyeria en Informatica - Curs 2004 / 2005                        */
/*                                                                      */
/*  Fitxer adac.c: programa principal que inicia l'execucio del         */
/*                 compilador. Cal indicar per parametre el nom del     */
/*                 fitxer amb codi de MiniAda per a compilar.           */
/*                                                                      */
/*  Autors: Sergio Blanco Cuaresma                                      */
/*          Francisco Jose Aguilar Celdran                              */
/*                                                                      */
/************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "global.h"
#include "taula.h"

#include "print_ts.h"
#include "print_c3a.h"
#include "print_ra.h"

Taula_Simbols *taulaSimbolActual, *taulaSimbolPrincipal;

extern void init_lexic (char *filename);
extern void init_sintactic (char *filename);
extern void end_lexic (void);
extern void end_sintactic (void);
extern FILE *yyin;
extern int yyparse(void);

void impressio_inici_fitxer(FILE *output, int tipus_fitxer);

int main (int argc, char **argv) {
	//String s;
	printf ("Inicialitzant...\n");
	if ((argc < 2) ) {
		fprintf (stderr, "Us: adac source_file\n Usant la entrada estandar (stdin):\n" );
		init_sintactic("stdin");
		init_lexic ("stdin");
		yyin=stdin;
	}
	else if (argc >=3){
		printf ("Us: adacc source_file\n" );
		printf("Sortint...\n");
		exit(0);
	}else{
		yyin=fopen(argv[1],"r");
		if (yyin==NULL){
			printf("El fitxer '%s' no existeix.\n",argv[1]);
			printf("Sortint...\n");
			return (0);
		}
		init_sintactic(argv[1]);
		init_lexic (argv[1]);
	}
	
	taulaSimbolPrincipal = (Taula_Simbols *)crear_taula_simbols("$MAIN$", NULL);
	inserir_tipus_basics(taulaSimbolPrincipal);
	taulaSimbolActual = taulaSimbolPrincipal;
	
	printf("... fet!\n");
	printf("Compilant %s...\n",argv[1]==NULL?"":argv[1]);
	if (yyparse() == 0) {
		printf(":: Sense errors ::\n");
	} else {
		printf("\n<< Errors trobats!! >>\n");
	}
	end_lexic();
	end_sintactic();
	
	// Creem fitxer de log amb extensió ".ts"
	char *outputName = calloc(strlen(argv[1])+4, sizeof(char));
    strcpy(outputName, argv[1]);
    strcat(outputName, ".ts");

    printf("Escribint contingut de la taula de simbols...\n");		
    FILE *output = fopen (outputName, "w");
    if (output==NULL) {
        fprintf(stderr,"El log de la taula de simbols no ha pogut ser creat.\n");
	} else {
		// Imprimim la Taula de simbols
		char cadena_ambit[256];
		sprintf(cadena_ambit,"    ");
		impressio_inici_fitxer(output, 0);
		print_taula_simbols(output, taulaSimbolPrincipal, cadena_ambit, 0, 0);
	}
	
	printf("... fet!\n");
	
    char *outputNameC3A = calloc(strlen(argv[1])+5, sizeof(char));
    strcpy(outputNameC3A, argv[1]);
    strcat(outputNameC3A, ".c3a");
					
	Taula_Simbols *programa_principal = (Taula_Simbols *)obtenir_TS_programa_principal(taulaSimbolPrincipal);
	
    printf("Generant arxiu amb codi de tres adreces...\n");
    FILE *outputC3A = fopen (outputNameC3A, "w");
    if (outputC3A==NULL) {
        fprintf(stderr,"El fitxer de codi de tres adreces no ha pogut ser creat.\n");
	} else {
		// Imprimim el codi de tres adreces
		char cadena_ambit[256];
		sprintf(cadena_ambit,"    ");
		impressio_inici_fitxer(outputC3A, 1);
		print_codi_tres_adreces(outputC3A, programa_principal, cadena_ambit, 0, 0, programa_principal->nom);
	}
	printf("... fet!\n");
	
    char *outputNameRA = calloc(strlen(argv[1])+4, sizeof(char));
    strcpy(outputNameRA, argv[1]);
    strcat(outputNameRA, ".ra");
	
    printf("Generant arxiu amb el contingut dels registres d'activacio...\n");
    FILE *outputRA = fopen (outputNameRA, "w");
    if (output==NULL) {
        fprintf(stderr,"El fitxer dels registres d'activacio no ha pogut ser creat.\n");
	} else {
		// Imprimim els registres d'activacio
		char cadena_ambit[256];
		sprintf(cadena_ambit,"    ");
		impressio_inici_fitxer(outputRA, 2);
		print_registres_activacio(outputRA, programa_principal, cadena_ambit, 0, 0);
	}
	printf("... fet!\n");
	
	return 0;
}

void impressio_inici_fitxer(FILE *output, int tipus_fitxer)
{
	fprintf(output,"\n- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
	fprintf(output,"Practica de Compiladors II - Compilador de MiniAda.\n");
	switch(tipus_fitxer) {
		case 0: fprintf(output,"Fitxer descriptor de la Taula de Simbols.\n"); break;
		case 1: fprintf(output,"Fitxer descriptor del Codi de Tres Adreces.\n"); break;
		case 2: fprintf(output,"Fitxer descriptor dels Registres d'Activacio.\n"); break;
		default: break;
	}
	fprintf(output,"Autors: Sergio Blanco Cuaresma i Francisco Jose Aguilar Celdran.\n");
	fprintf(output,"- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
	if(tipus_fitxer == 0) {
		fprintf(output,"\n");
	}
}

