
/************************************************************************/
/*                                                                      */
/*  Practica de Compiladors II                                          */
/*  Enginyeria en Informatica - Curs 2004 / 2005                        */
/*                                                                      */
/*  Fitxer cua.h: conté la implementacio del Tipus Abstracte            */
/*                de Dades Cua.                                         */
/*                                                                      */
/*  Autors: Sergio Blanco Cuaresma                                      */
/*          Francisco Jose Aguilar Celdran                              */
/*                                                                      */
/************************************************************************/

#define MODUL_CUA
#include <stdlib.h>
#include <stdio.h>
#include "cua.h"
    
/* ------------------------------ */
/* CREAR_CUA: Crea una nova cua.  */
/* ------------------------------ */
         
Cua *crear_cua(void)
{
   Cua *nova;

   nova =(Cua *)malloc(sizeof(Cua));
   nova -> primer = NULL;
   nova -> ultim = NULL;
   nova -> pdi = NULL;   
   return nova;
}

/* ---------------------------------------------- */
/* CUA_BUIDA: respon si la cua esta o no buida.   */
/* ---------------------------------------------- */

unsigned int cua_buida(Cua *c)
{
   if(c -> primer == NULL) 
      return 1;
   else 
      return 0;
}

/* ------------------------------------------------------------- */
/* ENCUAR: afefeix com a ultim element de la cua el node rebut.  */
/* ------------------------------------------------------------- */

void encuar(Cua *c, void *valors)
{
   Node_Cua *nou_node;
   
   nou_node = (Node_Cua *)malloc(sizeof(Node_Cua));
   nou_node -> node = valors;
   nou_node -> anterior = NULL;
   if (cua_buida(c))
   {
      c -> primer = nou_node;
      c -> ultim = nou_node;
      c -> pdi = nou_node;
   }
   else
   {
      c -> ultim -> anterior = nou_node;
      c -> ultim = nou_node;
   } 
}

/* ------------------------------------------------ */
/* DESENCUAR: elimina el primer element de la cua.  */
/* ------------------------------------------------ */

void desencuar(Cua *c)
{
   Node_Cua *aux;
   
   if (!cua_buida(c))
   {  
      if (c -> primer -> anterior == NULL)
      {
         c -> primer = NULL; 
         c -> ultim = NULL;
         c -> pdi = NULL;
      }  
      else
      {
         aux = c -> primer;
         c -> primer = c -> primer -> anterior;
         free(aux);
      }
   }
}

/* --------------------------------------------- */
/* PRIMER: retorna el primer element de la cua.  */
/* --------------------------------------------- */

void *primer(Cua *c)
{
   if (c -> primer != NULL)
      return c -> primer -> node;
   else
      return NULL;
}

/* ----------------------------------------------------------------------- */
/* REINICIAR_PDI: ubica el Punt d'Interes en el primer element de la cua.  */
/* ----------------------------------------------------------------------- */

void reiniciar_PDI(Cua *c)
{
   c -> pdi = c -> primer;   
}

/* ---------------------------------------------------------------------------- */
/* OBTENIR_ELEMENT_ PDI: retorna el node apuntat pel Punt d'Interes de la cua.  */
/* ---------------------------------------------------------------------------- */

void *obtenir_element_PDI(Cua *c)
{
   if(c != NULL)
   {
      if (c -> pdi != NULL)
         return c -> pdi -> node;
      else
         return NULL;
   }
   else
      return NULL;
}

/* ------------------------------------------------------------------ */
/* seguent_PDI: ubica el Punt d'Interes a l'ement seguent de la cua.  */
/* ------------------------------------------------------------------ */

int seguent_PDI(Cua *c)
{
   if(c != NULL)
   {
   	  if (c -> pdi != NULL)
      {
         c -> pdi = c -> pdi -> anterior; 
         return 1;
      }
   }
   else
   {
      return 0;   
   }
}

/* ---------------------------------------------------------------------------- */
/* OBTENIR_ELEMENT_POSICIO: retorna el node en la posicio X                     */
/* ---------------------------------------------------------------------------- */

void *obtenir_element_pos(Cua *cua, int posicio)
{
	int posicio_actual = 1; // Comencem a contar per 1
	
	reiniciar_PDI(cua);
	void *element = obtenir_element_PDI(cua);
	while (element != NULL) {
		if (posicio_actual == posicio) {
			return element;
		}
		posicio_actual++;
		seguent_PDI(cua);
		element = obtenir_element_PDI(cua);
	}
	
	return NULL;
}


/* ---------------------------------------------------------------------------- */
/* EXTREURE_ELEMENT_POSICIO: retorna una cua sense el node de la posicio X      */
/* ---------------------------------------------------------------------------- */

Cua *extreure_element_pos(Cua *cua, int posicio)
{
	Cua *nova_cua = (Cua*)crear_cua();
	int posicio_actual = 1; // Comencem a contar per 1
	
	reiniciar_PDI(cua);
	void *element = obtenir_element_PDI(cua);
	while (element != NULL) {
		if (posicio_actual != posicio) {
			encuar(nova_cua, element);
		}
		posicio_actual++;
		seguent_PDI(cua);
		element = obtenir_element_PDI(cua);
	}
	
	return nova_cua;
}

/* ------------------------------------------------------------------ */
/* fusionar: Crear una nova cua amb els elements de dues cues         */
/* ------------------------------------------------------------------ */

Cua *fusionar_cues(Cua *cua1, Cua *cua2)
{
	Cua *cua_nova = (Cua*)crear_cua();
	
	if (cua1 != NULL) {
		reiniciar_PDI(cua1);
		void *element = obtenir_element_PDI(cua1);
		while (element != NULL) {
			encuar(cua_nova, element);
			seguent_PDI(cua1);
			element = obtenir_element_PDI(cua1);
		}
	}
	
	if (cua2 != NULL) {
		reiniciar_PDI(cua2);
		void *element = obtenir_element_PDI(cua2);
		while (element != NULL) {
			encuar(cua_nova, element);
			seguent_PDI(cua2);
			element = obtenir_element_PDI(cua2);
		}
	}
	
	return cua_nova;
}


