
/************************************************************************/
/*                                                                      */
/*  Practica de Compiladors II                                          */
/*  Enginyeria en Informatica - Curs 2004 / 2005                        */
/*                                                                      */
/*  Fitxer lexic.l: conté la implementació de l'analitzador lexic       */
/*                  desenvolupat amb Flex.                              */
/*                                                                      */
/*  Autors: Sergio Blanco Cuaresma                                      */
/*          Francisco Jose Aguilar Celdran                              */
/*                                                                      */
/************************************************************************/

%option noyywrap
%x STRING_LITERAL
%{
#include <string.h>
#include <math.h>
#include <stdio.h>
#define _GNU_SOURCE
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include "taula.h"
#include "sintactic.h"

FILE *logsLexic;
void lexLog(char *text, int id);
void lexError(const char *format, ...);
int lineNumber;
int colNumber;
extern Taula_Simbols *taulaSimbolActual;
char *eliminar_caracter(char *s, char caracter);

void strToLower(char *que){
	int i;
	for (i=0;i<strlen(que);i++) {
		que[i]=tolower(que[i]);
	}
}


%}
DIGIT [0-9]
IDENTIFIER_LETTER [A-Za-z]
LETTER_OR_DIGIT {IDENTIFIER_LETTER}|{DIGIT}
STRING_ELEMENT \"\"|[^"\n]
HORIZONTAL_TABULATION [\t]
SPACE_CHARACTER [ ]
SPECIAL_CHARACTER [\33-\127^{IDENTIFIER_LETTER}^{DIGIT}]
GRAPHIC_CHARACTER {IDENTIFIER_LETTER}|{DIGIT}|{SPACE_CHARACTER}|{SPECIAL_CHARACTER}
NUMERAL {DIGIT}("_"|{DIGIT})*


SPACE             " "
EXPONENT          {SPACE}*E{SPACE}*"+"?{SPACE}*{NUMERAL}|{SPACE}*E{SPACE}*-{SPACE}*{NUMERAL}
DECIMAL_LITERAL   ("+"|"-"|""){NUMERAL}"."{NUMERAL}|("+"|"-"|""){NUMERAL}"."{NUMERAL}{EXPONENT}|("+"|"-"|""){NUMERAL}{EXPONENT}
NUMERIC_LITERAL   ("+"|"-"|""){NUMERAL}
%%

abs {colNumber += yyleng;lexLog("ABS",PR_ABS);return PR_ABS;}
access  {colNumber += yyleng;lexLog("ACCESS",PR_ACCESS);return PR_ACCESS;}
all {colNumber += yyleng;lexLog("ALL",PR_ALL);return PR_ALL;}
and {colNumber += yyleng;lexLog("AND",PR_AND);return PR_AND;}
array   {colNumber += yyleng;lexLog("ARRAY",PR_ARRAY);return PR_ARRAY;}
begin   {colNumber += yyleng;lexLog("BEGIN",PR_BEGIN);return PR_BEGIN;}
case    {colNumber += yyleng;lexLog("CASE",PR_CASE);return PR_CASE;}
constant {colNumber += yyleng;lexLog("CONSTANT",PR_CONSTANT);return PR_CONSTANT;}
declare {colNumber += yyleng;lexLog("DECLARE",PR_DECLARE);return PR_DECLARE;}
is  {colNumber += yyleng;lexLog("IS",PR_IS);return PR_IS;}
else    {colNumber += yyleng;lexLog("ELSE",PR_ELSE);return PR_ELSE;}
elsif   {colNumber += yyleng;lexLog("ELSIF",PR_ELSIF);return PR_ELSIF;}
end {colNumber += yyleng;lexLog("END",PR_END);return PR_END;}
exit    {colNumber += yyleng;lexLog("EXIT",PR_EXIT);return PR_EXIT;}
for {colNumber += yyleng;lexLog("FOR",PR_FOR);return PR_FOR;}
function {colNumber += yyleng;lexLog("FUNCTION",PR_FUNCTION);return PR_FUNCTION;}
if  {colNumber += yyleng;lexLog("IF",PR_IF);return PR_IF;}
in  {colNumber += yyleng;lexLog("IN",PR_IN);return PR_IN;}
loop    {colNumber += yyleng;lexLog("LOOP",PR_LOOP);return PR_LOOP;}
mod {colNumber += yyleng;lexLog("MOD",PR_MOD);return PR_MOD;}
new {colNumber += yyleng;lexLog("NEW",PR_NEW);return PR_NEW;}
not {colNumber += yyleng;lexLog("NOT",PR_NOT);return PR_NOT;}
null    {colNumber += yyleng;lexLog("NULL",PR_NULL);return PR_NULL;}
of  {colNumber += yyleng;lexLog("OF",PR_OF);return PR_OF;}
or  {colNumber += yyleng;lexLog("OR",PR_OR);return PR_OR;}
others  {colNumber += yyleng;lexLog("OTHERS",PR_OTHERS);return PR_OTHERS;}
procedure {colNumber += yyleng;lexLog("PROCEDURE",PR_PROCEDURE);return PR_PROCEDURE;}
range   {colNumber += yyleng;lexLog("RANGE",PR_RANGE);return PR_RANGE;}
record  {colNumber += yyleng;lexLog("RECORD",PR_RECORD);return PR_RECORD;}
return {colNumber += yyleng;lexLog("RETURN",PR_RETURN);return PR_RETURN;}
reverse {colNumber += yyleng;lexLog("REVERSE",PR_REVERSE);return PR_REVERSE;}
subtype {colNumber += yyleng;lexLog("SUBTYPE",PR_SUBTYPE);return PR_SUBTYPE;}
then    {colNumber += yyleng;lexLog("THEN",PR_THEN);return PR_THEN;}
type    {colNumber += yyleng;lexLog("TYPE",PR_TYPE);return PR_TYPE;}
when    {colNumber += yyleng;lexLog("WHEN",PR_WHEN);return PR_WHEN;}
while   {colNumber += yyleng;lexLog("WHILE",PR_WHILE);return PR_WHILE;}

".."	{colNumber += yyleng;lexLog("2_POINTS",PR_2_POINTS);return PR_2_POINTS;}
";"	{colNumber += yyleng;lexLog(";",';');return ';';}
"."	{colNumber += yyleng;lexLog(".",'.');return '.';}
"("	{colNumber += yyleng;lexLog("(",'(');return '(';}
")"	{colNumber += yyleng;lexLog(")",')');return ')';}
"="	{colNumber += yyleng;lexLog("=",'=');return '=';}
":"	{colNumber += yyleng;lexLog(":",':');return ':';}
"["	{colNumber += yyleng;lexLog("[",'[');return '[';}
"]"	{colNumber += yyleng;lexLog("]",']');return ']';}
":="	{colNumber += yyleng;lexLog("EQU",PR_EQU);return PR_EQU;}
","	{colNumber += yyleng;lexLog(",",',');return ',';}
"^"	{colNumber += yyleng;lexLog("^",'^');return '^';}
"/="	{colNumber += yyleng;lexLog("DIFFERENT",PR_DIFFERENT);return PR_DIFFERENT;}
"<"	{colNumber += yyleng;lexLog("<",'<');return '<';}
"<="	{colNumber += yyleng;lexLog("MINEQU",PR_MINEQU);return PR_MINEQU;}
">"	{colNumber += yyleng;lexLog(">",'>');return '>';}
">="	{colNumber += yyleng;lexLog("MAJEQU",PR_MAJEQU);return PR_MAJEQU;}
"=>"	{colNumber += yyleng;lexLog("ARROW",PR_ARROW);return PR_ARROW;}
"|"	{colNumber += yyleng;lexLog("|",'|');return '|';}
"+"	{colNumber += yyleng;lexLog("+",'+');return '+';}
"-"	{colNumber += yyleng;lexLog("-",'-');return '-';}
"*"	{colNumber += yyleng;lexLog("*",'*');return '*';}
"/"	{colNumber += yyleng;lexLog("/",'/');return '/';}
"&"	{colNumber += yyleng;lexLog("&",'&');return '&';}
"**"	{colNumber += yyleng;lexLog("EXP",PR_EXP);return PR_EXP;}
"_"	{colNumber += yyleng;lexLog("_",'_');return '_';}


{IDENTIFIER_LETTER}({LETTER_OR_DIGIT}|"_")* {
	strToLower(yytext); // Case insensitive
	Simbol *simbol = busca_totes_taules(taulaSimbolActual, yytext);
	if (simbol!=NULL){
		
		yylval.simbol_ptr = simbol;
		if (simbol->atributs.tipus == VARIABLE){
			lexLog(yytext,PR_ID_VARIABLE);
			colNumber += yyleng;
			return PR_ID_VARIABLE;
		}
		else if (simbol->atributs.tipus == PROCEDURE){
			lexLog(yytext,PR_ID_PROCEDURE);
			colNumber += yyleng;
			return PR_ID_PROCEDURE;
		}
		else if (simbol->atributs.tipus == FUNCTION){
			lexLog(yytext,PR_ID_FUNCTION);
			colNumber += yyleng;
			return PR_ID_FUNCTION;
		} 
		else if ((simbol->atributs.tipus == TYPE) || (simbol->atributs.tipus == ENUMERATED_TYPE)
		      || (simbol->atributs.tipus == RECORD_TYPE) || (simbol->atributs.tipus == ARRAY_TYPE)) {
			lexLog(yytext,PR_ID_TYPE);
			colNumber += yyleng;
			return PR_ID_TYPE;
		} else {
			lexError("Tipus de simbol \"%s\" desconegut!",yytext);
		}
	} else {
			lexLog(yytext,PR_ID);
			if (strlen(yytext)>255){
				lexLog ("Error. Id too long",0);
				printf ("Warning: ID too long: %s, id truncated to 255 charactrers\n",yytext);
			}
			// Segons l'enunciat els strings s'han de guardar junt amb la seva longitud a la posició 0
			// Pels identificadors no cal!!
			snprintf(yylval.string, STRING_LENGTH, "%s", yytext);
			
			colNumber += yyleng;
			return PR_ID;
	}
}

{DECIMAL_LITERAL} {
	yytext = eliminar_caracter(yytext,'_');
	colNumber += yyleng;
	
	char comprovacio[1024];
	strncpy(comprovacio,yytext,1022);
	strncpy(comprovacio,eliminar_caracter(comprovacio,'-'),1022);
	
	lexLog (yytext,PR_DECIMAL_LITERAL);
	
	yylval.vfloat = strtod(yytext, (char **)NULL);
	// Si hi ha hagut overflow...
	if ((errno == ERANGE) && (yylval.vfloat > 0)) {
		lexError("El numero decimal surt del rang permes -1.7E-308 i 1.7E+308");
	}
	else
	{
		double val = yylval.vfloat * 2;
		sprintf(comprovacio,"%f",val);
		val = strtod(comprovacio, (char **)NULL); 
		if(errno == ERANGE)
		{
			lexError("El numero decimal surt del rang permes");
		}
	}
	
	return PR_DECIMAL_LITERAL;
};
	
{NUMERIC_LITERAL} {
	yytext = eliminar_caracter(yytext,'_');
	
	colNumber += yyleng;
	lexLog (yytext,PR_NUMERIC_LITERAL);
	
	yylval.vinteger = strtol(yytext, (char **)NULL, 10);
	// Si hi ha hagut overflow...
	if (errno == ERANGE) {
		lexError("El numero decimal surt del rang permes -2147483648 i 2147483647.");
	}
	
	return PR_NUMERIC_LITERAL;
};


\'({GRAPHIC_CHARACTER})\'  {
	colNumber += yyleng;
	lexLog (yytext,PR_CHARACTER_LITERAL);
	yylval.character = yytext[1];
	return PR_CHARACTER_LITERAL;
};


\" { BEGIN(STRING_LITERAL); }

<STRING_LITERAL>[^\n"]*  { 
    if (yyleng > STRING_LENGTH) {
        lexError("String massa llarg, com a maxim poden haver-hi '%i' caracters", STRING_LENGTH);
    }
    else
    {
        colNumber += yyleng;
        sprintf(yylval.string, "%c%s", strlen(yytext), yytext);
        return PR_STRING_LITERAL;
    }
}

<STRING_LITERAL>\"   { BEGIN(INITIAL); }

<STRING_LITERAL>\n   { lexError("Error: Salt de linia abans de finalitzacio de String. (Linia: %d, Columna: %d)\n", lineNumber, colNumber); }


--[^\n]* {
	colNumber += yyleng;
    lexLog(yytext,0);
}

{HORIZONTAL_TABULATION}|{SPACE_CHARACTER} { colNumber++; }

[\n] {
    lineNumber++;
    colNumber = 0;
}


. {
    lexError ("Trobada la sequencia desconeguda '%s' (Linia: %d, Columna: %d)\n", yytext, lineNumber, colNumber);
}

%%

void init_lexic(char *file)
{
    char *temp;
    temp=calloc(strlen(file)+10,sizeof(char));
    strcpy(temp,file);
    strcat(temp,".lex.log");
    logsLexic=fopen (temp,"w");
    if (logsLexic==NULL)
        fprintf(stderr,"The lexical analizer log file has not been created\n");
    lineNumber=1;
    free(temp);
}

void lexError(const char *format, ...){
    va_list arg;
    char linea[4096];
    va_start(arg,format);
    sprintf(linea,"Error a la linia %i columna %i:", lineNumber, colNumber);
    vsprintf(&linea[strlen(linea)],format,arg);
    fprintf(stderr,"%s\n",linea);
    va_end(arg);
    exit(0);
}

void lexLog(char *text,int id)
{
    if (logsLexic!=NULL) {
        fprintf(logsLexic,"Token found: %s-->%d\n",text,id);
	}
}

void end_lexic(void){
    fclose(logsLexic);
}

char *eliminar_caracter(char *s, char caracter)
{
   char *cadena = (char *)malloc(sizeof(char) * strlen(s));
   int index_aux, index_s; //fprintf(yyout," a ");fflush(yyout);
   strcpy(cadena,s);
   index_aux = 0;
   index_s = 0;
   while(index_s < strlen(s)) { //fprintf(yyout," %d(%d) ",index_s,strlen(s));fflush(yyout);
      while(s[index_s] == caracter) {
         index_s++;
      }
      if(index_s < strlen(s)) {
         cadena[index_aux] = s[index_s];
         index_s++;
         index_aux++;
      }
   } //fprintf(yyout," fi: %d(%d) ",index_s,strlen(s));fflush(yyout);
   cadena[index_aux] = '\0'; //fprintf(yyout," aa ");fflush(yyout);
   //fprintf(yyout,"[%d,%d]",strlen(s),strlen(cadena));fflush(yyout);
   index_s = 0;
   //fprintf(yyout," %s %s ",s,cadena);fflush(yyout);
   while(index_s < strlen(s)) {
      //fprintf(yyout,"_%d_",index_s);fflush(yyout);
      s[index_s] = cadena[index_s];
      index_s++;
   }
   //fprintf(yyout," bb ");fflush(yyout);
   //strcpy(s,cadena);  fprintf(yyout," cc ");fflush(yyout);
   free(cadena); //fprintf(yyout," fi elminiar caracter ");fflush(yyout);
   return s;
}
