
/************************************************************************/
/*                                                                      */
/*  Practica de Compiladors II                                          */
/*  Enginyeria en Informatica - Curs 2004 / 2005                        */
/*                                                                      */
/*  Fitxer taula.c: conté la implementació de tots els mètodes que      */
/*                  permeten operar amb la taula de simbols.            */
/*                                                                      */
/*  Autors: Sergio Blanco Cuaresma                                      */
/*          Francisco Jose Aguilar Celdran                              */
/*                                                                      */
/************************************************************************/

#include <stdlib.h>
#include "taula.h"

// Crea una taula que penja d'una altra, si es la principal
// el pare ha de ser NULL
Taula_Simbols *crear_taula_simbols(char nom[STRING_LENGTH], Taula_Simbols *pare) {
	Taula_Simbols *nova_taula;
	int i;
	nova_taula = (Taula_Simbols *)malloc(sizeof(struct Taula_Simbols));
	sprintf(nova_taula->nom, nom);
	nova_taula -> anterior = pare;
	for (i = 0; i < TAM_TAULA_SIMBOLS; i++) {
		// Inicialment no hi ha cap cua creada
		nova_taula -> taula[i] = NULL;
	}
	return nova_taula;
}

void inserir_tipus_basics(Taula_Simbols *taula) {
	Simbol *sym, *sym_param;
	
	sym = crear_i_inserir_en_taula(taula, "boolean");
	sym->atributs.tipus = TYPE;
	sym->atributs.continguts.info_type.tipus = TIPUS_BOOLEAN;
	sym->atributs.continguts.info_type.rang.min = 0;
	sym->atributs.continguts.info_type.rang.max = 1;
	sym->atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;
	sym->atributs.continguts.info_type.tamany = 1;	// Bytes
	
	sym = crear_i_inserir_en_taula(taula, "integer");
	sym->atributs.tipus = TYPE;
	sym->atributs.continguts.info_type.tipus = TIPUS_INTEGER;
	sym->atributs.continguts.info_type.rang.min = -2147483648;
	sym->atributs.continguts.info_type.rang.max = 2147483647;
	sym->atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;
	sym->atributs.continguts.info_type.tamany = 4;	// Bytes
	
	sym = crear_i_inserir_en_taula(taula, "float");
	sym->atributs.tipus = TYPE;
	sym->atributs.continguts.info_type.tipus = TIPUS_FLOAT;
	sym->atributs.continguts.info_type.rang.min = -1.7E-308;
	sym->atributs.continguts.info_type.rang.max = 1.7E+308;
	sym->atributs.continguts.info_type.rang.tipus = TIPUS_FLOAT;
	sym->atributs.continguts.info_type.tamany = 8;	// Bytes
	
	sym = crear_i_inserir_en_taula(taula, "character");
	sym->atributs.tipus = TYPE;
	sym->atributs.continguts.info_type.tipus = TIPUS_CHARACTER;
	sym->atributs.continguts.info_type.rang.min = 0;
	sym->atributs.continguts.info_type.rang.max = 255;
	sym->atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;
	sym->atributs.continguts.info_type.tamany = 1;	// Bytes
	
	sym = crear_i_inserir_en_taula(taula, "string");
	sym->atributs.tipus = TYPE;
	sym->atributs.continguts.info_type.tipus = TIPUS_STRING;
	/*sym->atributs.continguts.info_type.rang.min = -2147483647;
	sym->atributs.continguts.info_type.rang.max = 2147483647;
	sym->atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;*/
	sym->atributs.continguts.info_type.tamany = 1*STRING_LENGTH;	// Bytes
	
	sym = crear_i_inserir_en_taula(taula, "true");
	sym->atributs.tipus = VARIABLE;
	sym->atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
	sym->atributs.continguts.info_variable.constant = TRUE;
	sym->atributs.continguts.info_variable.literal = TRUE;
	sym->atributs.continguts.info_variable.inicialitzat = TRUE;
	sym->atributs.continguts.info_variable.valor.boolean = TRUE;
	sprintf(sym->atributs.continguts.info_variable.nom_tipus, "boolean");
	
	sym = crear_i_inserir_en_taula(taula, "false");
	sym->atributs.tipus = VARIABLE;
	sym->atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
	sym->atributs.continguts.info_variable.constant = TRUE;
	sym->atributs.continguts.info_variable.literal = TRUE;
	sym->atributs.continguts.info_variable.inicialitzat = TRUE;
	sym->atributs.continguts.info_variable.valor.boolean = FALSE;
	sprintf(sym->atributs.continguts.info_variable.nom_tipus, "boolean");
	
	// function get return string
	sym = crear_i_inserir_en_taula(taula, "length");
	sym->atributs.tipus = FUNCTION;
	sym->atributs.continguts.info_function.id = 0;
	sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	sym->atributs.continguts.info_function.complet = TRUE;
	sym->atributs.continguts.info_function.return_indicat = TRUE;
	sym->atributs.continguts.info_function.num_parametres = 1;
	sym->atributs.continguts.info_function.parametres = (Cua*) crear_cua();
	sym->atributs.continguts.info_function.taula = (Taula_Simbols*)crear_taula_simbols("length", taula);
	sym->atributs.continguts.info_function.sobrecarregues = NULL;
	sym->atributs.continguts.info_function.tipus_retorn = TIPUS_INTEGER;
	
	sym_param = crear_i_inserir_en_taula(sym->atributs.continguts.info_function.taula, "item");
	sym_param->atributs.tipus = VARIABLE;
	sym_param->atributs.continguts.info_variable.tipus = TIPUS_STRING;
	sym_param->atributs.continguts.info_variable.constant = FALSE;
	sym_param->atributs.continguts.info_variable.literal = FALSE;
	sym_param->atributs.continguts.info_variable.inicialitzat = FALSE;
	
	encuar(sym->atributs.continguts.info_function.parametres, "item");
	
	
	//// Inici procediment sobrecarregat PUT
	sym = crear_i_inserir_en_taula(taula, "put");
	sym->atributs.tipus = PROCEDURE;
	sym->atributs.continguts.info_procedure.id = 0;
	sym->atributs.continguts.info_procedure.num_sobrecarregues = 3;
	sym->atributs.continguts.info_procedure.complet = TRUE;
	sym->atributs.continguts.info_procedure.num_parametres = 1;
	sym->atributs.continguts.info_procedure.parametres = (Cua*)crear_cua();
	sym->atributs.continguts.info_procedure.taula = (Taula_Simbols*)crear_taula_simbols("put", taula);
	
	// put(item: integer)
	sym_param = crear_i_inserir_en_taula(sym->atributs.continguts.info_procedure.taula, "item");
	sym_param->atributs.tipus = VARIABLE;
	sym_param->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
	sym_param->atributs.continguts.info_variable.constant = FALSE;
	sym_param->atributs.continguts.info_variable.literal = FALSE;
	sym_param->atributs.continguts.info_variable.inicialitzat = FALSE;
	
	encuar(sym->atributs.continguts.info_procedure.parametres, "item");
	
	sym->atributs.continguts.info_procedure.sobrecarregues = (Cua*)crear_cua();
	Info_Procedure *info_procedure = (Info_Procedure*)malloc(sizeof(struct Info_Procedure));
	
	// put(item: float)
	info_procedure->id = 1;
	info_procedure->num_sobrecarregues = 3;
	info_procedure->complet = TRUE;
	info_procedure->num_parametres = 1;
	info_procedure->parametres = (Cua*)crear_cua();
	info_procedure->taula = (Taula_Simbols*)crear_taula_simbols("put", taula);
	
	sym_param = crear_i_inserir_en_taula(info_procedure->taula, "item");
	sym_param->atributs.tipus = VARIABLE;
	sym_param->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
	sym_param->atributs.continguts.info_variable.constant = FALSE;
	sym_param->atributs.continguts.info_variable.literal = FALSE;
	sym_param->atributs.continguts.info_variable.inicialitzat = FALSE;
	
	encuar(info_procedure->parametres, "item");
	encuar(sym->atributs.continguts.info_procedure.sobrecarregues, info_procedure);
	
	// put(item: character)
	info_procedure = (Info_Procedure*)malloc(sizeof(struct Info_Procedure));
	info_procedure->id = 2;
	info_procedure->num_sobrecarregues = 3;
	info_procedure->complet = TRUE;
	info_procedure->num_parametres = 1;
	info_procedure->parametres = (Cua*)crear_cua();
	info_procedure->taula = (Taula_Simbols*)crear_taula_simbols("put", taula);
	
	sym_param = crear_i_inserir_en_taula(info_procedure->taula, "item");
	sym_param->atributs.tipus = VARIABLE;
	sym_param->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
	sym_param->atributs.continguts.info_variable.constant = FALSE;
	sym_param->atributs.continguts.info_variable.literal = FALSE;
	sym_param->atributs.continguts.info_variable.inicialitzat = FALSE;
	
	encuar(info_procedure->parametres, "item");
	encuar(sym->atributs.continguts.info_procedure.sobrecarregues, info_procedure);
	
	// put(item: string)
	info_procedure = (Info_Procedure*)malloc(sizeof(Info_Procedure));
	info_procedure->id = 3;
	info_procedure->num_sobrecarregues = 3;
	info_procedure->complet = TRUE;
	info_procedure->num_parametres = 1;
	info_procedure->parametres = (Cua*)crear_cua();
	info_procedure->taula = (Taula_Simbols*)crear_taula_simbols("put", taula);
	
	sym_param = crear_i_inserir_en_taula(info_procedure->taula, "item");
	sym_param->atributs.tipus = VARIABLE;
	sym_param->atributs.continguts.info_variable.tipus = TIPUS_STRING;
	sym_param->atributs.continguts.info_variable.constant = FALSE;
	sym_param->atributs.continguts.info_variable.literal = FALSE;
	sym_param->atributs.continguts.info_variable.inicialitzat = FALSE;
	
	encuar(info_procedure->parametres, "item");
	encuar(sym->atributs.continguts.info_procedure.sobrecarregues, info_procedure);
	//// Fi procediment sobrecarregat PUT

	// function get return integer
	sym = crear_i_inserir_en_taula(taula, "get_integer");
	sym->atributs.tipus = FUNCTION;
	sym->atributs.continguts.info_function.id = 0;
	sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	sym->atributs.continguts.info_function.complet = TRUE;
	sym->atributs.continguts.info_function.return_indicat = TRUE;
	sym->atributs.continguts.info_function.num_parametres = 0;
	sym->atributs.continguts.info_function.parametres = NULL;
	sym->atributs.continguts.info_function.taula = (Taula_Simbols*)crear_taula_simbols("get_integer", taula);
	sym->atributs.continguts.info_function.sobrecarregues = NULL;
	sym->atributs.continguts.info_function.tipus_retorn = TIPUS_INTEGER;
	
	// function get return float
	sym = crear_i_inserir_en_taula(taula, "get_float");
	sym->atributs.tipus = FUNCTION;
	sym->atributs.continguts.info_function.id = 0;
	sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	sym->atributs.continguts.info_function.complet = TRUE;
	sym->atributs.continguts.info_function.return_indicat = TRUE;
	sym->atributs.continguts.info_function.num_parametres = 0;
	sym->atributs.continguts.info_function.parametres = NULL;
	sym->atributs.continguts.info_function.taula = (Taula_Simbols*)crear_taula_simbols("get_float", taula);
	sym->atributs.continguts.info_function.sobrecarregues = NULL;
	sym->atributs.continguts.info_function.tipus_retorn = TIPUS_FLOAT;
	
	// function get return character
	sym = crear_i_inserir_en_taula(taula, "get_character");
	sym->atributs.tipus = FUNCTION;
	sym->atributs.continguts.info_function.id = 0;
	sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	sym->atributs.continguts.info_function.complet = TRUE;
	sym->atributs.continguts.info_function.return_indicat = TRUE;
	sym->atributs.continguts.info_function.num_parametres = 0;
	sym->atributs.continguts.info_function.parametres = NULL;
	sym->atributs.continguts.info_function.taula = (Taula_Simbols*)crear_taula_simbols("get_character", taula);
	sym->atributs.continguts.info_function.sobrecarregues = NULL;
	sym->atributs.continguts.info_function.tipus_retorn = TIPUS_CHARACTER;
	
	// function get return string
	sym = crear_i_inserir_en_taula(taula, "get_string");
	sym->atributs.tipus = FUNCTION;
	sym->atributs.continguts.info_function.id = 0;
	sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	sym->atributs.continguts.info_function.complet = TRUE;
	sym->atributs.continguts.info_function.return_indicat = TRUE;
	sym->atributs.continguts.info_function.num_parametres = 0;
	sym->atributs.continguts.info_function.parametres = NULL;
	sym->atributs.continguts.info_function.taula = (Taula_Simbols*)crear_taula_simbols("get_string", taula);
	sym->atributs.continguts.info_function.sobrecarregues = NULL;
	sym->atributs.continguts.info_function.tipus_retorn = TIPUS_STRING;
}

int calcular_modul(int numero, int modul) {
	if (numero < modul)
		return numero;
	else {
		int resta = numero % modul;
		resta = resta - 1;
		if (resta < 1)
			resta = modul - 1;
		return resta;
	}
}

// Funció HASH
int dispersio(char *nom_element) {
	int i = 0;
	int valor_dispersio = 0;
	int num_caracters = strlen(nom_element);
	for (i = 0; i < num_caracters; i++) {
		valor_dispersio = valor_dispersio + (int)nom_element[i];
	}
	return calcular_modul(valor_dispersio, TAM_TAULA_SIMBOLS);
}

Simbol* crear_i_inserir_en_taula(Taula_Simbols *simbols, char *nom) {
	
	// Comprobació de que ja existeixi un simbol igual
	int posicio_hash = dispersio(nom);
	
	if (simbols -> taula[posicio_hash] == NULL) {
		simbols -> taula[posicio_hash] = (Cua *)crear_cua();
	}
	Cua *colisions = simbols -> taula[posicio_hash];
	
	reiniciar_PDI(colisions);
	Simbol *simbol = (Simbol *)obtenir_element_PDI(colisions);
	while(simbol != NULL) {
		if (strcmp(simbol -> nom, nom) == 0) {
			// Ja existeix un simbol amb aquest nom
			return NULL;
		}
		seguent_PDI(colisions);
		simbol = (Simbol *)obtenir_element_PDI(colisions);
	}
	
	// No s'ha trobat cap simbol igual
	Simbol *element = (Simbol*)malloc(sizeof(struct Simbol));
	sprintf (element->nom, nom);
	encuar(colisions, element);
	
	return element;
}

// Inserta un simbol a la taula, retorna false si ya existeix.
Boolean inserir_simbol_en_taula(Taula_Simbols *simbols, Simbol *simbol) {
	
	// Comprobació de que ja existeixi un simbol igual
	int posicio_hash = dispersio(simbol->nom);
	
	if (simbols -> taula[posicio_hash] == NULL) {
		simbols -> taula[posicio_hash] = (Cua *)crear_cua();
	}
	Cua *colisions = simbols -> taula[posicio_hash];
	
	reiniciar_PDI(colisions);
	Simbol *simbol_tmp = (Simbol *)obtenir_element_PDI(colisions);
	while(simbol_tmp != NULL) {
		if (strcmp(simbol_tmp->nom, simbol->nom) == 0) {
			// Ja existeix un simbol amb aquest nom
			return FALSE;
		}
		seguent_PDI(colisions);
		simbol_tmp = (Simbol *)obtenir_element_PDI(colisions);
	}
	
	// No s'ha trobat cap simbol igual
	encuar(colisions, simbol);
	
	return TRUE;
}


// Cerca un simbol
Boolean comprovar_simbol_existent(Taula_Simbols *simbols, char *nom) {
	Simbol *simbol;
	int posicio_hash = dispersio(nom);
	if (simbols -> taula[posicio_hash] == NULL) {
		// No hi ha cua creada, no hi ha cap element
		return FALSE;
	}
	Cua *colisions = simbols -> taula[posicio_hash];
	reiniciar_PDI(colisions);
	simbol = (Simbol *)obtenir_element_PDI(colisions);
	while(simbol != NULL) {
		if (strcmp(simbol -> nom, nom) == 0) {
			return TRUE;
		}
		seguent_PDI(colisions);
		simbol = (Simbol *)obtenir_element_PDI(colisions);
	}
	return FALSE;
}

// Cerca un simbol, retorna NULL si no existeix o el simbol en cas contrari
Simbol* obtenir_simbol(Taula_Simbols *simbols, char *nom) {
	Simbol *simbol;
	int posicio_hash = dispersio(nom);
	if (simbols -> taula[posicio_hash] == NULL) {
		// No hi ha cua creada, no hi ha cap element
		return NULL;
	}
	Cua *colisions = simbols -> taula[posicio_hash];
	reiniciar_PDI(colisions);
	simbol = (Simbol *)obtenir_element_PDI(colisions);
	while(simbol != NULL) {
		if (strcmp(simbol -> nom, nom) == 0) {
			return simbol;
		}
		seguent_PDI(colisions);
		simbol = (Simbol *)obtenir_element_PDI(colisions);
	}
	return NULL;
}

// Cerca un simbol a totes les taules
Simbol* busca_totes_taules(Taula_Simbols *t, char *nom) {
	int index;
	Boolean fi = FALSE;
	Simbol *element;
	Taula_Simbols *taula = t;

	while (!fi) {
		// hash sobre name
		index = dispersio(nom);
		
		// busquem l'element
		if (taula->taula[index] != NULL) {
			Cua *colisions = taula->taula[index];
			reiniciar_PDI(colisions);
			element = (Simbol *)obtenir_element_PDI(colisions);
			while(element != NULL) {
				if (strcmp(element -> nom, nom) == 0) {
					return element;
				}
				seguent_PDI(colisions);
				element = (Simbol *)obtenir_element_PDI(colisions);
			}
		}

		if (taula->anterior == NULL) {
			fi = TRUE;
		} else {
			taula = taula->anterior;
		}
	}
	return NULL;
}

Taula_Simbols *obtenir_TS_programa_principal(Taula_Simbols *taula_actual)
{
	Taula_Simbols *punter_taula = taula_actual;
	while(punter_taula -> anterior != NULL) {
		punter_taula = punter_taula -> anterior;
	}
	return punter_taula -> programa_principal;
}

