#ifndef __TAULA_H__
#define __TAULA_H__

#include "global.h"
#include "cua.h" 

typedef enum {
	TIPUS_INTEGER = 0,
	TIPUS_INTEGER_DERIVAT = 1,	// 'integer' amb rang redefinit o element d'enumerat
	TIPUS_FLOAT = 2,
	TIPUS_BOOLEAN = 3,
	TIPUS_CHARACTER = 4, 
	TIPUS_STRING = 5,
	TIPUS_DERIVAT = 6	// RECORD, ENUMERATED, ACCESS, ARRAY
	//TIPUS_ERROR = 9
} Tipus_Dades;


typedef enum {
	VARIABLE = 0, 
	TYPE = 1,	// e.g. Tipus bàsic boolean, integer... 
	RECORD_TYPE = 2,
	ARRAY_TYPE = 3, 
	ENUMERATED_TYPE = 4,
	//ACCESS_TYPE = 5,	// No hi ha tipus access a les especificacions reduides
	PROCEDURE = 6,
	FUNCTION = 7,
} Tipus_Simbol;

typedef enum {
	START,
	CHSI,
	CHSF,
	I2F,
	F2I,
	I2C,
	C2I,
	ADDI,
	SUBI,
	MULI,
	DIVI,
	MODI,
	ADDF,
	SUBF,
	MULF,
	DIVF,
	ASSIGN,
	ASSIGN_TO_TABLE,
	ASSIGN_FROM_TABLE,
	GOTO,
	IF_EQ,
	IF_NE,
	IF_LTI,
	IF_LEI,
	IF_GTI,
	IF_GEI,
	IF_LTF,
	IF_LEF,
	IF_GTF,
	IF_GEF,
	PARAM,
	CALL_FUNC,
	RETURN_FUNC,
	CALL_PROC,
	RETURN_PROC,
	PUTC,
	PUTI,
	PUTF,
	GETC,
	GETI,
	GETF,
	HALT
} Tipus_Operacio;

typedef struct Tipus_Sentencia {
	Tipus_Operacio operacio;
	char arg1[STRING_LENGTH];
	char arg2[STRING_LENGTH];
	char arg3[STRING_LENGTH];
	Boolean esborrat;	// Si per optimització s'esborra una sentencia, es fica a TRUE
} Tipus_Sentencia;


typedef struct Rang {
	double min;
	double max;
	Tipus_Dades tipus;  // type of the dimension range
	//// No necessitem arrays multidimensionals
} Rang; 

typedef struct Info_Array {
	Tipus_Dades tipus;
	char nom_tipus[STRING_LENGTH]; // Utilitzat per variables que no son de tipus predefinits (integer, float...) i per elements de enumerats
	Rang rang;
} Info_Array;

typedef struct Info_Record {
	struct Taula_Simbols *elements;
	Cua *llista_elements;
} Info_Record;

typedef struct Info_Enumerated {
	Cua *elements;
} Info_Enumerated;

typedef struct Info_Function {
	char nom_unic[STRING_LENGTH];	// Utilitzat pel codi de 3 adreces
	int id; // Identificador de les diferentes sobrecarregues
	int num_sobrecarregues; // Numero total de sobrecarregues (només inicialitzat al info_function principal)
	struct Taula_Simbols *taula;
	Cua *parametres;	// Cua d'strings amb els IDs corresponents als parametres (els simbols es troben a la taula)
	int num_parametres;
	Tipus_Dades tipus_retorn;
	char nom_tipus_retorn[STRING_LENGTH]; // Utilitzat per variables que no son de tipus predefinits (integer, float...) i per elements de enumerats
	Boolean complet;	// Definicio i cos complet de la funcio ja indicat?
	Boolean return_indicat;	// S'ha ficat un return al cos de la funció?
	Cua *sobrecarregues;	// Cua d'info_functions de sobrecarregats
	
	Cua *variables_locals;	// Cua de simbols de variables locals
	Cua *variables_temporals;	// Cua de simbols de variables temporals
	
	Cua *sentencies;	// Cua d'elements tipus_sentencia (Codi 3 adreces)
} Info_Function;

typedef struct Info_Procedure {
	char nom_unic[STRING_LENGTH];	// Utilitzat pel codi de 3 adreces
	int id; // Identificador de les diferentes sobrecarregues
	int num_sobrecarregues; // Numero total de sobrecarregues (només inicialitzat al info_procedure principal)
	struct Taula_Simbols *taula;
	Cua *parametres;	// Cua d'strings amb els IDs corresponents als parametres (els simbols es troben a la taula)
	int num_parametres;
	Boolean complet;	// Definicio i cos complet del procedure ja indicat?
	Cua *sobrecarregues;	// Cua d'info_procedures de sobrecarregats
	
	Cua *variables_locals;	// Cua de simbols de variables locals
	Cua *variables_temporals;	// Cua de simbols de variables temporals
	
	Cua *sentencies;	// Cua d'elements tipus_sentencia (Codi 3 adreces)
} Info_Procedure;

typedef union {
	Boolean boolean;
	int integer;
	char string[CONTENIDOR_STRING_LENGTH];
	char character;
	double vfloat;
	struct Taula_Simbols *taula;	// Si la variable es del tipus RECORD, aquí es guarden les variables del RECORD
} Valor;

typedef struct Info_Variable {
	Tipus_Dades tipus;
	Boolean constant;
	Boolean literal;		// Variable amb nom != literal, valor = literal
	Boolean inicialitzat;
	Valor valor;
	char nom_tipus[STRING_LENGTH]; // Utilitzat per variables que no son de tipus predefinits (integer, float...) i per elements de enumerats
	Boolean offset_literal;
	int offset; // Utilitzat per arrays, guardem la posició a la que es vol accedir (e.g. vector(1), posicio = 1)
	char nom_offset[STRING_LENGTH];
	struct Simbol *offset_simbol;  // Simbol amb el tipus de dada que s'espera al offset indicat
} Info_Variable;

typedef struct Info_Type {
	Tipus_Dades tipus;
	Rang rang;
	int tamany;		// Bytes que ocuparà
} Info_Type;

typedef union {
	Info_Type info_type;
	Info_Variable info_variable;
	Info_Procedure info_procedure;
	Info_Function info_function;
	Info_Enumerated info_enumerated;
	Info_Array info_array;
	Info_Record info_record;
} Continguts;


typedef struct {
	Tipus_Simbol tipus;
	Continguts continguts;
} Atributs;


typedef struct Simbol {
	char nom[STRING_LENGTH];
	Cua *sentencies;	// Cua d'elements tipus_sentencia (Codi 3 adreces)
	Cua *llista_cert;	// Back patching
	Cua *llista_fals;	// Back patching
	Cua *llista_seguents;	// Back patching
	Atributs atributs;
} Simbol;

/* Estructura de la taula de hash. */ 
typedef struct Taula_Simbols {
	Cua *taula[TAM_TAULA_SIMBOLS];
	struct Taula_Simbols *anterior; 
	char nom[STRING_LENGTH];
	struct Taula_Simbols *programa_principal;
} Taula_Simbols;


Simbol* crear_i_inserir_en_taula(Taula_Simbols *simbols, char *nom);
Boolean inserir_simbol_en_taula(Taula_Simbols *simbols, Simbol *simbol);
void inserir_tipus_basics(Taula_Simbols *taula);
Simbol* inserir_en_taula_simbols(Taula_Simbols *simbols, char *nom);
Boolean comprovar_simbol_existent(Taula_Simbols *simbols, char *nom);
Simbol* obtenir_simbol(Taula_Simbols *simbols, char *nom);
Simbol* busca_totes_taules(Taula_Simbols *t, char *nom);


#endif

