
/************************************************************************/
/*                                                                      */
/*  Practica de Compiladors II                                          */
/*  Enginyeria en Informatica - Curs 2004 / 2005                        */
/*                                                                      */
/*  Fitxer sintactic.y: conte la implementacio de l'analitzador         */
/*                      sintactic desenvolupat amb Bison.               */
/*                                                                      */
/*  Autors: Sergio Blanco Cuaresma                                      */
/*          Francisco Jose Aguilar Celdran                              */
/*                                                                      */
/************************************************************************/

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "taula.h"
#include "global.h"
#include <stdarg.h>
#include "pila.h"

FILE *logsSintactic;
extern int lineNumber;
extern int colNumber;
void sinLog(const char *format, ...);
void sinError(const char *format, ...);
extern Taula_Simbols *taulaSimbolActual, *taulaSimbolPrincipal;

extern int calcular_tamany_tipus(Taula_Simbols *taula_actual, char *nom_tipus_derivat);
extern int obtenir_tamany_tipus_basic(Tipus_Dades tipus_variable);
extern int obtenir_offset_fins_element_record(Simbol *record, char *nom_element, Taula_Simbols *taula_actual);

Cua *cua_tmp, *cua_strings_tmp, *cua_defining_identifier_tmp;
int num_elements_cua_strings_tmp;
Taula_Simbols *taula_tmp;
int contador;  // Utilitzat per contar enumerats a la definicio, parametres a la crida de funcions/procediments
int contador_noms_unics;	// Util per assignar noms unics als procediments/funcions. Usat en codi 3 adreces.

Cua *cua_record_tmp;

// Les 6 seguents variables son utils per poder accedir a les cues de variables temporals/locals de funcions i procediments
// Necessari per poder construir el registre d'activacio
Info_Function *infoFunctionActual;  // Apunta al info_function de la funcio que esta sent tratada. Nomes s'accedeix quan es troba un RETURN. Util per quan es troba un RETURN al cos, es pugui marcar el simbol de la funcio a TRUE
Info_Procedure *infoProcedureActual;
Pila *pila_cues_var_locals;
Pila *pila_cues_var_tmp;
Cua *cua_actual_var_locals;
Cua *cua_actual_var_tmp;

Cua *codi;
int seguent_quadruple;

int num_temporal;	// Numero del temporal d'un procediment/funcio. Usat per poder generar codi 3 adreces.


Simbol *cercar_simbol_a_cua(Cua *cua, char *nom);
Taula_Simbols *intentar_duplicar_taula_simbols (Taula_Simbols *ts);
Boolean parametres_diferents(Cua *strings_parametres1, Cua *strings_parametres2, Taula_Simbols *parametres1, Taula_Simbols *parametres2);
Boolean crida_parametres_diferents(Cua *valors, Cua *strings_parametres, Taula_Simbols *parametres);
void comprovacio_complets(Taula_Simbols *ts);
void sinLog(const char *format, ...);
void sinError(const char *format, ...);
void completa_goto(Cua *codi, Cua *posicions, int valor);
void emite(Cua *codi, Tipus_Operacio operacio, char *arg1, char *arg2, char *arg3);
char *nou_temp();
void assignar_programa(Taula_Simbols *taula_actual, Taula_Simbols *principal);
void marcar_sentencies_esborrades(Cua *codi, Cua* posicions);
void incrementar_elements(Cua * posicions);

%}

%union {
	// Inicialitzat des de flex
    Simbol *simbol_ptr; 		// Si el simbol de l'identificador ja existeix, el referenciem
	Simbol simbol; 				// Valors que pasem cap adalt en l'arbre sintactic (bison).
    double  vfloat; 	// Valor numeric decimal
    int     vinteger;		// Valor numeric senser
    char    string[CONTENIDOR_STRING_LENGTH];	// Valor caracters, strings o nom de nou identificador
	char    character;	// Valor caracters, strings o nom de nou identificador
}

%token <string>	PR_ID
%token <simbol_ptr>		PR_ID_VARIABLE
%token <simbol_ptr>		PR_ID_FUNCTION
%token <simbol_ptr>		PR_ID_PROCEDURE
%token <simbol_ptr>		PR_ID_TYPE

%token <vinteger>  	PR_NUMERIC_LITERAL
%token <vfloat>		PR_DECIMAL_LITERAL
%token <character>    	PR_CHARACTER_LITERAL
%token <string>       PR_STRING_LITERAL

// El lexic no retornara cap valor als seguents tokens
%token <simbol_ptr>        PR_ABS
%token <simbol_ptr>        PR_ACCESS
%token <simbol_ptr>        PR_ALL
%token <simbol_ptr>        PR_AND
%token <simbol_ptr>        PR_ARRAY
%token <simbol_ptr>        PR_BEGIN
%token <simbol_ptr>        PR_CASE
%token <simbol_ptr>        PR_CONSTANT
%token <simbol_ptr>        PR_DECLARE
%token <simbol_ptr>        PR_IS
%token <simbol_ptr>        PR_ELSE
%token <simbol_ptr>        PR_ELSIF
%token <simbol_ptr>        PR_END
%token <simbol_ptr>        PR_EXIT
%token <simbol_ptr>        PR_FOR
%token <simbol_ptr>        PR_FUNCTION
%token <simbol_ptr>        PR_IF
%token <simbol_ptr>        PR_IN
%token <simbol_ptr>        PR_LOOP
%token <simbol_ptr>        PR_MOD
%token <simbol_ptr>        PR_NEW
%token <simbol_ptr>        PR_NOT
%token <simbol_ptr>        PR_NULL
%token <simbol_ptr>        PR_OF
%token <simbol_ptr>        PR_OR
%token <simbol_ptr>        PR_OTHERS
%token <simbol_ptr>        PR_PROCEDURE
%token <simbol_ptr>        PR_RANGE
%token <simbol_ptr>        PR_RECORD
%token <simbol_ptr>        PR_RETURN
%token <simbol_ptr>        PR_REVERSE
%token <simbol_ptr>        PR_SUBTYPE
%token <simbol_ptr>        PR_THEN
%token <simbol_ptr>        PR_TYPE
%token <simbol_ptr>        PR_WHEN
%token <simbol_ptr>        PR_WHILE

%token <simbol_ptr>	    PR_EQU
%token <simbol_ptr>    	PR_ARROW
%token <simbol_ptr>    	PR_MAJEQU
%token <simbol_ptr>    	PR_MINEQU
%token <simbol_ptr>		PR_DIFFERENT
%token <simbol_ptr>    	PR_EXP
%token <simbol_ptr>		PR_2_POINTS

// Reduce/Shift expected conflicts
%expect 2

%left '-' '+'
%left '*' '/'
%right PR_EXP

%left PR_IF
%left PR_THEN
%left PR_ELSE
%left PR_ELSIF


%type <simbol> type_definition enumeration_type_definition enumeration_literal_specification record_definition record_type_definition range direct_name name primary factor term simple_expression component_definition range_constraint scalar_constraint subtype_mark subtype_indication defining_identifier_element constraint indexed_component prefix relation expression selected_component type_conversion function_call array_type_definition integer_type_definition signed_integer_type_definition constrained_array_definition discrete_subtype_definition parameter_and_result_profile subprogram_specification procedure_call_statement designator subprogram_body_element program_element condition
%type <string> defining_identifier defining_character_literal selector_name defining_program_unit_name defining_designator
%type <vinteger> actual_parameter_part_element function_call_element mark

%type <simbol> sequence_of_statements if_statement n_mark loop_statement iteration_scheme loop_parameter_specification compound_statement statement sequence_of_statements_element

%start program

%{
    //#define YYERROR_VERBOSE 1
%}

%%

basic_declaration:
   type_declaration { sinLog(" -> basic_declaration -> type_declaration"); } 
   | object_declaration {sinLog(" -> basic_declaration -> object_declaration");} 
   | subprogram_declaration {sinLog(" -> basic_declaration -> subprogram_declaration");}
;

subprogram_declaration:
   subprogram_specification ';' {sinLog(" -> subprogram_declaration -> subprogram_specification");}
;

subprogram_specification:
   // Inicialitzar taula simbols entre PR_PROCEDURE i defining_program_unit_name (nou ambit)
   // Controlar errors de procedures/functions ja declarats o incomplets (son valids)
   // El final de la branca de 'defining_program_unit_name' es un PR_ID
   PR_PROCEDURE defining_program_unit_name {
	   // Nova definicio
	   taula_tmp = (Taula_Simbols *)crear_taula_simbols($2, taulaSimbolActual);
	   cua_strings_tmp = (Cua *)crear_cua();
	   num_elements_cua_strings_tmp = 0;
   }
   parameter_profile {
	   sinLog("subprogram_specification -> PR_PROCEDURE defining_program_unit_name parameter_profile");
	   Simbol *sym = crear_i_inserir_en_taula(taulaSimbolActual, taula_tmp->nom);
	   sym->atributs.tipus = PROCEDURE;
	   sprintf(sym->atributs.continguts.info_procedure.nom_unic, "%s_%i", $2, contador_noms_unics++);
	   sym->atributs.continguts.info_procedure.id = 0;
	   sym->atributs.continguts.info_procedure.num_sobrecarregues = 0;
	   sym->atributs.continguts.info_procedure.complet = FALSE;
	   sym->atributs.continguts.info_procedure.num_parametres = num_elements_cua_strings_tmp;
	   num_elements_cua_strings_tmp = 0;
	   sym->atributs.continguts.info_procedure.parametres = cua_strings_tmp;
	   cua_strings_tmp = NULL;
	   sym->atributs.continguts.info_procedure.taula = taula_tmp;
	   taula_tmp = NULL;
	   sym->atributs.continguts.info_procedure.sobrecarregues = NULL;
	   sym->atributs.continguts.info_procedure.variables_locals = (Cua*)crear_cua();
	   sym->atributs.continguts.info_procedure.variables_temporals = (Cua*)crear_cua();
	   
	   infoProcedureActual = &(sym->atributs.continguts.info_procedure);
	   
	   // Tornem el nou simbol
	   memcpy(&$$, sym, sizeof(struct Simbol));
   }
   | PR_PROCEDURE PR_ID_PROCEDURE {
	   // NOTE: Possible redefinicio o sobrecarrega
	   Info_Procedure *info_procedure = (Info_Procedure*)malloc(sizeof(struct Info_Procedure)); 
	   taula_tmp = (Taula_Simbols *)crear_taula_simbols($2, taulaSimbolActual);
	   cua_strings_tmp = (Cua *)crear_cua();
	   num_elements_cua_strings_tmp = 0;
   }
   parameter_profile {
	   sinLog("subprogram_specification -> PR_PROCEDURE PR_ID_PROCEDURE parameter_profile");
	   Boolean trobada_sobrecarrega_definida = FALSE;
	   Info_Procedure *info_procedure_sobrecarregat = &($2->atributs.continguts.info_procedure); // Apuntara al info de la sobrecarrega ja definida en cas d'existir
	   int total_sobrecarregues = info_procedure_sobrecarregat->num_sobrecarregues;
	   
	   // Busquem el Info_Function amb mateix parametres que el procedure trobat per la regla
	   // 1.- Comparem amb el Info_Function principal
	   if (num_elements_cua_strings_tmp == info_procedure_sobrecarregat->num_parametres) {
		   // Num. parametres iguals
		   if (parametres_diferents(info_procedure_sobrecarregat->parametres, cua_strings_tmp, info_procedure_sobrecarregat->taula, taula_tmp) == FALSE) {
			   // Tipus parametres iguals
			   trobada_sobrecarrega_definida = TRUE;
		   }
	   }

	   // 2.- Comparem amb els Info_Functions del llistat de sobrecarregues
	   if (trobada_sobrecarrega_definida == FALSE) {
		   if ($2->atributs.continguts.info_procedure.sobrecarregues != NULL) {
			   reiniciar_PDI($2->atributs.continguts.info_procedure.sobrecarregues);
			   info_procedure_sobrecarregat = (Info_Procedure *)obtenir_element_PDI($2->atributs.continguts.info_procedure.sobrecarregues);		   
			   while (info_procedure_sobrecarregat != NULL) {
				   if (num_elements_cua_strings_tmp == info_procedure_sobrecarregat->num_parametres) {
					   // Num. parametres iguals
					   if (parametres_diferents(info_procedure_sobrecarregat->parametres, cua_strings_tmp, info_procedure_sobrecarregat->taula, taula_tmp) == FALSE) {
						   // Tipus parametres iguals
						   trobada_sobrecarrega_definida = TRUE;
						   break; // No cal continuar buscant
					   }
				   }
				   seguent_PDI($2->atributs.continguts.info_procedure.sobrecarregues);
				   info_procedure_sobrecarregat = (Info_Procedure *)obtenir_element_PDI($2->atributs.continguts.info_procedure.sobrecarregues);
			   }
		   }
	   }
	   
	   // Si no hem trobat cap sobrecarrega igual
	   if (trobada_sobrecarrega_definida == FALSE) {
		   // Afegim nova sobrecarrega
		   if ($2->atributs.continguts.info_procedure.sobrecarregues == NULL) {
			   $2->atributs.continguts.info_procedure.sobrecarregues = (Cua *)crear_cua();
		   }
		   $2->atributs.continguts.info_procedure.num_sobrecarregues++;
		   
		   Info_Procedure *info_procedure = (Info_Procedure*)malloc(sizeof(struct Info_Procedure));
		   sprintf(info_procedure->nom_unic, "%s_%i", $2->nom, contador_noms_unics++);
		   info_procedure->id = total_sobrecarregues + 1;
		   info_procedure->complet = FALSE;
		   info_procedure->num_parametres = num_elements_cua_strings_tmp;
		   num_elements_cua_strings_tmp = 0;
		   info_procedure->parametres = cua_strings_tmp;
		   cua_strings_tmp = NULL;
		   info_procedure->taula = taula_tmp;
		   taula_tmp = NULL;
		   info_procedure->variables_locals = (Cua*)crear_cua();
		   info_procedure->variables_temporals = (Cua*)crear_cua();
		   
		   encuar($2->atributs.continguts.info_procedure.sobrecarregues, info_procedure);
		   
		   infoProcedureActual = info_procedure;
		   
		   // Retornem el simbol original amb els Info_Function de la sobrecarrega nova
		   memcpy(&$$, $2, sizeof(struct Simbol));
		   memcpy(&($$.atributs.continguts.info_procedure), info_procedure , sizeof(struct Info_Procedure));
	   } else {
		   // Hem trobat la sobrecarrega ya definida, no cal definir-la de nou
		   // TODO: Alliberar taula_tmp i cua_strings_tmp
		   // Retornem el simbol original amb els Info_Function de la sobrecarrega trobada
		   memcpy(&$$, $2, sizeof(struct Simbol));
		   memcpy(&($$.atributs.continguts.info_procedure), info_procedure_sobrecarregat , sizeof(struct Info_Procedure));
		   
		   infoProcedureActual = info_procedure_sobrecarregat;
	   }
   }
   | PR_FUNCTION defining_designator {
	   taula_tmp = (Taula_Simbols *)crear_taula_simbols($2, taulaSimbolActual);
	   cua_strings_tmp = (Cua *)crear_cua();
	   num_elements_cua_strings_tmp = 0;
   }
   parameter_and_result_profile {
	   sinLog("subprogram_specification -> PR_PROCEDURE defining_program_unit_name parameter_and_result_profile");
	   Simbol *sym = crear_i_inserir_en_taula(taulaSimbolActual, taula_tmp->nom);
	   sym->atributs.tipus = FUNCTION;
	   sprintf(sym->atributs.continguts.info_function.nom_unic, "%s_%i", $2, contador_noms_unics++);
	   sym->atributs.continguts.info_function.id = 0;
	   sym->atributs.continguts.info_function.num_sobrecarregues = 0;
	   sym->atributs.continguts.info_function.complet = FALSE;
	   sym->atributs.continguts.info_function.return_indicat = FALSE;
	   sym->atributs.continguts.info_function.num_parametres = num_elements_cua_strings_tmp;
	   num_elements_cua_strings_tmp = 0;
	   sym->atributs.continguts.info_function.parametres = cua_strings_tmp;
	   cua_strings_tmp = NULL;
	   sym->atributs.continguts.info_function.taula = taula_tmp;
	   taula_tmp = NULL;
	   sym->atributs.continguts.info_function.sobrecarregues = NULL;
	   sym->atributs.continguts.info_function.variables_locals = (Cua*)crear_cua();
	   sym->atributs.continguts.info_function.variables_temporals = (Cua*)crear_cua();
	   
	   // $4 es 'parameter_and_result_profile'
	   if ($4.atributs.tipus == TYPE) {
			// Tipus basic
			sym->atributs.continguts.info_function.tipus_retorn = $4.atributs.continguts.info_type.tipus;
			if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
				memcpy(sym->atributs.continguts.info_function.nom_tipus_retorn, $4.nom, STRING_LENGTH);
			}
	   } else if ($4.atributs.tipus == ENUMERATED_TYPE) {
			// A partir del nom podem extreure el simbol del tipus
			sym->atributs.continguts.info_function.tipus_retorn = TIPUS_DERIVAT;
			memcpy(sym->atributs.continguts.info_function.nom_tipus_retorn, $4.nom, STRING_LENGTH);
	   } else if ($4.atributs.tipus == RECORD_TYPE) {
			sym->atributs.continguts.info_function.tipus_retorn = TIPUS_DERIVAT;
			memcpy(sym->atributs.continguts.info_function.nom_tipus_retorn, $4.nom, STRING_LENGTH);
	   } else if ($4.atributs.tipus == ARRAY_TYPE) {
			sym->atributs.continguts.info_function.tipus_retorn = TIPUS_DERIVAT;
			if (strcmp($4.nom, "") == 0) {
				// NOTE: Considerem que no poden haver-hi tipus anonims a la signatura.
				sinError("No poden haver-hi tipus anonims a la signatura.");
			} else {
				// Array ya definit
				memcpy(sym->atributs.continguts.info_function.nom_tipus_retorn, $4.nom, STRING_LENGTH);
			}
	   } else {
			sinError("Sense tipus valid de retorn! %i", $4.atributs.tipus);
	   }
	   
	   // Util per quan es troba un RETURN al cos, es pugui marcar el simbol de la funcio a TRUE
	   infoFunctionActual = &(sym->atributs.continguts.info_function);
	   
	   // Tornem el nou simbol
	   memcpy(&$$, sym, sizeof(struct Simbol));
   }
   | PR_FUNCTION PR_ID_FUNCTION {
	   // NOTE: Posible redefinicio o sobrecarrega
	   taula_tmp = (Taula_Simbols *)crear_taula_simbols($2, taulaSimbolActual);
	   cua_strings_tmp = (Cua *)crear_cua();
	   num_elements_cua_strings_tmp = 0;
   }
   parameter_and_result_profile {
	   sinLog("subprogram_specification -> PR_FUNCTION PR_ID_FUNCTION parameter_and_result_profile");
	   Boolean trobada_sobrecarrega_definida = FALSE;
	   Info_Function *info_function_sobrecarregat = &($2->atributs.continguts.info_function); // Apuntara al info de la sobrecarrega ja definida en cas d'existir
	   int total_sobrecarregues = info_function_sobrecarregat->num_sobrecarregues;
	   
	   // Busquem el info_function amb mateix parametres que la function trobada per la regla
	   // 1.- Comparem amb el info_function principal
	   if (num_elements_cua_strings_tmp == info_function_sobrecarregat->num_parametres) {
		   // Num. parametres iguals
		   if (parametres_diferents(info_function_sobrecarregat->parametres, cua_strings_tmp, info_function_sobrecarregat->taula, taula_tmp) == FALSE) {
			   // Tipus parametres iguals
			   trobada_sobrecarrega_definida = TRUE;
			   
			   // NOTE: No suportem sobrecarrega de tipus de retorn
			   // El tipus retornat ha de ser igual, sino generem error
			   if ($4.atributs.tipus == TYPE) {
				   if ($4.atributs.continguts.info_type.tipus == info_function_sobrecarregat->tipus_retorn) {
					   if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
						   if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) != 0) {
							   // Tipus retorn no iguals
							   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
						   }
					   }
				   } else {
					   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
				   }
			   } else if (($4.atributs.tipus == ENUMERATED_TYPE) || ($4.atributs.tipus == RECORD_TYPE)
			   				|| ($4.atributs.tipus == ARRAY_TYPE)) {
				   		if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) != 0) {
							// Tipus retorn no iguals
							   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
						}
			   } else {
					   sinError("Sense tipus valid de retorn! %i", $4.atributs.tipus);
			   }
			   
			   
			   // Comprovar tipus retorn
			   /*if ($4.atributs.tipus == TYPE) {
				   if ($4.atributs.continguts.info_type.tipus == info_function_sobrecarregat->tipus_retorn) {
					   if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
						   if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) == 0) {
							   // Tipus retorn iguals
							   trobada_sobrecarrega_definida = TRUE;
						   }
					   } else {
						   // Tipus retorn iguals
						   trobada_sobrecarrega_definida = TRUE;
					   }
				   }
			   } else if (($4.atributs.tipus == ENUMERATED_TYPE) || ($4.atributs.tipus == RECORD_TYPE)
			   				|| ($4.atributs.tipus == ARRAY_TYPE)) {
				   		if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) == 0) {
							// Tipus retorn iguals
							trobada_sobrecarrega_definida = TRUE;
						}
			   } else {
					   sinError("Sense tipus valid de retorn! %i", $4.atributs.tipus);
			   }*/
		   }
	   }

	   // 2.- Comparem amb els info_function del llistat de sobrecarregues
	   if (trobada_sobrecarrega_definida == FALSE) {
		   if ($2->atributs.continguts.info_function.sobrecarregues != NULL) {
			   reiniciar_PDI($2->atributs.continguts.info_function.sobrecarregues);
			   info_function_sobrecarregat = (Info_Function *)obtenir_element_PDI($2->atributs.continguts.info_function.sobrecarregues);		   
			   while (info_function_sobrecarregat != NULL) {
				   if (num_elements_cua_strings_tmp == info_function_sobrecarregat->num_parametres) {
					   // Num. parametres iguals
					   
					   if (parametres_diferents(info_function_sobrecarregat->parametres, cua_strings_tmp, info_function_sobrecarregat->taula, taula_tmp) == FALSE) {
						   // Tipus parametres iguals
						   trobada_sobrecarrega_definida = TRUE;
			   
						   // NOTE: No suportem sobrecarrega de tipus de retorn
						   // El tipus retornat ha de ser igual, sino generem error
						   if ($4.atributs.tipus == TYPE) {
							   if ($4.atributs.continguts.info_type.tipus == info_function_sobrecarregat->tipus_retorn) {
								   if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
									   if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) != 0) {
										   // Tipus retorn no iguals
										   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
									   }
								   }
							   } else {
								   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
							   }
						   } else if (($4.atributs.tipus == ENUMERATED_TYPE) || ($4.atributs.tipus == RECORD_TYPE)
										|| ($4.atributs.tipus == ARRAY_TYPE)) {
									if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) != 0) {
										// Tipus retorn no iguals
										   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
									}
						   } else {
								   sinError("El tipus de retorn de la funcio no coincideix amb funcio '%s' anteriorment definida.", $2->nom);
						   }
						   
						   // NOTE: No suportem sobrecarrega de tipus de retorn
						   /*// Comprovar tipus retorn
						   if ($4.atributs.tipus == TYPE) {
							   if ($4.atributs.continguts.info_type.tipus == info_function_sobrecarregat->tipus_retorn) {
								   if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
									   if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) == 0) {
										   // Tipus retorn iguals
										   trobada_sobrecarrega_definida = TRUE;
										   break; // No cal continuar buscant
									   }
								   } else {
									   // Tipus retorn iguals
									   trobada_sobrecarrega_definida = TRUE;
									   break; // No cal continuar buscant
								   }
							   }
						   } else if (($4.atributs.tipus == ENUMERATED_TYPE) || ($4.atributs.tipus == RECORD_TYPE)
										|| ($4.atributs.tipus == ARRAY_TYPE)) {
									if (strcmp($4.nom, info_function_sobrecarregat->nom_tipus_retorn) == 0) {
										// Tipus retorn iguals
										trobada_sobrecarrega_definida = TRUE;
										break; // No cal continuar buscant
									}
						   } else {
								   sinError("Sense tipus valid de retorn! %i", $4.atributs.tipus);
						   }*/
						   
					   }
				   }
				   seguent_PDI($2->atributs.continguts.info_function.sobrecarregues);
				   info_function_sobrecarregat = (Info_Function *)obtenir_element_PDI($2->atributs.continguts.info_function.sobrecarregues);
			   }
		   }
	   }
	   
	   // Si no hem trobat cap sobrecarrega igual
	   if (trobada_sobrecarrega_definida == FALSE) {
		   // Afegim nova sobrecarrega
		   if ($2->atributs.continguts.info_function.sobrecarregues == NULL) {
			   $2->atributs.continguts.info_function.sobrecarregues = (Cua *)crear_cua();
		   }
		   
		   $2->atributs.continguts.info_function.num_sobrecarregues++;
		   
		   Info_Function *info_function = (Info_Function*)malloc(sizeof(struct Info_Function));
		   sprintf(info_function->nom_unic, "%s_%i", $2->nom, contador_noms_unics++);
		   info_function->id = total_sobrecarregues + 1;
		   info_function->complet = FALSE;
		   info_function->return_indicat = FALSE;
		   info_function->num_parametres = num_elements_cua_strings_tmp;
		   num_elements_cua_strings_tmp = 0;
		   info_function->parametres = cua_strings_tmp;
		   cua_strings_tmp = NULL;
		   info_function->taula = taula_tmp;
		   taula_tmp = NULL;
		   info_function->variables_locals = (Cua*)crear_cua();
		   info_function->variables_temporals = (Cua*)crear_cua();
		   
		   // $4 es 'parameter_and_result_profile'
		   if ($4.atributs.tipus == TYPE) {
				// Tipus basic
				info_function->tipus_retorn = $4.atributs.continguts.info_type.tipus;
				if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
					memcpy(info_function->nom_tipus_retorn, $4.nom, STRING_LENGTH);
				}
		   } else if ($4.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				info_function->tipus_retorn = TIPUS_DERIVAT;
				memcpy(info_function->nom_tipus_retorn, $4.nom, STRING_LENGTH);
		   } else if ($4.atributs.tipus == RECORD_TYPE) {
				info_function->tipus_retorn = TIPUS_DERIVAT;
				memcpy(info_function->nom_tipus_retorn, $4.nom, STRING_LENGTH);
		   } else if ($4.atributs.tipus == ARRAY_TYPE) {
				info_function->tipus_retorn = TIPUS_DERIVAT;
				if (strcmp($4.nom, "") == 0) {
					// NOTE: Considerem que no poden haver-hi tipus anonims a la signatura.
					sinError("No poden haver-hi tipus anonims a la signatura.");
				} else {
					// Array ya definit
					memcpy(info_function->nom_tipus_retorn, $4.nom, STRING_LENGTH);
				}
		   } else {
				sinError("Sense tipus valid de retorn! %i", $4.atributs.tipus);
		   }
		   
		   encuar($2->atributs.continguts.info_function.sobrecarregues, info_function);
		   
		   // Util per quan es troba un RETURN al cos, es pugui marcar el simbol de la funcio a TRUE
		   infoFunctionActual = info_function;
		   
		   // Retornem el simbol original amb els info_function de la sobrecarrega nova
		   memcpy(&$$, $2, sizeof(struct Simbol));
		   memcpy(&($$.atributs.continguts.info_function), info_function , sizeof(struct Info_Function));
	   } else {
		   // Hem trobat la sobrecarrega ya definida, no cal definir-la de nou
		   // TODO: Alliberar taula_tmp i cua_strings_tmp
		   // Retornem el simbol original amb els Info_Function de la sobrecarrega trobada
		   memcpy(&$$, $2, sizeof(struct Simbol));
		   memcpy(&($$.atributs.continguts.info_function), info_function_sobrecarregat , sizeof(struct Info_Function));
		   
		   infoFunctionActual = info_function_sobrecarregat;
	   }
   }
;

defining_designator:
   defining_program_unit_name {
	    sinLog(" -> defining_designator -> defining_program_unit_name");
		// Pasem el nom de l'identificador cap adalt
		snprintf($$, STRING_LENGTH, "%s" , $1);
   }
;

parameter_profile:
   formal_part {sinLog(" -> parameter_profile -> formal_part");}
   | {sinLog(" -> parameter_profile");}
;

parameter_and_result_profile:
   formal_part PR_RETURN subtype_mark {
	   sinLog("parameter_and_result_profile -> formal_part PR_RETURN subtype_mark");
	   memcpy(&$$, &$3, sizeof(struct Simbol));
   }
   | PR_RETURN subtype_mark {
	   sinLog("parameter_and_result_profile -> PR_RETURN subtype_mark");
	   memcpy(&$$, &$2, sizeof(struct Simbol));
   }
;

formal_part:
   '(' parameter_specification formal_part_element {sinLog(" -> formal_part -> '(' parameter_specification formal_part_element");}
;

formal_part_element:
   ';' parameter_specification formal_part_element { sinLog(" -> formal_part_element -> ';' parameter_specification formal_part_element"); }
   | ';' parameter_specification {sinLog(" -> formal_part_element -> parameter_specification");}
   | ')' {sinLog(" -> formal_part_element -> ')'");}
;

parameter_specification:
   defining_identifier_list ':' subtype_mark {
	    // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp); // Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			if ($3.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $3.atributs.continguts.info_type.tipus;
				if ($3.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
					memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
				}
			} else if ($3.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else if ($3.atributs.tipus == RECORD_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else if ($3.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				if (strcmp($3.nom, "") == 0) {
					// NOTE: Considerem que no poden haver-hi tipus anonims a la signatura.
					sinError("No poden haver-hi tipus anonims a la signatura.");
					/*// Array definit in situ (TIPUS ANONIM)
					
					// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
					Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
					sprintf(symTipus->nom, "$%s", element->nom);
					memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
					inserir_simbol_en_taula(taula_tmp, symTipus);
					
					// Associem la variable al tipus anterior
					memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);*/
				} else {
					// Array ya definit
					memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
				}
			} else {
				sinError("Sense tipus valid! %i", $3.atributs.tipus);
			}
			
			// Encuem el nom del parametre per tindre un llistat de parametres de procediment/funcio
			encuar(cua_strings_tmp, element->nom);
			num_elements_cua_strings_tmp++;
			
			inserir_simbol_en_taula(taula_tmp, element);
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
;

object_declaration:
   defining_identifier_list ':' PR_CONSTANT subtype_indication PR_EQU expression ';' {
	    // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = TRUE;
			element->atributs.continguts.info_variable.inicialitzat = TRUE;
			
			if ($4.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $4.atributs.continguts.info_type.tipus;
				
				// La expressio de $5 mai sera TIPUS_INTEGER_DERIVAT, s'accepta si la variable es integer derivat i
				// es comprova els rangs mes tard.
				if (!(($6.atributs.continguts.info_type.tipus == TIPUS_INTEGER)
					&& ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT))) {
				
					if ($4.atributs.continguts.info_type.tipus != $6.atributs.continguts.info_variable.tipus) {
						sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
					}
				}
				
				
				if ($6.atributs.continguts.info_variable.literal == TRUE) {
					switch ($4.atributs.continguts.info_type.tipus) {
						// NOTE: La comprovacio de rangs es fa al lexic excepte pels INTEGER_DERIVAT
						case TIPUS_INTEGER:
							element->atributs.continguts.info_variable.valor.integer = $6.atributs.continguts.info_variable.valor.integer;
							break;
						case TIPUS_INTEGER_DERIVAT:
							if (($6.atributs.continguts.info_variable.valor.integer < (int)$4.atributs.continguts.info_type.rang.min)
									|| ($6.atributs.continguts.info_variable.valor.integer > (int)$4.atributs.continguts.info_type.rang.max)) {
								sinError("Valor fora de rang, ha d'estar entre %i i %i.", (int)$4.atributs.continguts.info_type.rang.min, (int)$4.atributs.continguts.info_type.rang.max);
							}
							if (strcmp($4.nom, "") == 0) {
								// Subtipus integer definit in situ (TIPUS ANONIM)
					
								// Afegim subtipus integer definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
								Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
								sprintf(symTipus->nom, "$%s", element->nom);
								memcpy(&(symTipus->atributs), &($4.atributs), sizeof(Atributs));
								inserir_simbol_en_taula(taulaSimbolActual, symTipus);
								
								// Associem la variable al tipus anterior
								memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
							} else {
								// Subtipus integer ja definit
								memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
							}
							element->atributs.continguts.info_variable.valor.integer = $6.atributs.continguts.info_variable.valor.integer;
							break;
						case TIPUS_FLOAT:
							element->atributs.continguts.info_variable.valor.vfloat = $6.atributs.continguts.info_variable.valor.vfloat;
							break;
						case TIPUS_BOOLEAN:
							element->atributs.continguts.info_variable.valor.boolean = $6.atributs.continguts.info_variable.valor.boolean;
							break;
						case TIPUS_CHARACTER:
							element->atributs.continguts.info_variable.valor.character = $6.atributs.continguts.info_variable.valor.character;
							break;
						case TIPUS_STRING:
							snprintf(element->atributs.continguts.info_variable.valor.string, CONTENIDOR_STRING_LENGTH, "%s", $6.atributs.continguts.info_variable.valor.string);
							break;
						case TIPUS_DERIVAT:
							// Es possible trobar-se en aquest punt? Crec que no.
							sinError("Comprovar error.");
							break;
					}
				}
			} else if ($4.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
				
				Simbol *symEnumElement = busca_totes_taules(taulaSimbolActual, $6.nom);
				if (symEnumElement == NULL) {
					sinError("El valor que s'intenta assignar no existeix com enumerador.");
				}
				
				if (strcmp($4.nom, symEnumElement->atributs.continguts.info_variable.nom_tipus) != 0) {
					sinError("El valor que s'intenta assignar no pertany al mateix enumerat que defineix el tipus de la variable.");
				}
				
				// NOTE: A una variable del tipus enumerat guardem a 'integer' el valor correponent al element del enumerat assignat
				element->atributs.continguts.info_variable.valor.integer = symEnumElement->atributs.continguts.info_variable.valor.integer;
				
			} else if ($4.atributs.tipus == RECORD_TYPE) {
				// NOTE: No permetem inicialitzacio de RECORDs a la declaracio
				sinError("No es pot inicialitzar un tipus RECORD a la seva declaracio.");
			} else if ($4.atributs.tipus == ARRAY_TYPE) {
				// NOTE: No permetem inicialitzacio de ARRAY a la declaracio
				sinError("No es pot inicialitzar un tipus ARRAY a la seva declaracio.");
			} else {
				sinError("Sense tipus valid! %i", $4.atributs.tipus);
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' PR_CONSTANT subtype_indication PR_EQU expression ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' subtype_indication PR_EQU expression ';' {
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = FALSE;
			element->atributs.continguts.info_variable.inicialitzat = TRUE;
			
			if ($3.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $3.atributs.continguts.info_type.tipus;
				
				// La expressio de $5 mai sera TIPUS_INTEGER_DERIVAT, s'accepta si la variable es integer derivat i
				// es comprova els rangs mes tard.
				if (!(($5.atributs.continguts.info_type.tipus == TIPUS_INTEGER)
					&& ($3.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT))) {
				
					if ($3.atributs.continguts.info_type.tipus != $5.atributs.continguts.info_variable.tipus) {
						sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
					}
				}
				
				if ($5.atributs.continguts.info_variable.literal == TRUE) {
					switch ($3.atributs.continguts.info_type.tipus) {
						// NOTE: La comprovacio de rangs es fa al lexic excepte pels INTEGER_DERIVAT
						case TIPUS_INTEGER:
							element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
							break;
						case TIPUS_INTEGER_DERIVAT:
							if (($5.atributs.continguts.info_variable.valor.integer < (int)$3.atributs.continguts.info_type.rang.min)
									|| ($5.atributs.continguts.info_variable.valor.integer > (int)$3.atributs.continguts.info_type.rang.max)) {
								sinError("Valor fora de rang, ha d'estar entre %i i %i.", (int)$3.atributs.continguts.info_type.rang.min, (int)$3.atributs.continguts.info_type.rang.max);
							}
							if (strcmp($3.nom, "") == 0) {
								// Subtipus integer definit in situ (TIPUS ANONIM)
					
								// Afegim subtipus integer definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
								Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
								sprintf(symTipus->nom, "$%s", element->nom);
								memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
								inserir_simbol_en_taula(taulaSimbolActual, symTipus);
								
								// Associem la variable al tipus anterior
								memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
							} else {
								// Subtipus integer ja definit
								memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
							}
							element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
							break;
						case TIPUS_FLOAT:
							element->atributs.continguts.info_variable.valor.vfloat = $5.atributs.continguts.info_variable.valor.vfloat;
							break;
						case TIPUS_BOOLEAN:
							element->atributs.continguts.info_variable.valor.boolean = $5.atributs.continguts.info_variable.valor.boolean;
							break;
						case TIPUS_CHARACTER:
							element->atributs.continguts.info_variable.valor.character = $5.atributs.continguts.info_variable.valor.character;
							break;
						case TIPUS_STRING:
							snprintf(element->atributs.continguts.info_variable.valor.string, CONTENIDOR_STRING_LENGTH, "%s", $5.atributs.continguts.info_variable.valor.string);
							break;
						case TIPUS_DERIVAT:
							// Es possible trobar-se en aquest punt? Crec que no.
							sinError("Comprovar error.");
							break;
					}
				}
			} else if ($3.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
				
				Simbol *symEnumElement = busca_totes_taules(taulaSimbolActual, $5.nom);
				if (symEnumElement == NULL) {
					sinError("El valor que s'intenta assignar no existeix com enumerador.");
				}
				
				if (strcmp($3.nom, symEnumElement->atributs.continguts.info_variable.nom_tipus) != 0) {
					sinError("El valor que s'intenta assignar no pertany al mateix enumerat que defineix el tipus de la variable.");
				}
				
				// NOTE: A una variable del tipus enumerat guardem a 'integer' el valor correponent al element del enumerat assignat
				element->atributs.continguts.info_variable.valor.integer = symEnumElement->atributs.continguts.info_variable.valor.integer;
				
			} else if ($3.atributs.tipus == RECORD_TYPE) {
				// NOTE: No permetem inicialitzacio de RECORDs a la declaracio
				sinError("No es pot inicialitzar un tipus RECORD a la seva declaracio.");
			} else if ($3.atributs.tipus == ARRAY_TYPE) {
				// NOTE: No permetem inicialitzacio de ARRAY a la declaracio
				sinError("No es pot inicialitzar un tipus ARRAY a la seva declaracio.");
			} else {
				sinError("Sense tipus valid! %i", $3.atributs.tipus);
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' subtype_indication PR_EQU expression ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' PR_CONSTANT subtype_indication ';' {
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = TRUE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($4.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $4.atributs.continguts.info_type.tipus;
				if ($4.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
					if (strcmp($4.nom, "") == 0) {
						// Subtipus integer definit in situ (TIPUS ANONIM)
			
						// Afegim subtipus integer definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
						Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
						sprintf(symTipus->nom, "$%s", element->nom);
						memcpy(&(symTipus->atributs), &($4.atributs), sizeof(Atributs));
						inserir_simbol_en_taula(taulaSimbolActual, symTipus);
						
						// Associem la variable al tipus anterior
						memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
					} else {
						// Subtipus integer ja definit
						memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
					}
				}
			} else if ($4.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
			} else if ($4.atributs.tipus == RECORD_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
				// Dupliquem la taula de simbols del RECORD, donat que aquesta conte les variables del RECORD
				// i les podrem utilitzar per guardar valors.
				element->atributs.continguts.info_variable.valor.taula = intentar_duplicar_taula_simbols($4.atributs.continguts.info_record.elements);
				
				if (element->atributs.continguts.info_variable.valor.taula == NULL) {
					// NOTE: Comprovar que funciona.
					sinError("Imposible duplicar taula!");
				}
			} else if ($4.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Arrays anonims tractats a una altra branca, aqui sempre es troben ja definits
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $4.nom, STRING_LENGTH);
			} else {
				sinError("Sense tipus valid! %i", $4.atributs.tipus);
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' PR_CONSTANT subtype_indication ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' subtype_indication ';' {
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = FALSE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($3.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $3.atributs.continguts.info_type.tipus;
				if ($3.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
					if (strcmp($3.nom, "") == 0) {
						// Subtipus integer definit in situ (TIPUS ANONIM)
			
						// Afegim subtipus integer definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
						Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
						sprintf(symTipus->nom, "$%s", element->nom);
						memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
						inserir_simbol_en_taula(taulaSimbolActual, symTipus);
						
						// Associem la variable al tipus anterior
						memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
					} else {
						// Subtipus integer ja definit
						memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
					}
				}
			} else if ($3.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else if ($3.atributs.tipus == RECORD_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
				// Dupliquem la taula de simbols del RECORD, donat que aquesta conte les variables del RECORD
				// i les podrem utilitzar per guardar valors.
				element->atributs.continguts.info_variable.valor.taula = intentar_duplicar_taula_simbols($3.atributs.continguts.info_record.elements);
				
				if (element->atributs.continguts.info_variable.valor.taula == NULL) {
					// NOTE: Comprovar que funciona.
					sinError("Imposible duplicar taula!");
				}
			} else if ($3.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Arrays anonims tractats a una altra branca, aqui sempre es troben ja definits
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else {
				sinError("Sense tipus valid! %i", $3.atributs.tipus);
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' subtype_indication ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' PR_CONSTANT array_type_definition PR_EQU expression ';' {
	   // ARRAY anonim
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = TRUE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($4.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Array definit in situ (TIPUS ANONIM)
				// $3.nom sempre ha de ser igual a "" en aquest cas
					
				// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
				Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
				sprintf(symTipus->nom, "$%s", element->nom);
				memcpy(&(symTipus->atributs), &($4.atributs), sizeof(Atributs));
				inserir_simbol_en_taula(taulaSimbolActual, symTipus);
				
				// Associem la variable al tipus anterior
				memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
			} else {
				sinError("Nomes s'espera un tipus array anonim aqui.");
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
		
		
		if ($4.atributs.continguts.info_array.tipus != $6.atributs.continguts.info_variable.tipus) {
			sinError("El tipus assignat no es del mateix tipus que l'array.");
		}
		
		// Assignacio de valors
		switch ($6.atributs.continguts.info_array.tipus) {
			// TODO: Assignar valor a totes les posicions del array
			case TIPUS_INTEGER:
				// Buscar integer a la taula de simbols i comprovar rangs
				//element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
				break;
			case TIPUS_INTEGER_DERIVAT:
				// Buscar subtipus integer a la taula de simbols i comprovar rangs
				/*if (($6.atributs.continguts.info_variable.valor.integer < $4.atributs.continguts.info_type.rang.min)
						|| ($6.atributs.continguts.info_variable.valor.integer > $4.atributs.continguts.info_type.rang.max)) {
					sinError("Valor fora de rang, ha d'estar entre %i i %i.", $4.atributs.continguts.info_type.rang.min, $4.atributs.continguts.info_type.rang.max);
				}*/
				//element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
				break;
			case TIPUS_FLOAT:
				//element->atributs.continguts.info_variable.valor.vfloat = $6.atributs.continguts.info_variable.valor.vfloat;
				break;
			case TIPUS_BOOLEAN:
				//element->atributs.continguts.info_variable.valor.boolean = $6.atributs.continguts.info_variable.valor.boolean;
				break;
			case TIPUS_CHARACTER:
				//element->atributs.continguts.info_variable.valor.character = $6.atributs.continguts.info_variable.valor.character;
				break;
			case TIPUS_STRING:
				//snprintf(element->atributs.continguts.info_variable.valor.string, STRING_LENGTH, "%s", $6.atributs.continguts.info_variable.valor.string);
				break;
			case TIPUS_DERIVAT:
				// NOTE: No es permeten arrays de tipus que no siguin basics (integer, character, float, boolean i string).
				sinError("No es permeten arrays de tipus que no siguin basics (integer, character, float, boolean i string).");
				break;
		}
		
   }
   | defined_identifier ':' PR_CONSTANT array_type_definition PR_EQU expression ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' array_type_definition PR_EQU expression ';' {
	   // ARRAY anonim
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = FALSE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($3.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Array definit in situ (TIPUS ANONIM)
				// $3.nom sempre ha de ser igual a "" en aquest cas
					
				// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
				Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
				sprintf(symTipus->nom, "$%s", element->nom);
				memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
				inserir_simbol_en_taula(taulaSimbolActual, symTipus);
				
				// Associem la variable al tipus anterior
				memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
			} else {
				sinError("Nomes s'espera un tipus array anonim aqui.");
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
		
		if ($3.atributs.continguts.info_array.tipus != $5.atributs.continguts.info_variable.tipus) {
			sinError("El tipus assignat no es del mateix tipus que l'array.");
		}
		
		// Assignacio de valors
		switch ($5.atributs.continguts.info_array.tipus) {
			// TODO: Assignar valor a totes les posicions del array
			case TIPUS_INTEGER:
				// Buscar integer a la taula de simbols i comprovar rangs
				//element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
				break;
			case TIPUS_INTEGER_DERIVAT:
				// Buscar subtipus integer a la taula de simbols i comprovar rangs
				/*if (($5.atributs.continguts.info_variable.valor.integer < $3.atributs.continguts.info_type.rang.min)
						|| ($5.atributs.continguts.info_variable.valor.integer > $3.atributs.continguts.info_type.rang.max)) {
					sinError("Valor fora de rang, ha d'estar entre %i i %i.", $3.atributs.continguts.info_type.rang.min, $3.atributs.continguts.info_type.rang.max);
				}*/
				//element->atributs.continguts.info_variable.valor.integer = $5.atributs.continguts.info_variable.valor.integer;
				break;
			case TIPUS_FLOAT:
				//element->atributs.continguts.info_variable.valor.vfloat = $5.atributs.continguts.info_variable.valor.vfloat;
				break;
			case TIPUS_BOOLEAN:
				//element->atributs.continguts.info_variable.valor.boolean = $5.atributs.continguts.info_variable.valor.boolean;
				break;
			case TIPUS_CHARACTER:
				//element->atributs.continguts.info_variable.valor.character = $5.atributs.continguts.info_variable.valor.character;
				break;
			case TIPUS_STRING:
				//snprintf(element->atributs.continguts.info_variable.valor.string, STRING_LENGTH, "%s", $5.atributs.continguts.info_variable.valor.string);
				break;
			case TIPUS_DERIVAT:
				// NOTE: No es permeten arrays de tipus que no siguin basics (integer, character, float, boolean i string).
				sinError("No es permeten arrays de tipus que no siguin basics (integer, character, float, boolean i string).");
				break;
		}
   }
   | defined_identifier ':' array_type_definition PR_EQU expression ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' PR_CONSTANT array_type_definition ';' {
	   // ARRAY anonim
	   // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = TRUE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($4.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Array definit in situ (TIPUS ANONIM)
				// $3.nom sempre ha de ser igual a "" en aquest cas
					
				// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
				Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
				sprintf(symTipus->nom, "$%s", element->nom);
				memcpy(&(symTipus->atributs), &($4.atributs), sizeof(Atributs));
				inserir_simbol_en_taula(taulaSimbolActual, symTipus);
				
				// Associem la variable al tipus anterior
				memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
			} else {
				sinError("Nomes s'espera un tipus array anonim aqui.");
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' PR_CONSTANT array_type_definition ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
   | defining_identifier_list ':' array_type_definition ';' {
	    // ARRAY anonim
	    // Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp);	// Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			element->atributs.continguts.info_variable.constant = FALSE;
			element->atributs.continguts.info_variable.inicialitzat = FALSE;
			
			if ($3.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				// Array definit in situ (TIPUS ANONIM)
				// $3.nom sempre ha de ser igual a "" en aquest cas
					
				// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
				Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
				sprintf(symTipus->nom, "$%s", element->nom);
				memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
				inserir_simbol_en_taula(taulaSimbolActual, symTipus);
				
				// Associem la variable al tipus anterior
				memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
			} else {
				sinError("Nomes s'espera un tipus array anonim aqui.");
			}
			
			inserir_simbol_en_taula(taulaSimbolActual, element);
			encuar(cua_actual_var_locals, element); // Llistat de variables locals
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
   }
   | defined_identifier ':' array_type_definition ';' {
	   // Controlar errors de variables definides amb mateix nom
	   // Saltara l'error a defined_identifier
   }
;

type_declaration:
   full_type_declaration {sinLog(" -> type_declaration -> full_type_declaration");}
;

full_type_declaration:
	// NOTE: Segons l'enunciat no sembla que sagi de soportar els tipus incomplets
   PR_TYPE defining_identifier PR_IS type_definition ';' {
	   Simbol *sym, *sym_tmp;
	   int i;
	   
	   sinLog(" -> full_type_declaration -> PR_TYPE defining_identifier PR_IS type_definition ';' ");

	   sinLog("El tipus complet \"%s\" s'ha insertat a la taula de simbols.", $2);
	   sym = crear_i_inserir_en_taula(taulaSimbolActual, $2);
	   //sym->atributs.continguts.info_type.tipus = TIPUS_BOOLEAN;
	   sym->atributs.tipus = $4.atributs.tipus;
	   
	   if (sym->atributs.tipus == ENUMERATED_TYPE) {
			sym->atributs.continguts.info_enumerated.elements = $4.atributs.continguts.info_enumerated.elements;
			
			Cua *elements = sym->atributs.continguts.info_enumerated.elements;
			reiniciar_PDI(elements);
			Simbol *element = (Simbol *)obtenir_element_PDI(elements);
			int posicio = 0;
			while(element != NULL) {
				if (posicio > 255) {
					// NOTE: Segons l'enunciat un enumerat es de 8 bits, no poden haver mes de 256 elements
					sinError("Un enumerat no pot tindre mes de 256 elements.");
				}
				// Cada element del enumerat s'inserta com un simbol de tipus variable a la TS
				sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, element -> nom);
				memcpy(&(sym_tmp->atributs), &(element->atributs), sizeof(Atributs));
				sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER_DERIVAT;
				sym_tmp->atributs.continguts.info_variable.literal = TRUE;
				sym_tmp->atributs.continguts.info_variable.constant = TRUE;
				sym_tmp->atributs.continguts.info_variable.valor.integer = posicio;
				// Es lliga al tipus enumerat:
				sprintf(sym_tmp->atributs.continguts.info_variable.nom_tipus, "%s", $2);
				posicio++;
				seguent_PDI(elements);
				element = (Simbol *)obtenir_element_PDI(elements);
			}
			
		}
		
		if (sym->atributs.tipus == TYPE) {
			// Tipus 'integer' amb un rang redefinit
			sym->atributs.continguts.info_type.tipus = $4.atributs.continguts.info_type.tipus;
			sym->atributs.continguts.info_type.rang.min = $4.atributs.continguts.info_type.rang.min;
			sym->atributs.continguts.info_type.rang.max = $4.atributs.continguts.info_type.rang.max;
			sym->atributs.continguts.info_type.rang.tipus = $4.atributs.continguts.info_type.rang.tipus;
		}
		
		if (sym->atributs.tipus == RECORD_TYPE) {
			sym->atributs.continguts.info_record.elements = $4.atributs.continguts.info_record.elements;
			sym->atributs.continguts.info_record.llista_elements = $4.atributs.continguts.info_record.llista_elements;
		}
		
		if (sym->atributs.tipus == ARRAY_TYPE) {
			memcpy(&(sym->atributs), &($4.atributs), sizeof(Atributs));
		}
		
	}
	
	| PR_TYPE defined_identifier PR_IS type_definition ';' {
		// Controlar errors de tipus definits amb el mateix nom
		// Saltara l'error a defined_identifier
	}
;

type_definition:
   enumeration_type_definition { 
	   sinLog(" -> type_definition -> enumeration_type_definition");
	   $$.atributs.tipus = $1.atributs.tipus; // ENUMERATED_TYPE
	   $$.atributs.continguts.info_enumerated.elements = $1.atributs.continguts.info_enumerated.elements;
   } 
   | integer_type_definition {
	   sinLog(" -> type_definition -> integer_type_definition");
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   } 
   | array_type_definition {
	   sinLog(" -> type_definition -> array_type_definition");
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   } 
   | record_type_definition {
	   sinLog("record_type_definition -> type_definition");
	   $$.atributs.tipus = RECORD_TYPE;
	   $$.atributs.continguts.info_record.elements = $1.atributs.continguts.info_record.elements;
	   $$.atributs.continguts.info_record.llista_elements = $1.atributs.continguts.info_record.llista_elements;
   }
;

record_type_definition:
   record_definition {}
;

record_definition:
   // Potser cal inicialitzar taula de simbols entre PR_RECORD i component_list
   PR_RECORD {
	   taula_tmp = (Taula_Simbols *)crear_taula_simbols("RECORD", taulaSimbolActual);
	   cua_record_tmp = (Cua*)crear_cua();
   }
   component_list PR_END PR_RECORD {
	   sinLog(" -> record_definition -> T_RECORD component_list T_END T_RECORD");
	   $$.atributs.continguts.info_record.elements = taula_tmp;
	   $$.atributs.continguts.info_record.llista_elements = cua_record_tmp;
	   taula_tmp = NULL;
	   cua_record_tmp = NULL;
   }
;

component_list:
   component_item {}
   | component_item component_list {}
;

component_item:
   component_declaration {}
;

component_declaration:
	defining_identifier_list ':' component_definition ';' {
		// Treure de la cua i ficar a la TS temporal
		reiniciar_PDI(cua_defining_identifier_tmp); // Cua inicialitzada i muntada a defining_identifier_list
		Simbol *element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		while (element != NULL) {
			if ($3.atributs.tipus == TYPE) {
				// Tipus basic
				element->atributs.continguts.info_variable.tipus = $3.atributs.continguts.info_type.tipus;
				if ($3.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
					
					if (strcmp($3.nom, "") == 0) {
						// Subtipus integer definit in situ (TIPUS ANONIM)
			
						// Afegim subtipus integer definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
						Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
						sprintf(symTipus->nom, "$%s", element->nom);
						memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
						inserir_simbol_en_taula(taulaSimbolActual, symTipus);
						
						// Associem la variable al tipus anterior
						memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
					} else {
						// Subtipus integer ja definit
						memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
					}
				}
				
			} else if ($3.atributs.tipus == ENUMERATED_TYPE) {
				// A partir del nom podem extreure el simbol del tipus
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else if ($3.atributs.tipus == RECORD_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
			} else if ($3.atributs.tipus == ARRAY_TYPE) {
				element->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
				if (strcmp($3.nom, "") == 0) {
					// Array definit in situ (TIPUS ANONIM)
					
					// Afegim array definit in situ a la TS com si fos un tipus array definit amb el nom "$" + "nom_variable"
					Simbol *symTipus = (Simbol*)malloc(sizeof(struct Simbol));
					sprintf(symTipus->nom, "$%s", element->nom);
					memcpy(&(symTipus->atributs), &($3.atributs), sizeof(Atributs));
					inserir_simbol_en_taula(taulaSimbolActual, symTipus);
					
					// Associem la variable al tipus anterior
					memcpy(element->atributs.continguts.info_variable.nom_tipus, symTipus->nom, STRING_LENGTH);
				} else {
					// Array ya definit
					memcpy(element->atributs.continguts.info_variable.nom_tipus, $3.nom, STRING_LENGTH);
				}
			} else {
				sinError("Sense tipus valid! %i", $3.atributs.tipus);
			}
			
			inserir_simbol_en_taula(taula_tmp, element);
			encuar(cua_record_tmp, element);
			seguent_PDI(cua_defining_identifier_tmp);
			element = (Simbol *)obtenir_element_PDI(cua_defining_identifier_tmp);
		}
		cua_defining_identifier_tmp = NULL;
	}
;

defining_identifier_list:
   // Insertem tots els identificadors del mateix tipus a una cua per despres
   // (quan sapiguem de quin tipus son) insertar-los tots a la taula de simbols del record
   defining_identifier defining_identifier_element {
	   if (cua_defining_identifier_tmp == NULL) {
		   // RECORDATORI: Quan acabem d'utilitzar la cua, hem d'assignar-la a NULL
		   cua_defining_identifier_tmp = (Cua *)crear_cua();
	   }
	   sinLog(" -> defining_identifier_list -> defining_identifier defining_identifier_element");
	   if (cercar_simbol_a_cua(cua_defining_identifier_tmp, $1) == NULL) {
		   Simbol *sym = (Simbol*)malloc(sizeof(struct Simbol));
		   memcpy(sym->nom, $1, STRING_LENGTH); 
		   sym->atributs.tipus = VARIABLE;
		   sym->atributs.continguts.info_variable.valor.taula = NULL;
		   encuar(cua_defining_identifier_tmp, sym);
	   } else {
		   sinError("Redefinicio del simbol %s", $1);
	   }
   }
;

defining_identifier_element: /* Gramatica auxiliar de 'defining_identifier_list' */
   ',' defining_identifier defining_identifier_element {
	   if (cua_defining_identifier_tmp == NULL) {
		   // RECORDATORI: Quan acabem d'utilitzar la cua, hem d'assignar-la a NULL
		   cua_defining_identifier_tmp = (Cua *)crear_cua();
	   }
	   sinLog(" -> defining_identifier_element -> ',' defining_identifier defining_identifier_element");
	   if (cercar_simbol_a_cua(cua_defining_identifier_tmp, $2) == NULL) {
		   Simbol *sym = (Simbol*)malloc(sizeof(struct Simbol));
		   memcpy(sym->nom, $2, STRING_LENGTH); 
		   sym->atributs.tipus = VARIABLE;
		   sym->atributs.continguts.info_variable.valor.taula = NULL;
		   encuar(cua_defining_identifier_tmp, sym);
	   } else {
		   sinError("Redefinicio del simbol %s", $2);
	   }
   }
   | {}
;

array_type_definition:
   constrained_array_definition {
	    sinLog(" -> array_type_definition -> constrained_array_definition");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

constrained_array_definition:
	// De discrete_subtype_definition ens arriba un ARRAY_TYPE
   PR_ARRAY '(' discrete_subtype_definition ')' PR_OF component_definition {
	   sinLog(" -> array_type_definition -> PR_ARRAY '(' discrete_subtype_definition ')' PR_OF component_definition");
	   memcpy(&$$, &$3, sizeof(struct Simbol));
	   
	   
	   $$.atributs.continguts.info_array.tipus = $6.atributs.tipus; // ARRAY_TYPE
	   if (($6.atributs.tipus == TYPE) && ($6.atributs.continguts.info_type.tipus == TIPUS_DERIVAT)) {
		   // NOTE: Comprovar que aquest cas no es pot donar
		   sinError("No pot ser TYPE i TIPUS_DERIVAT");
		   //snprintf($$.atributs.continguts.info_array.nom_tipus, STRING_LENGTH, "%s", $6.nom);
	   
	   } else if (($6.atributs.tipus == TYPE) && ($6.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT)) {
		   $$.atributs.continguts.info_array.tipus = $6.atributs.continguts.info_type.tipus;
		   snprintf($$.atributs.continguts.info_array.nom_tipus, STRING_LENGTH, "%s", $6.nom);
	   } else if ($6.atributs.tipus == TYPE) {
		   $$.atributs.continguts.info_array.tipus = $6.atributs.continguts.info_type.tipus;
	   } else if ($6.atributs.tipus == ENUMERATED_TYPE) {
	   		$$.atributs.continguts.info_array.tipus = TIPUS_DERIVAT;
		   snprintf($$.atributs.continguts.info_array.nom_tipus, STRING_LENGTH, "%s", $6.nom);
	   } else if (($6.atributs.tipus == RECORD_TYPE) || ($6.atributs.tipus == ARRAY_TYPE)) {
		   //sinError("No es possible definir un array de tipus array o record");
		   $$.atributs.continguts.info_array.tipus = TIPUS_DERIVAT;
		   snprintf($$.atributs.continguts.info_array.nom_tipus, STRING_LENGTH, "%s", $6.nom);
	   }
   }
;

discrete_subtype_definition:
   subtype_indication {
	    sinLog(" -> discrete_subtype_definition -> subtype_indication");
		
		// A $1 hauriem de tindre TYPE, TIPUS_INTEGER.
		// Retornarem un ARRAY_TYPE
		if ($1.atributs.tipus == TYPE) {
			if (($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT)) {
				// Agafem el rang del tipus
				sprintf($$.nom, "");	// No es encara un simbol de la TS
				$$.atributs.tipus = ARRAY_TYPE;
				$$.atributs.continguts.info_array.rang.min = $1.atributs.continguts.info_type.rang.min;
				$$.atributs.continguts.info_array.rang.max = $1.atributs.continguts.info_type.rang.max;
				$$.atributs.continguts.info_array.rang.tipus = $1.atributs.continguts.info_type.rang.tipus;
			} else {
				sinError("El tipus indicat no te cap rang definit (nomes es poden usar tipus 'integer' o derivats).");
			}
		} else {
			sinError("El tipus indicat no te cap rang definit (nomes es poden usar tipus 'integer' o derivats).");
		}
		
		
	} /*discrete_subtype_indication {}*/
   | range {
	    sinLog(" -> discrete_subtype_definition -> range");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

subtype_indication:
	// Expected shift/reduce conflict
	subtype_mark {
		sinLog(" -> subtype_indication -> subtype_mark");
		memcpy(&$$, &$1, sizeof(struct Simbol));
	}
	| subtype_mark constraint {
		sinLog(" -> subtype_indication -> subtype_mark constraint");
		/*if ($1.fullType==FALSE)
			sinLog("The incomplete %s type can not be constrained",$1.name_type);*/
		
		memcpy(&$$, &$2, sizeof(struct Simbol)); // ARRAY_TYPE
		
		if ($1.atributs.tipus != TYPE) {
			sinError("Nomes es poden especificar subrangs a partir del tipus 'integer'.");
		}
		if ($1.atributs.continguts.info_type.tipus != TIPUS_INTEGER) {
			sinError("Nomes es poden especificar subrangs a partir del tipus 'integer'.");
		}
		
		// En 'contraint' ($2) ens arriba un ARRAY_TYPE
		sprintf($$.nom, "");	// No es encara un simbol de la TS
		$$.atributs.tipus = TYPE;
		$$.atributs.continguts.info_type.tipus = TIPUS_INTEGER_DERIVAT;
		$$.atributs.continguts.info_type.rang.min = $2.atributs.continguts.info_array.rang.min;
		$$.atributs.continguts.info_type.rang.max = $2.atributs.continguts.info_array.rang.min;
		$$.atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;
	}
;

subtype_mark:
   PR_ID_TYPE {
	   // El lexic haura trobat el simbol corresponent a la TS i ens l'ha retornat a $1
	   // Passem cap adalt.
	   sinLog(" -> subtype_mark -> T_ID_TYPE"); 
	   memcpy(&$$, $1, sizeof(struct Simbol));
   }
   /*subtype_name {}*/
;

constraint:
   scalar_constraint {
	   sinLog(" -> constraint -> scalar_constraint");
	   memcpy(&$$, &$1, sizeof(struct Simbol)); // ARRAY_TYPE
   }
;

scalar_constraint:
   range_constraint {
	   sinLog(" -> scalar_constraint -> range_constraint");
	   memcpy(&$$, &$1, sizeof(struct Simbol)); // ARRAY_TYPE
   }
;

range_constraint:
   PR_RANGE range {
	    sinLog("-> range_constraint -> PR_RANGE range");
		memcpy(&$$, &$2, sizeof(struct Simbol)); // ARRAY_TYPE
   }
;

component_definition:
   subtype_indication {
	    sinLog(" -> component_definition -> subtype_indication");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

integer_type_definition:
   signed_integer_type_definition {
	    sinLog(" -> integer_type_definition -> signed_integer_type_definition");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

signed_integer_type_definition:
   PR_RANGE simple_expression PR_2_POINTS simple_expression {
   		sinLog(" -> signed_integer_type_definition -> PR_RANGE simple_expression PR_2_POINTS simple_exprTession");
		if (($2.atributs.continguts.info_variable.literal == FALSE)
			|| ($4.atributs.continguts.info_variable.literal == FALSE)) {
			
			sinError("Nomes es permeten expressions literals a la definicio de rangs d'enters.");
		}
		
		if (($2.atributs.continguts.info_variable.tipus != TIPUS_INTEGER)
			&& ($4.atributs.continguts.info_variable.tipus != TIPUS_INTEGER)) {
			
			sinError("El limits del rang han de ser de tipus 'integer'.");
		}
		
		if ($2.atributs.continguts.info_variable.valor.integer > $4.atributs.continguts.info_variable.valor.integer) {
			sinError("El limit inferior del rang ha de ser mes petit que el limit superior.");
		}
		
		$$.atributs.tipus = TYPE;
		$$.atributs.continguts.info_type.tipus = TIPUS_INTEGER_DERIVAT;
		$$.atributs.continguts.info_type.rang.min = $2.atributs.continguts.info_variable.valor.integer;
		$$.atributs.continguts.info_type.rang.max = $4.atributs.continguts.info_variable.valor.integer;
		$$.atributs.continguts.info_type.rang.tipus = TIPUS_INTEGER;
   }
		
;

simple_expression:
	// NOTE: Nomes es fan comprovacions de rang pels 'integer'
   '+' term {
	    sinLog(" -> simple_expression -> '+' term ");
		if (($2.atributs.continguts.info_variable.tipus != TIPUS_INTEGER)
			&& ($2.atributs.continguts.info_variable.tipus != TIPUS_INTEGER_DERIVAT)
			&& ($2.atributs.continguts.info_variable.tipus != TIPUS_FLOAT)) {
			sinError("Operador unari '+' no pot ser aplicat a tipus diferent de 'integer' o 'float'.");
		}
		memcpy(&$$, &$2, sizeof(struct Simbol));
   }
   | '-' term {
	    sinLog(" -> simple_expression -> '-' term ");
		if (($2.atributs.continguts.info_variable.tipus != TIPUS_INTEGER)
			&& ($2.atributs.continguts.info_variable.tipus != TIPUS_INTEGER_DERIVAT)
			&& ($2.atributs.continguts.info_variable.tipus != TIPUS_FLOAT)) {
			sinError("Operador unari '-' no pot ser aplicat a tipus diferent de 'integer' o 'float'.");
		}
		memcpy(&$$, &$2, sizeof(struct Simbol));
		
		$$.atributs.continguts.info_variable.literal = FALSE;
		if ($2.atributs.continguts.info_variable.literal) {
			// Invertim signe
			if (($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				// Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			    // ja es tornara a comprovar amb el seu rang especific al assignment
			    Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			    long long int resultat = (long long int)-$2.atributs.continguts.info_variable.valor.integer;
			    if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
					|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("El canvi de signe sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
				}
				$2.atributs.continguts.info_variable.valor.integer = (int)resultat;
			} else if ($2.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				$2.atributs.continguts.info_variable.valor.vfloat = -$2.atributs.continguts.info_variable.valor.vfloat;
			} else {
				sinError("Tipus no valid.");
			}
			$$.atributs.continguts.info_variable.literal = TRUE;
		} else {
			// Codi 3 adreces
			memcpy(&$$, &$2, sizeof(struct Simbol));
			
			if (($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				emite(codi, CHSI, $$.nom, $2.nom, NULL);
				seguent_quadruple++;
			} else if ($2.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				emite(codi, CHSF, $$.nom, $2.nom, NULL);
				seguent_quadruple++;
			} else {
				sinError("Tipus no valid.");
			}
		}
   }
   | term {
   		sinLog(" -> simple_expression -> term");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   | simple_expression '+' term {
	   sinLog(" -> simple_expression -> simple_expression '+' term");
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   // Segons els tipus dels elements, el resultat sera integer o float
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   
			   // Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			   // ja es tornara a comprovar amb el seu rang especific al assignment
			   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			   long long int resultat = (long long int)$1.atributs.continguts.info_variable.valor.integer + $3.atributs.continguts.info_variable.valor.integer;
			   if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
				   	|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("La suma sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
			   }
			   
			   $$.atributs.continguts.info_variable.valor.integer = (int)resultat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // NOTE: No es fa comprovacio de limits per floats donat els problemes de precisio d'aquests tipus a C
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat + $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.integer + $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat + $3.atributs.continguts.info_variable.valor.integer;
		   } else {
			   sinError("L'operador '+' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
	   } else {
		   // Codi 3 adreces
		   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   memcpy(&(sym_tmp->atributs), &($1.atributs), sizeof(Atributs));
		   encuar(cua_actual_var_tmp, sym_tmp);
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ADDI, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ADDI, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, ADDI, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, ADDF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ADDF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, ADDF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, ADDF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("L'operador '+' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   sym_tmp->atributs.tipus = VARIABLE;
		   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
		   memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	   }
   }
   | simple_expression '-' term {
	   sinLog(" -> simple_expression -> simple_expression '-' term");
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   // Segons els tipus dels elements, el resultat sera integer o float
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			   // ja es tornara a comprovar amb el seu rang especific al assignment
			   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			   long long int resultat = (long long int)$1.atributs.continguts.info_variable.valor.integer - $3.atributs.continguts.info_variable.valor.integer;
			   if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
				   	|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("La resta sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
			   }
			   
			   $$.atributs.continguts.info_variable.valor.integer = (int)resultat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat - $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.integer - $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat - $3.atributs.continguts.info_variable.valor.integer;
		   } else {
			   sinError("L'operador '-' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
	   }  else {
		   // Codi 3 adreces
		   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   memcpy(&(sym_tmp->atributs), &($1.atributs), sizeof(Atributs));
		   encuar(cua_actual_var_tmp, sym_tmp);
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, SUBI, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, SUBI, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, SUBI, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, SUBF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, SUBF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, SUBF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, SUBF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("L'operador '-' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   sym_tmp->atributs.tipus = VARIABLE;
		   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
		   memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	   }
   }
   | simple_expression '&' term {
	   // Concatenacio
	   sinLog(" -> simple_expression -> simple_expression '&' term");

	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {	   
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) && ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING)) {
			   // STRING & STRING
			   int totalLength = (unsigned char)($1.atributs.continguts.info_variable.valor.string[0]) + (unsigned char)($3.atributs.continguts.info_variable.valor.string[0]);
			   if ((totalLength) > 255) {
				   sinError("La concatenacio excedeix el 255 caracters.");
			   }
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // NOTE: Comprovar que es realitza be la concatenacio
			   sprintf($$.atributs.continguts.info_variable.valor.string, "%c%s%s", totalLength, &($1.atributs.continguts.info_variable.valor.string[1]), &($3.atributs.continguts.info_variable.valor.string[1]));
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) && ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER)) {
			   // STRING & CHARACTER
			   int totalLength = (unsigned char)($1.atributs.continguts.info_variable.valor.string[0]) + 1;
			   if (totalLength > 255) {
				   sinError("La concatenacio excedeix el 255 caracters.");
			   }
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // NOTE: Comprovar que es realitza be la concatenacio
			   sprintf($$.atributs.continguts.info_variable.valor.string, "%c%s%c", totalLength, &($1.atributs.continguts.info_variable.valor.string[1]), $3.atributs.continguts.info_variable.valor.character);
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING)) {
			   // CHARACTER & STRING
			   int totalLength = (unsigned char)($3.atributs.continguts.info_variable.valor.string[0]) + 1;
			   if (totalLength > 255) {
				   sinError("La concatenacio excedeix el 255 caracters.");
			   }
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // NOTE: Comprovar que es realitza be la concatenacio
			   sprintf($$.atributs.continguts.info_variable.valor.string, "%c%c%s", totalLength, $1.atributs.continguts.info_variable.valor.character, &($3.atributs.continguts.info_variable.valor.string[1]));
		   } else {
			   sinError("No es possible concatenar aquests tipus.");
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) && ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING)) {
			   char string1[STRING_LENGTH];
			   char string2[STRING_LENGTH];
			   
			   // A string1 i string2 fiquem els noms de les 2 variables string a comparar
			   if (($1.atributs.continguts.info_variable.literal == FALSE) && ($3.atributs.continguts.info_variable.literal == FALSE)) {
				   sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
			   } else if ($1.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(string1, "%s", $1.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string2, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }

			   } else if ($3.atributs.continguts.info_variable.literal == FALSE) {
				   //sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string1, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($1.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $1.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   sinError("No es pot donar aquest cas.");
			   }
			   
			   
			   // Generem codi per la concatenacio
			   
			   // Variables temporals que necessitarem
			   Simbol *sym_result = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   sprintf($$.nom, "%s", sym_result->nom);  // Copiar el nom de la variable a on es guardara el resultat, al valor de retorn
			   encuar(cua_actual_var_tmp, sym_result);
			   sym_result->atributs.tipus = VARIABLE;
			   sym_result->atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   sym_result->atributs.continguts.info_variable.literal = FALSE;
			   sym_result->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_contador = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_contador);
			   sym_contador->atributs.tipus = VARIABLE;
			   sym_contador->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_contador->atributs.continguts.info_variable.literal = FALSE;
			   sym_contador->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_length_total = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_length_total);
			   sym_length_total->atributs.tipus = VARIABLE;
			   sym_length_total->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_length_total->atributs.continguts.info_variable.literal = FALSE;
			   sym_length_total->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_length01 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_length01);
			   sym_length01->atributs.tipus = VARIABLE;
			   sym_length01->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_length01->atributs.continguts.info_variable.literal = FALSE;
			   sym_length01->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_length02 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_length02);
			   sym_length02->atributs.tipus = VARIABLE;
			   sym_length02->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_length02->atributs.continguts.info_variable.literal = FALSE;
			   sym_length02->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_char_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_char_tmp);
			   sym_char_tmp->atributs.tipus = VARIABLE;
			   sym_char_tmp->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_char_tmp->atributs.continguts.info_variable.literal = FALSE;
			   sym_char_tmp->atributs.continguts.info_variable.constant = FALSE;
			   
			   // length01 := string1[0];
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_length01->nom, string1, arg3);
			   seguent_quadruple++;
			   
			   // length02 := string2[0];
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_length02->nom, string2, arg3);
			   seguent_quadruple++;
			   
			   // length_total := length02 + length02
			   emite(codi, ADDI, sym_length_total->nom, sym_length01->nom, sym_length02->nom);
			   seguent_quadruple++;
			   
			   //// Comencem a copiar l'string 1
			   // contador := 1;
			   sprintf(arg2, "1");
			   emite(codi, ASSIGN, sym_contador->nom, arg2, NULL);
			   seguent_quadruple++;
			   
			   // if contador = length01 then goto [next]
			   sprintf(arg3, "%i", seguent_quadruple + 5);
			   emite(codi, IF_GTI, sym_contador->nom, sym_length01->nom, arg3);
			   seguent_quadruple++;
			   
			   // temporal := string1[contador];
			   emite(codi, ASSIGN_FROM_TABLE, sym_char_tmp->nom, string1, sym_contador->nom);
			   seguent_quadruple++;
			   
			   // resultat[contador] := temporal
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, sym_contador->nom, sym_char_tmp->nom);
			   seguent_quadruple++;
			   
			   // contador++
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_contador->nom, sym_contador->nom, arg3);
			   seguent_quadruple++;
			   
			   // goto [principi]
			   sprintf(arg1, "%i", seguent_quadruple - 4);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
			   
			   //// Comencem a copiar l'string 2

			   // if contador = length_total then goto [next]
			   sprintf(arg3, "%i", seguent_quadruple + 5);
			   emite(codi, IF_GTI, sym_contador->nom, sym_length_total->nom, arg3);
			   seguent_quadruple++;
			   
			   // temporal := string2[contador];
			   emite(codi, ASSIGN_FROM_TABLE, sym_char_tmp->nom, string2, sym_contador->nom);
			   seguent_quadruple++;
			   
			   // resultat[contador] := temporal
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, sym_contador->nom, sym_char_tmp->nom);
			   seguent_quadruple++;
			   
			   // contador++
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_contador->nom, sym_contador->nom, arg3);
			   seguent_quadruple++;
			   
			   // goto [principi]
			   sprintf(arg1, "%i", seguent_quadruple - 4);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) && ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   
			   char arg1[STRING_LENGTH];
			   char arg2[STRING_LENGTH];
			   char arg3[STRING_LENGTH];
		   
			   char string1[STRING_LENGTH];
			   char caracter[STRING_LENGTH];

			   // A string1 i string2 fiquem els noms de les 2 variables string a comparar
			   if (($1.atributs.continguts.info_variable.literal == FALSE) && ($3.atributs.continguts.info_variable.literal == FALSE)) {
				   sprintf(string1, "%s", $1.nom);
				   sprintf(caracter, "%s", $3.nom);
			   } else if ($1.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(string1, "%s", $1.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_char = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(caracter, "%s", sym_tmp_char->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_char);
				   sym_tmp_char->atributs.tipus = VARIABLE;
				   sym_tmp_char->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
				   sym_tmp_char->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_char->atributs.continguts.info_variable.constant = FALSE;
				   
				   sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				   emite(codi, ASSIGN, sym_tmp_char->nom, arg2, NULL);
				   seguent_quadruple++;

			   } else if ($3.atributs.continguts.info_variable.literal == FALSE) {
				   //sprintf(string1, "%s", $1.nom);
				   sprintf(caracter, "%s", $3.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string1, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($1.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $1.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   sinError("No es pot donar aquest cas.");
			   }
			   
			   // Generem codi per la concatenacio
			   
			   // Variables temporals que necessitarem
			   Simbol *sym_result = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   sprintf($$.nom, "%s", sym_result->nom);  // Copiar el nom de la variable a on es guardara el resultat, al valor de retorn
			   encuar(cua_actual_var_tmp, sym_result);
			   sym_result->atributs.tipus = VARIABLE;
			   sym_result->atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   sym_result->atributs.continguts.info_variable.literal = FALSE;
			   sym_result->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_contador = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_contador);
			   sym_contador->atributs.tipus = VARIABLE;
			   sym_contador->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_contador->atributs.continguts.info_variable.literal = FALSE;
			   sym_contador->atributs.continguts.info_variable.constant = FALSE;
			   
			   
			   Simbol *sym_length01 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_length01);
			   sym_length01->atributs.tipus = VARIABLE;
			   sym_length01->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_length01->atributs.continguts.info_variable.literal = FALSE;
			   sym_length01->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_char_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_char_tmp);
			   sym_char_tmp->atributs.tipus = VARIABLE;
			   sym_char_tmp->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_char_tmp->atributs.continguts.info_variable.literal = FALSE;
			   sym_char_tmp->atributs.continguts.info_variable.constant = FALSE;
			   
			   // length01 := string1[0];
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_length01->nom, string1, arg3);
			   seguent_quadruple++;
			   
			   //// Comencem a copiar l'string 1
			   // contador := 1;
			   sprintf(arg2, "1");
			   emite(codi, ASSIGN, sym_contador->nom, arg2, NULL);
			   seguent_quadruple++;
			   
			   // if contador = length01 then goto [next]
			   sprintf(arg3, "%i", seguent_quadruple + 5);
			   emite(codi, IF_GTI, sym_contador->nom, sym_length01->nom, arg3);
			   seguent_quadruple++;
			   
			   // temporal := string1[contador];
			   emite(codi, ASSIGN_FROM_TABLE, sym_char_tmp->nom, string1, sym_contador->nom);
			   seguent_quadruple++;
			   
			   // resultat[contador] := temporal
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, sym_contador->nom, sym_char_tmp->nom);
			   seguent_quadruple++;
			   
			   // contador++
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_contador->nom, sym_contador->nom, arg3);
			   seguent_quadruple++;
			   
			   // goto [principi]
			   sprintf(arg1, "%i", seguent_quadruple - 4);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
			   
			   //// Afegim caracter
			   // resultat[contador] := caracter
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, sym_contador->nom, caracter);
			   seguent_quadruple++;
			   
			   //// Tamany
			   // length01 := length01 + 1;
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_length01->nom, sym_length01->nom, arg3);
			   seguent_quadruple++;
			   
			   // result[0] := length01;
			   sprintf(arg2, "0");
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, arg2, sym_length01->nom);
			   seguent_quadruple++;
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   
			   char arg1[STRING_LENGTH];
			   char arg2[STRING_LENGTH];
			   char arg3[STRING_LENGTH];
		   
			   char caracter[STRING_LENGTH];
			   char string2[STRING_LENGTH];

			   // A string1 i string2 fiquem els noms de les 2 variables string a comparar
			   if (($1.atributs.continguts.info_variable.literal == FALSE) && ($3.atributs.continguts.info_variable.literal == FALSE)) {
				   sprintf(caracter, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
			   } else if ($1.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(caracter, "%s", $1.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string2, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else if ($3.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(string2, "%s", $3.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_char = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(caracter, "%s", sym_tmp_char->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_char);
				   sym_tmp_char->atributs.tipus = VARIABLE;
				   sym_tmp_char->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
				   sym_tmp_char->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_char->atributs.continguts.info_variable.constant = FALSE;
				   
				   sprintf(arg2, "'%c'", $1.atributs.continguts.info_variable.valor.character);
				   emite(codi, ASSIGN, sym_tmp_char->nom, arg2, NULL);
				   seguent_quadruple++;

			   } else {
				   sinError("No es pot donar aquest cas.");
			   }
			   
			   // Generem codi per la concatenacio
			   
			   // Variables temporals que necessitarem
			   Simbol *sym_result = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   sprintf($$.nom, "%s", sym_result->nom);  // Copiar el nom de la variable a on es guardara el resultat, al valor de retorn
			   encuar(cua_actual_var_tmp, sym_result);
			   sym_result->atributs.tipus = VARIABLE;
			   sym_result->atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   sym_result->atributs.continguts.info_variable.literal = FALSE;
			   sym_result->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_contador = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_contador);
			   sym_contador->atributs.tipus = VARIABLE;
			   sym_contador->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_contador->atributs.continguts.info_variable.literal = FALSE;
			   sym_contador->atributs.continguts.info_variable.constant = FALSE;
			   
			   
			   Simbol *sym_length02 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_length02);
			   sym_length02->atributs.tipus = VARIABLE;
			   sym_length02->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_length02->atributs.continguts.info_variable.literal = FALSE;
			   sym_length02->atributs.continguts.info_variable.constant = FALSE;
			   
			   Simbol *sym_char_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_char_tmp);
			   sym_char_tmp->atributs.tipus = VARIABLE;
			   sym_char_tmp->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_char_tmp->atributs.continguts.info_variable.literal = FALSE;
			   sym_char_tmp->atributs.continguts.info_variable.constant = FALSE;
			   
			   // length02 := string2[0];
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_length02->nom, string2, arg3);
			   seguent_quadruple++;
			   
			   //// Afegim caracter
			   sprintf(arg2, "1");
			   // resultat[contador] := caracter
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, arg2, caracter);
			   seguent_quadruple++;
			   
			   //// Comencem a copiar l'string 1
			   // contador := 2;
			   sprintf(arg2, "2");
			   emite(codi, ASSIGN, sym_contador->nom, arg2, NULL);
			   seguent_quadruple++;
			   
			   // if contador = length02 then goto [next]
			   sprintf(arg3, "%i", seguent_quadruple + 5);
			   emite(codi, IF_GTI, sym_contador->nom, sym_length02->nom, arg3);
			   seguent_quadruple++;
			   
			   // temporal := string2[contador];
			   emite(codi, ASSIGN_FROM_TABLE, sym_char_tmp->nom, string2, sym_contador->nom);
			   seguent_quadruple++;
			   
			   // resultat[contador] := temporal
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, sym_contador->nom, sym_char_tmp->nom);
			   seguent_quadruple++;
			   
			   // contador++
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_contador->nom, sym_contador->nom, arg3);
			   seguent_quadruple++;
			   
			   // goto [principi]
			   sprintf(arg1, "%i", seguent_quadruple - 4);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
			   
			   //// Tamany
			   // length01 := length01 + 1;
			   sprintf(arg3, "1");
			   emite(codi, ADDI, sym_length02->nom, sym_length02->nom, arg3);
			   seguent_quadruple++;
			   
			   // result[0] := length01;
			   sprintf(arg2, "0");
			   emite(codi, ASSIGN_TO_TABLE, sym_result->nom, arg2, sym_length02->nom);
			   seguent_quadruple++;
			   
		   } else {
			   sinError("No es possible concatenar aquests tipus.");
		   }
	   }
   }
;

term:
   // No es necessari un term_element i desglossar fins els operadors
   factor { 
	    sinLog(" -> term -> factor");
		memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   | factor '*' term {
	   sinLog(" -> term -> factor '*' term");
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   // Segons els tipus dels elements, el resultat sera integer o float
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			   // ja es tornara a comprovar amb el seu rang especific al assignment
			   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			   long long int resultat = (long long int)$1.atributs.continguts.info_variable.valor.integer * $3.atributs.continguts.info_variable.valor.integer;
			   if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
				   	|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("La multiplicacio sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
			   }
			   $$.atributs.continguts.info_variable.valor.integer = (int)resultat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat * $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   printf("Avis: Multiplicacio de tipus diferents (integer / float).");
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.integer * $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   printf("Avis: Multiplicacio de tipus diferents (float / integer).");
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = $1.atributs.continguts.info_variable.valor.vfloat * $3.atributs.continguts.info_variable.valor.integer;
		   } else {
			   sinError("L'operador '*' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
	   } else {
		   // Codi 3 adreces
		   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   memcpy(&(sym_tmp->atributs), &($1.atributs), sizeof(Atributs));
		   encuar(cua_actual_var_tmp, sym_tmp);
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MULI, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MULI, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, MULI, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, MULF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, MULF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, MULF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MULF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, MULF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, MULF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, MULF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MULF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, MULF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("L'operador '*' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   sym_tmp->atributs.tipus = VARIABLE;
		   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
		   memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	   }
   
   }
   | factor '/' term {
	   sinLog(" -> term -> factor '/' term");
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   // Controlem la divisio entre 0
		   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)
		       || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)){
			   if ($3.atributs.continguts.info_variable.valor.integer == 0) {
				   sinError("Divisio entre 0");
			   }
		   } else if  ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   if ($3.atributs.continguts.info_variable.valor.vfloat == 0) {
				   sinError("Divisio entre 0");
			   }
		   }
		   
		   double double_tmp;
		   double divisio;
		   
		   // NOTE: No controlem limits de la divisio
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   	&& (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		   		divisio = $1.atributs.continguts.info_variable.valor.integer / $3.atributs.continguts.info_variable.valor.integer;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   divisio = $1.atributs.continguts.info_variable.valor.vfloat / $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   printf("Avis: Divisio de tipus diferents (integer / float).");
			   divisio = $1.atributs.continguts.info_variable.valor.integer / $3.atributs.continguts.info_variable.valor.vfloat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   printf("Avis: Divisio de tipus diferents (float / integer).");
			   divisio = $1.atributs.continguts.info_variable.valor.vfloat / $3.atributs.continguts.info_variable.valor.integer;
		   } else {
			   sinError("L'operador '/' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   // Segons el resultat de la divisio, el tipus sera integer o float
		   if (modf(divisio, &double_tmp) > 0){
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = divisio;
		   } else {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.integer = divisio;
		   }
	   } else {
		   // Codi 3 adreces
		   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   memcpy(&(sym_tmp->atributs), &($1.atributs), sizeof(Atributs));
		   encuar(cua_actual_var_tmp, sym_tmp);
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: La operacio divisio entre TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT dona com a resultat sempre un TIPUS_FLOAT
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, DIVI, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   if ($3.atributs.continguts.info_variable.valor.integer == 0) {
					   sinError("Divisio entre 0");
				   }
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, DIVI, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, DIVI, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
			   
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, DIVF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   if ($3.atributs.continguts.info_variable.valor.vfloat == 0) {
					   sinError("Divisio entre 0");
				   }
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, DIVF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   if ($3.atributs.continguts.info_variable.valor.vfloat == 0) {
					   sinError("Divisio entre 0");
				   }
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, DIVF, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   if ($3.atributs.continguts.info_variable.valor.integer == 0) {
					   sinError("Divisio entre 0");
				   }
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, DIVF, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("L'operador '/' no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   sym_tmp->atributs.tipus = VARIABLE;
		   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
		   memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	   }
   }
   | factor PR_MOD term {
	   	sinLog(" -> term -> factor PR_MOD term");
	    if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
			// Controlem la divisio entre 0
			if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)
		       || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)){
			   if ($3.atributs.continguts.info_variable.valor.integer == 0) {
				   sinError("Divisio entre 0");
			   }
			} else if  ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   if ($3.atributs.continguts.info_variable.valor.vfloat == 0) {
				   sinError("Divisio entre 0");
			   }
		   	}
		   
			if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
				$$.atributs.tipus = VARIABLE;
				$$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
				$$.atributs.continguts.info_variable.literal = TRUE;
				// Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			    // ja es tornara a comprovar amb el seu rang especific al assignment
			    Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			    long long int resultat = (long long int)$1.atributs.continguts.info_variable.valor.integer % $3.atributs.continguts.info_variable.valor.integer;
			    if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
					|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("El modul sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
			    }
				$$.atributs.continguts.info_variable.valor.integer = (int)resultat;
			} else {
				sinError("L'operador MOD nomes pot ser aplicat a tipus 'integer' mod 'integer'");
			}
		} else {
		   // Codi 3 adreces
		   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   memcpy(&(sym_tmp->atributs), &($1.atributs), sizeof(Atributs));
		   encuar(cua_actual_var_tmp, sym_tmp);
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: La operacio divisio entre TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT dona com a resultat sempre un TIPUS_INTEGER
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MODI, sym_tmp->nom, arg2, $3.nom);
				   seguent_quadruple++;
			   } else if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   if ($3.atributs.continguts.info_variable.valor.integer == 0) {
					   sinError("Divisio entre 0");
				   }
				   char arg3[STRING_LENGTH];
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, MODI, sym_tmp->nom, $1.nom, arg3);
				   seguent_quadruple++;
			   } else {
				   // Els 2 no son literals
				   emite(codi, MODI, sym_tmp->nom, $1.nom, $3.nom);
				   seguent_quadruple++;
			   }
			   
		   } else {
			   sinError("L'operador MOD nomes pot ser aplicat a tipus 'integer' mod 'integer'");
		   }
		   
		   sym_tmp->atributs.tipus = VARIABLE;
		   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
		   memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	   }
   }
;

factor:
   primary { 
	   sinLog(" -> factor -> primary");
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   | primary PR_EXP primary {
	   sinLog(" -> factor -> primary PR_EXP primary");
	   
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   // Segons els tipus dels elements, el resultat sera integer o float
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   // Primer comprovar limit 'integer', si es vol assignar a un integer_derivat
			   // ja es tornara a comprovar amb el seu rang especific al assignment
			   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
			   long long int resultat = (long long int)pow($1.atributs.continguts.info_variable.valor.integer, $3.atributs.continguts.info_variable.valor.integer);
			   if (( resultat > (int)symTipus->atributs.continguts.info_type.rang.max )
					|| ( resultat < (int)symTipus->atributs.continguts.info_type.rang.min )){
				   sinError("La exponenciacio sobrepassa el limit dels 'integers', el resultat ha d'estar entre -2147483647 i 2147483647.");
			   }
			   $$.atributs.continguts.info_variable.valor.integer = (int)resultat;
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = pow($1.atributs.continguts.info_variable.valor.vfloat, $3.atributs.continguts.info_variable.valor.vfloat);
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = pow($1.atributs.continguts.info_variable.valor.integer, $3.atributs.continguts.info_variable.valor.vfloat);
		   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = pow($1.atributs.continguts.info_variable.valor.vfloat, $3.atributs.continguts.info_variable.valor.integer);
		   } else {
			   sinError("L'operador EXP no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   
		   char num[STRING_LENGTH];
		   char exp[STRING_LENGTH];
		   
		   Tipus_Operacio multiplicacio;
		   Tipus_Operacio divisio;
		   
		   if ( (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
			   && (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) ) {
		       // NOTE: Les operacions amb TIPUS_INTEGER i TIPUS_INTEGER_DERIVAT donen com a resultat sempre un TIPUS_INTEGER
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   
			   
			   if ($1.atributs.continguts.info_variable.literal == TRUE) {
				   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(num, "%s", sym_tmp->nom);
				   encuar(cua_actual_var_tmp, sym_tmp);
				   sym_tmp->atributs.tipus = VARIABLE;
				   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
				   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
				   
				   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ASSIGN, sym_tmp->nom, arg2, NULL);
				   seguent_quadruple++;
			   } else {
				   sprintf(num, "%s", $1.nom);
			   }
			   
			   if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(exp, "%s", sym_tmp->nom);
				   encuar(cua_actual_var_tmp, sym_tmp);
				   sym_tmp->atributs.tipus = VARIABLE;
				   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
				   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
				   
				   if ($3.atributs.continguts.info_variable.valor.integer < 0) {
					   sinError("No es troba suportat exponents negatius per aquest cas.");
				   }
				   
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ASSIGN, sym_tmp->nom, arg2, NULL);
				   seguent_quadruple++;
			   } else {
				   printf("Avis: Si l'exponent es negatius a %s de la linea %i columna %i, el resultat sera imprevisible.\n", $3.nom, lineNumber, colNumber);
				   sprintf(exp, "%s", $3.nom);
			   }
			   
		   } else {
			   // NOTE: Codi 3 adreces nomes soportem EXP si els 2 numeros son integers
			   sinError("L'operador EXP no pot ser aplicat a tipus de dades diferent de 'integer'");
		   }
		   
		   // Generem codi per la exponenciacio
			   
		   // Variables temporals que necessitarem
		   Simbol *sym_result = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   sprintf($$.nom, "%s", sym_result->nom);  // Copiar el nom de la variable a on es guardara el resultat, al valor de retorn
		   encuar(cua_actual_var_tmp, sym_result);
		   sym_result->atributs.tipus = VARIABLE;
		   sym_result->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
		   sym_result->atributs.continguts.info_variable.literal = FALSE;
		   sym_result->atributs.continguts.info_variable.constant = FALSE;
		   
		   Simbol *sym_contador = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   encuar(cua_actual_var_tmp, sym_contador);
		   sym_contador->atributs.tipus = VARIABLE;
		   sym_contador->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
		   sym_contador->atributs.continguts.info_variable.literal = FALSE;
		   sym_contador->atributs.continguts.info_variable.constant = FALSE;
		   
		   // contador := 1
		   sprintf(arg2, "1");
		   emite(codi, ASSIGN, sym_contador->nom, arg2, NULL);
		   seguent_quadruple++;
		   
		   // result := num;
		   emite(codi, ASSIGN, sym_result->nom, num, NULL);
		   seguent_quadruple++;
		   
		   // if contador <= exp then goto [fer]
		   sprintf(arg3, "%i", seguent_quadruple + 2);
		   emite(codi, IF_LEI, sym_contador->nom, exp, arg3);
		   seguent_quadruple++;
		   
		   // goto [end]
		   sprintf(arg1, "%i", seguent_quadruple + 4);
		   emite(codi, GOTO, arg1, NULL, NULL);
		   seguent_quadruple++;
		   
		   // result := result * 10
		   sprintf(arg3, "10");
		   emite(codi, MULI, sym_result->nom, sym_result->nom, arg3);
		   seguent_quadruple++;
		   
		   // count++
		   sprintf(arg3, "1");
		   emite(codi, ADDI, sym_contador->nom, sym_contador->nom, arg3);
		   seguent_quadruple++;
		   
		   // goto [begin]
		   sprintf(arg1, "%i", seguent_quadruple - 4);
		   emite(codi, GOTO, arg1, NULL, NULL);
		   seguent_quadruple++;
	   }
	   
   }
   | PR_ABS primary {
	   sinLog(" -> factor -> PR_ABS primary");
	   
	   if ($2.atributs.continguts.info_variable.literal) {
		   // Segons els tipus dels elements, el resultat sera integer o float
		   if (($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)
			   || ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)){
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.integer = abs($2.atributs.continguts.info_variable.valor.integer);
		   } else if ($2.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   $$.atributs.continguts.info_variable.valor.vfloat = abs($2.atributs.continguts.info_variable.valor.vfloat);
		   } else {
			   sinError("L'operador ABS no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   
		   Tipus_Operacio operador_IF_GT;
		   Tipus_Operacio operador_CHS;
		   if (($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)
			   		|| ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)){
			   $$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   operador_IF_GT = IF_GEI;
			   operador_CHS = CHSI;
		   } else if ($2.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
			   operador_IF_GT = IF_GEF;
			   operador_CHS = CHSF;
		   } else {
			   sinError("L'operador ABS no pot ser aplicat a tipus de dades diferent de 'integer' o 'float'");
		   }
		   
		   // Generem codi per la exponenciacio
			   
		   // Variables temporals que necessitarem
		   Simbol *sym_result = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
		   sprintf($$.nom, "%s", sym_result->nom);  // Copiar el nom de la variable a on es guardara el resultat, al valor de retorn
		   encuar(cua_actual_var_tmp, sym_result);
		   sym_result->atributs.tipus = VARIABLE;
		   sym_result->atributs.continguts.info_variable.tipus = TIPUS_STRING;
		   sym_result->atributs.continguts.info_variable.literal = FALSE;
		   sym_result->atributs.continguts.info_variable.constant = FALSE;
		   
		   // if contador >= exp then goto [igual]
		   sprintf(arg2, "0");
		   sprintf(arg3, "%i", seguent_quadruple + 3);
		   emite(codi, operador_IF_GT, $2.nom, arg2, arg3);
		   seguent_quadruple++;
		   
		   // result := - (numero)
		   emite(codi, operador_CHS, sym_result->nom, $2.nom, NULL);
		   seguent_quadruple++;
		   
		   // goto [end]
		   sprintf(arg1, "%i", seguent_quadruple + 2);
		   emite(codi, GOTO, arg1, NULL, NULL);
		   seguent_quadruple++;
		   
		   // result := numero
		   emite(codi, ASSIGN, sym_result->nom, $2.nom, NULL);
		   seguent_quadruple++;
	   }
   }
   | PR_NOT primary {
	   sinLog(" -> factor -> PR_NOT primary");
	   
	   if ($2.atributs.continguts.info_variable.literal) {
		   // Nomes aplicable a booleans
		   if ($2.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($2.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
			   }
		   } else {
			   sinError("L'operador ABS no pot ser aplicat a tipus de dades diferent de 'boolean'");
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   sinLog("\t** No son literals.");
	   }
	   
	   // Back Patching
	   $$.llista_cert = $2.llista_fals;
	   $$.llista_fals = $2.llista_cert;
   }
;

primary:
   PR_DECIMAL_LITERAL { 
	    sinLog(" -> primary -> DECIMAL LITERAL");
		$$.atributs.tipus = VARIABLE;
		$$.atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
		$$.atributs.continguts.info_variable.valor.vfloat = $1;
		sprintf($$.nom, "");	// No esta a la TS
		$$.atributs.continguts.info_variable.literal = TRUE;
		$$.atributs.continguts.info_variable.inicialitzat = TRUE;
   }
   | PR_NUMERIC_LITERAL { 
	    sinLog(" -> primary -> NUMERIC LITERAL");
		$$.atributs.tipus = VARIABLE;
		$$.atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
		$$.atributs.continguts.info_variable.valor.integer = (int)$1;
		sprintf($$.nom, "");	// No esta a la TS
		$$.atributs.continguts.info_variable.literal = TRUE;
		$$.atributs.continguts.info_variable.inicialitzat = TRUE;
   }
   | PR_STRING_LITERAL { 
	   sinLog(" -> primary -> STRING LITERAL");
	   $$.atributs.tipus = VARIABLE;
	   $$.atributs.continguts.info_variable.tipus = TIPUS_STRING;
	   
	   if (strlen($1)-2 > 255) {  // - 1 (longitud) -1 (final string \0)
			sinError("String literal \"%s\" massa llarg. Maxim 255 caracters i te %d caracters", $1, strlen($1));
	   }
	   memcpy(&($$.atributs.continguts.info_variable.valor.string), $1, strlen($1));
	   $$.atributs.continguts.info_variable.literal = TRUE;
	   $$.atributs.continguts.info_variable.inicialitzat = TRUE;
	   sprintf($$.nom, "");	// No esta a la TS
   }
   | name {
	   sinLog(" -> primary -> name");
	   memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | '(' expression ')' {
	   sinLog(" -> primary -> '(' expression ')'");
	   memcpy(&$$, &$2,sizeof(struct Simbol));
   }
;

name:
   direct_name {
	   // Retorna VARIABLE
	    sinLog(" -> name -> direct_name");
		memcpy(&$$, &$1,sizeof(struct Simbol));
		
   }
   | indexed_component {
	   // Retorna VARIABLE
	   sinLog(" -> name -> indexed_component");
	   memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | selected_component {
	   // Retorna VARIABLE
	   sinLog(" -> name -> selected_component");
	   memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | type_conversion {
	   // Retorna VARIABLE
	   sinLog(" -> name -> type_conversion");
	   memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | function_call {
	   // Retorna VARIABLE
	   sinLog(" -> name -> function_call");
	   memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | PR_CHARACTER_LITERAL {
	    sinLog(" -> name -> PR_CHARACTER_LITERAL");
		$$.atributs.tipus = VARIABLE;
		$$.atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
		$$.atributs.continguts.info_variable.valor.character = $1;
		$$.atributs.continguts.info_variable.literal = TRUE;
		sprintf($$.nom, "");	// No esta a la TS
   }
;

function_call:
	// Aqui es produeix un "Expected shift/reduce conflict"
	PR_ID_FUNCTION {
		Boolean trobat = FALSE;
		Info_Function *info_function;
		
		info_function = &($1->atributs.continguts.info_function);
		if (info_function->num_parametres == 0) {
			trobat = TRUE;
		}
		
		if (trobat == FALSE) {
			if ($1->atributs.continguts.info_function.sobrecarregues == NULL) {
				sinError("La funcio '%s' requereix '%i' parametre(s).", $1->nom, $1->atributs.continguts.info_function.num_parametres);
			} else {
				// Recorrer sobrecarregues
				reiniciar_PDI($1->atributs.continguts.info_function.sobrecarregues);
				info_function = (Info_Function *)obtenir_element_PDI($1->atributs.continguts.info_function.sobrecarregues);
				while(info_function != NULL) {
					// Comprovar parametres
					if (info_function->num_parametres == 0) {
						trobat = TRUE;
						break;
					}
					
					seguent_PDI($1->atributs.continguts.info_function.sobrecarregues);
					info_function = (Info_Function *)obtenir_element_PDI($1->atributs.continguts.info_function.sobrecarregues);
				}
			}
		}
		
		if (trobat) {
			// NOTE: Es pot cridar una funcio amb cos no definit si la signatura si ha sigut definida
			// Al final del compilador ja es comprova que tots tenen cos definit.
			/*if (info_function->complet == FALSE) {
				sinError("La funcio '%s' no te encara el cos definit.", $1->nom);
			}*/
			
			$$.atributs.tipus = VARIABLE;
			$$.atributs.continguts.info_variable.tipus = info_function->tipus_retorn;
			$$.atributs.continguts.info_variable.literal = FALSE;
			$$.atributs.continguts.info_variable.inicialitzat = FALSE;
		
			if (($1->atributs.continguts.info_function.tipus_retorn == TIPUS_DERIVAT)
				|| ($1->atributs.continguts.info_function.tipus_retorn == TIPUS_INTEGER_DERIVAT)) {
				memcpy($$.atributs.continguts.info_variable.nom_tipus, info_function->nom_tipus_retorn, STRING_LENGTH);
			}
			
			// Codi 3 adreces
			char arg2[STRING_LENGTH];
			char arg3[STRING_LENGTH];
			
			// Temporal per guardar valor de retorn
			Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			memcpy(&(sym_tmp->atributs), &($$.atributs), sizeof(Atributs));
			encuar(cua_actual_var_tmp, sym_tmp);

			if (strcmp($1->nom, "get_integer") == 0) {
				sprintf(arg2, "0");
				emite(codi, GETI, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if (strcmp($1->nom, "get_float") == 0) {
				sprintf(arg2, "0");
				emite(codi, GETF, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if (strcmp($1->nom, "get_character") == 0) {
				sprintf(arg2, "0");
				emite(codi, GETC, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else {
				sprintf(arg2, "%s", info_function->nom_unic);
				sprintf(arg3, "0");
				emite(codi, CALL_FUNC, sym_tmp->nom, arg2, arg3);
				seguent_quadruple++;
			}
			
			sprintf($$.nom, "%s", sym_tmp->nom);  // El simbol que retornem ha de tindre el nom generat amb nou_temp()
		} else {
			sinError("La funcio '%s' no te cap sobrecarrega que permeti 0 parametres.", $1->nom);
		}
	}
	| PR_ID_FUNCTION function_call_element {
		
		Boolean trobat = FALSE;
		Info_Function *info_function;
		
		info_function = &($1->atributs.continguts.info_function);
		// Mateix nombre de parametres?

		if (info_function->num_parametres == contador) {
			// Comprovar tipus parametres
			if (crida_parametres_diferents(cua_tmp, info_function->parametres, info_function->taula) == FALSE) {
				// Coincideixen el tipus dels parametres amb la definicio
				trobat = TRUE;
			}
		}
		
		if (trobat == FALSE) {
			if ($1->atributs.continguts.info_function.sobrecarregues == NULL) {
				sinError("La funcio '%s' requereix '%i' parametre(s) i no s'han trobat cap tipus coincident.", $1->nom, $1->atributs.continguts.info_function.num_parametres);
			} else {
				// Recorrer sobrecarregues
				reiniciar_PDI($1->atributs.continguts.info_function.sobrecarregues);
				info_function = (Info_Function *)obtenir_element_PDI($1->atributs.continguts.info_function.sobrecarregues);
				while(info_function != NULL) {
					// Mateix nombre de parametres?
					if (info_function->num_parametres == contador) {
						// Comprovar tipus parametres
						if (crida_parametres_diferents(cua_tmp, info_function->parametres, info_function->taula) == FALSE) {
							// Coincideixen el tipus dels parametres amb la definicio
							trobat = TRUE;
							break;
						}
					}
					
					seguent_PDI($1->atributs.continguts.info_function.sobrecarregues);
					info_function = (Info_Function *)obtenir_element_PDI($1->atributs.continguts.info_function.sobrecarregues);
				}
			}
		}
		
		if (trobat) {
			// NOTE: Es pot cridar una funcio amb cos no definit si la signatura si ha sigut definida
			// Al final del compilador ja es comprova que tots tenen cos definit.
			/*if (info_function->complet == FALSE) {
				sinError("La funcio '%s' no te encara el cos definit.", $1->nom);
			}*/
			
			$$.atributs.tipus = VARIABLE;
			$$.atributs.continguts.info_variable.tipus = info_function->tipus_retorn;
			$$.atributs.continguts.info_variable.literal = FALSE;
			$$.atributs.continguts.info_variable.inicialitzat = FALSE;
		
			if (($1->atributs.continguts.info_function.tipus_retorn == TIPUS_DERIVAT)
				|| ($1->atributs.continguts.info_function.tipus_retorn == TIPUS_INTEGER_DERIVAT)) {
				memcpy($$.atributs.continguts.info_variable.nom_tipus, info_function->nom_tipus_retorn, STRING_LENGTH);
			}
			
			char arg1[STRING_LENGTH];
			char arg2[STRING_LENGTH];
			char arg3[STRING_LENGTH];
			
			// PARAMS
			reiniciar_PDI(cua_tmp);
			Simbol *sym_param = (Simbol *)obtenir_element_PDI(cua_tmp);
			Simbol *sym_param_especial = sym_param; // Usat per la funcio especial LENGTH
			while (sym_param!=NULL) {
				if (sym_param->atributs.continguts.info_variable.literal == TRUE) {
					if ((sym_param->atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || (sym_param->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg1, "%i", sym_param->atributs.continguts.info_variable.valor.integer);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg1, "%g", sym_param->atributs.continguts.info_variable.valor.vfloat);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg1, "'%c'", sym_param->atributs.continguts.info_variable.valor.character);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg1, "%i", sym_param->atributs.continguts.info_variable.valor.boolean);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Generem un temporal per guardar l'string i fem el PARAM amb aquest (excepte funcio especial LENGTH)
					   if (strcmp($1->nom, "length") != 0) {
						   int caracters = strlen(sym_param->atributs.continguts.info_variable.valor.string);
						   int i;
						   
						   char arg2[STRING_LENGTH];
						   char arg3[STRING_LENGTH];
					   
						   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
						   encuar(cua_actual_var_tmp, sym_tmp);
						   sym_tmp->atributs.tipus = VARIABLE;
						   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_STRING;
						   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
						   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
						   
						   // A la primera posicio va el tamany
						   sprintf(arg2, "0");
						   sprintf(arg3, "%i", caracters - 1);
						   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
						   seguent_quadruple++;
						   for (i = 1; i < caracters; i++) {
							   sprintf(arg2, "%i", i);
							   sprintf(arg3, "'%c'", sym_param->atributs.continguts.info_variable.valor.string[i]);
							   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
							   seguent_quadruple++;
						   }
						   
						   sprintf(arg1, "%s", sym_tmp->nom);
					   }
				   }
				} else {
					sprintf(arg1, "%s", sym_param->nom);
				}
				
				if (strcmp($1->nom, "length") != 0) {	// No per la funcio especial LENGTH
					emite(codi, PARAM, arg1, NULL, NULL);
					seguent_quadruple++;
				}
				
				seguent_PDI(cua_tmp);
				sym_param_especial = sym_param;
				sym_param = (Simbol *)obtenir_element_PDI(cua_tmp);
				
			}
			
			// Temporal per guardar valor de retorn
			Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			memcpy(&(sym_tmp->atributs), &($$.atributs), sizeof(Atributs));
			encuar(cua_actual_var_tmp, sym_tmp);
			
			if (strcmp($1->nom, "length") == 0) {
				
				if (sym_param_especial->atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "%i", (int)sym_param_especial->atributs.continguts.info_variable.valor.string[0]);
					emite(codi, ASSIGN, sym_tmp->nom, arg2, NULL);
					seguent_quadruple++;
				} else {
					sprintf(arg3, "0");
					emite(codi, ASSIGN_FROM_TABLE, sym_tmp->nom, sym_param_especial->nom, arg3);
					seguent_quadruple++;
				}
			} else {
				sprintf(arg2, "%s", info_function->nom_unic);
				sprintf(arg3, "%i", info_function->num_parametres);
				emite(codi, CALL_FUNC, sym_tmp->nom, arg2, arg3);
				seguent_quadruple++;
			}
			
			sprintf($$.nom, "%s", sym_tmp->nom);  // El simbol que retornem ha de tindre el nom generat amb nou_temp()
			cua_tmp = NULL;
		} else {
			sinError("La funcio '%s' no te cap sobrecarrega que permeti '%i' parametre(s) amb aquests tipus.", $1->nom, contador);
		}
	}
;

function_call_element:
	'(' {
		cua_tmp = (Cua *)crear_cua();
		contador = 0;
	}
	actual_parameter_part_element ')' {}
;

actual_parameter_part_element: // Gramatica auxiliar de 'actual_parameter_part'
	parameter_association {}
   | actual_parameter_part_element ',' parameter_association  {}
;


parameter_association:
   explicit_actual_parameter {}
;

explicit_actual_parameter:
   expression {
	   sinLog(" -> explicit_actual_parameter -> expression");
	   
	   // Cua d'arguments de la crida d'una funcio o procediment
	   Simbol *sym = (Simbol *)malloc(sizeof(struct Simbol));
	   memcpy(sym, &$1, sizeof(struct Simbol));
	   encuar(cua_tmp, sym);
	   contador++; // Nombre de parametres a la crida a la funcio/procediment
   }
   //| PR_ID_VARIABLE {}  //esta incluida en la anterior ?? Segun la practica vieja si.
;
	

type_conversion:
	subtype_mark '(' name ')' {
		sinLog(" -> type_conversion -> subtype_mark '(' name ')'");
		Simbol *sym_tmp;
		if (($1.atributs.tipus == TYPE) && ($3.atributs.tipus == VARIABLE)) {
			sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			encuar(cua_actual_var_tmp, sym_tmp);
			
			sym_tmp->atributs.tipus = VARIABLE;
			sym_tmp->atributs.continguts.info_variable.literal = $3.atributs.continguts.info_variable.literal;
			if (($1.atributs.continguts.info_type.tipus == TIPUS_CHARACTER) 
				&& (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))) {
				// INTEGER||INTEGER_DERIVAT -> CHARACTER				
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
				sym_tmp->atributs.continguts.info_variable.valor.character = (char)$3.atributs.continguts.info_variable.valor.integer;
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				} else {
					sprintf(arg2, "%s", $3.nom);
				}
				emite(codi, I2C, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if ((($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT))
				&& ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER)) {
				// CHARACTER -> INTEGER||INTEGER_DERIVAT
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
				sym_tmp->atributs.continguts.info_variable.valor.integer = $3.atributs.continguts.info_variable.valor.character;
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				} else {
					sprintf(arg2, "%s", $3.nom);
				}
				emite(codi, C2I, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if ((($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT)) 
				&& ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT)) {
				// FLOAT -> INTEGER||INTEGER_DERIVAT
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
				sym_tmp->atributs.continguts.info_variable.valor.integer = (int)$3.atributs.continguts.info_variable.valor.vfloat;
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				} else {
					sprintf(arg2, "%s", $3.nom);
				}
				emite(codi, F2I, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if (($1.atributs.continguts.info_type.tipus == TIPUS_FLOAT) 
				&& (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER))) {
				// INTEGER||INTEGER_DERIVAT -> FLOAT
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_FLOAT;
				sym_tmp->atributs.continguts.info_variable.valor.vfloat = $3.atributs.continguts.info_variable.valor.integer;
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				} else {
					sprintf(arg2, "%s", $3.nom);
				}
				emite(codi, I2F, sym_tmp->nom, arg2, NULL);
				seguent_quadruple++;
			} else if (($1.atributs.continguts.info_type.tipus == TIPUS_STRING) && ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER)) {
				// CHARACTER -> STRING
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				sprintf(sym_tmp->atributs.continguts.info_variable.valor.string, "%c", $3.atributs.continguts.info_variable.valor.character);
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				char arg3[STRING_LENGTH];
				// Assignem tamany
				sprintf(arg2, "0");
				sprintf(arg3, "1");	// tamany
				emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
				seguent_quadruple++;
				
				// Assignem valor
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				} else {
					sprintf(arg3, "%s", $3.nom);
				}
				sprintf(arg2, "1");
				emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
				seguent_quadruple++;
			} else if (($1.atributs.continguts.info_type.tipus == TIPUS_CHARACTER) && ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING)) {
				// STRING -> CHARACTER
				printf("La conversio de la linea %i columa %i, pot provocar perdues d'informacio.\n", lineNumber, colNumber);
				
			    sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
				//sprintf(sym_tmp->atributs.continguts.info_variable.valor.string, "%c", $3.atributs.continguts.info_variable.valor.character);
				
				// Codi 3 adreces
				char arg2[STRING_LENGTH];
				char arg3[STRING_LENGTH];
				
				// Assignem valor (nomes agafem el caracter de la primera posicio)
				if ($3.atributs.continguts.info_variable.literal == TRUE) {
					sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.string[1]);
				} else {
					sprintf(arg2, "%s", $3.nom);
				}
				sprintf(arg3, "1");
				emite(codi, ASSIGN_FROM_TABLE, sym_tmp->nom, arg2, arg3);
				seguent_quadruple++;
				
			} else {
				sinError("La conversio no es pot realitzar");
			}
		} else {
			sinError("La conversio no es pot realitzar");
		}
		
		memcpy(&$$, sym_tmp, sizeof(struct Simbol));
	}
;


selected_component:
   prefix '.' selector_name {
	   sinLog(" -> selected_component -> name '.' PR_ID");
	   if ($1.atributs.tipus != VARIABLE) {
		   sinError("Prefix ha de ser una variable");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != TIPUS_DERIVAT) {
		   sinError("Prefix ha de ser tipus RECORD.");
	   }
	   
	   // Busquem el simbol del tipus del "prefix"
	   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1.atributs.continguts.info_variable.nom_tipus);
	   if ((symTipus == NULL) || ((symTipus->atributs.tipus != RECORD_TYPE) && (symTipus->atributs.tipus != ARRAY_TYPE))) {
		   sinError("No es possible tractar com un RECORD '%s' donat que no es d'aquest tipus.", $1.atributs.continguts.info_variable.nom_tipus);
	   }
	   
	   if (symTipus->atributs.tipus == ARRAY_TYPE) {
		   Simbol *symTipusDelArray;
		   
		   if (symTipus->atributs.continguts.info_array.tipus == TIPUS_DERIVAT) {
			   symTipusDelArray = busca_totes_taules(taulaSimbolActual, symTipus->atributs.continguts.info_array.nom_tipus);
			   
			   if (symTipusDelArray->atributs.tipus != RECORD_TYPE) {
				   sinError("L'array '%s' no es de RECORDs.", $1.atributs.continguts.info_variable.nom_tipus);
			   }
		   } else {
			   sinError("L'array '%s' no es de RECORDs.", $1.atributs.continguts.info_variable.nom_tipus);
		   }
		   
		   // Busquem entre els elements del RECORD per veure si existeix selector_name
		   Simbol *symSubvariable = obtenir_simbol(symTipusDelArray->atributs.continguts.info_record.elements, $3);
		   if (symSubvariable == NULL) {
			   sinError("%s no es un membre del RECORD tipus '%s'", $3, symTipusDelArray->nom);
		   }
		   
		   Simbol *symSubtipus;
		   //Taula_Simbols *ts = symTipusDelArray->atributs.continguts.info_record.elements;
		   if ((symSubvariable->atributs.continguts.info_variable.tipus == TIPUS_DERIVAT)
				   || (symSubvariable->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   symSubtipus = busca_totes_taules(taulaSimbolActual, symSubvariable->atributs.continguts.info_variable.nom_tipus);
		   } else {
			   switch(symSubvariable->atributs.continguts.info_array.tipus) {
				   case TIPUS_INTEGER:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "integer");
					   break;
				   case TIPUS_FLOAT:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "float");
					   break;
				   case TIPUS_CHARACTER:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "character");
					   break;
				   case TIPUS_STRING:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "string");
					   break;
				   case TIPUS_BOOLEAN:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "boolean");
					   break;
				   default:
					   sinError("??");
					   break;
			   }
		   }
		    
		   
		   // Codi 3 adreces
		   // Bytes que ocupa cada posicio del array
		   int bytes = calcular_tamany_tipus(taulaSimbolActual, symTipus->atributs.continguts.info_array.nom_tipus);
		   
		   // Pels arrays que no comencen per 0:
		   int rang_min = (int)symTipus->atributs.continguts.info_array.rang.min;

		   char offset[STRING_LENGTH];
		   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
			   int offset_array = $1.atributs.continguts.info_variable.offset;
			   int offset_record = obtenir_offset_fins_element_record(symTipusDelArray, $3, taulaSimbolActual);
			   
			   // Retornem nom del array amb l'offset literal adient
			   $$.atributs.tipus = VARIABLE;
			   sprintf($$.nom, "%s", $1.nom);
			   $$.atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   $$.atributs.continguts.info_variable.constant = FALSE;
			   sprintf($$.atributs.continguts.info_variable.nom_tipus, "%s", symTipus->nom);
			   $$.atributs.continguts.info_variable.offset_literal = TRUE;
			   $$.atributs.continguts.info_variable.offset = offset_array + offset_record;
			   $$.atributs.continguts.info_variable.offset_simbol = symSubtipus;
		   } else {
			   sprintf(offset, "%s", $1.atributs.continguts.info_variable.nom_offset);
			   
			   int offset_record = obtenir_offset_fins_element_record(symTipusDelArray, $3, taulaSimbolActual);
			   
			   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   sym_tmp->atributs.tipus = VARIABLE;
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
			   encuar(cua_actual_var_tmp, sym_tmp);
			   
			   char arg3[STRING_LENGTH];
			   sprintf(arg3, "%i", offset_record);
			   
			   // offset es calcula a una variable temporal
			   emite(codi, ADDI, sym_tmp->nom, offset, arg3);
			   seguent_quadruple++;
			   
			   // Retornem nom del array amb l'offset (NO literal) adient
			   $$.atributs.tipus = VARIABLE;
			   sprintf($$.nom, "%s", $1.nom);
			   $$.atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
			   $$.atributs.continguts.info_variable.literal = FALSE;
			   $$.atributs.continguts.info_variable.constant = FALSE;
			   sprintf($$.atributs.continguts.info_variable.nom_tipus, "%s", symTipus->nom);
			   $$.atributs.continguts.info_variable.offset_literal = FALSE;
			   sprintf($$.atributs.continguts.info_variable.nom_offset, "%s", sym_tmp->nom);
			   $$.atributs.continguts.info_variable.offset_simbol = symSubtipus;
		   }
		   
	   } else {
	   
		   // Busquem entre els elements del RECORD per veure si existeix selector_name
		   Simbol *symSubvariable = obtenir_simbol(symTipus->atributs.continguts.info_record.elements, $3);
		   if (symSubvariable == NULL) {
			   sinError("%s no es un membre del RECORD tipus '%s'", $3, $1.atributs.continguts.info_variable.nom_tipus);
		   }
		   
		   Simbol *symSubtipus;
		   if ((symSubvariable->atributs.continguts.info_variable.tipus == TIPUS_DERIVAT)
				   || (symSubvariable->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   symSubtipus = busca_totes_taules(taulaSimbolActual, symSubvariable->atributs.continguts.info_variable.nom_tipus);
		   } else {
			   switch(symSubvariable->atributs.continguts.info_array.tipus) {
				   case TIPUS_INTEGER:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "integer");
					   break;
				   case TIPUS_FLOAT:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "float");
					   break;
				   case TIPUS_CHARACTER:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "character");
					   break;
				   case TIPUS_STRING:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "string");
					   break;
				   case TIPUS_BOOLEAN:
					   symSubtipus = busca_totes_taules(taulaSimbolActual, "boolean");
					   break;
				   default:
					   sinError("??");
					   break;
			   }
		   }
		   
		   // Retornem nom del RECORD amb l'offset (literal) adient
		   $$.atributs.tipus = VARIABLE;
		   sprintf($$.nom, "%s", $1.nom);
		   $$.atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   $$.atributs.continguts.info_variable.constant = FALSE;
		   //sprintf($$.atributs.continguts.info_variable.nom_tipus, "%s", symSubtipus->nom);
		   sprintf($$.atributs.continguts.info_variable.nom_tipus, "%s", $1.atributs.continguts.info_variable.nom_tipus);
		   $$.atributs.continguts.info_variable.offset_literal = TRUE;
		   $$.atributs.continguts.info_variable.offset = obtenir_offset_fins_element_record(symTipus, $3, taulaSimbolActual);
		   $$.atributs.continguts.info_variable.offset_simbol = symSubtipus;
	   }
   }
;

selector_name:
   PR_ID {
	   sinLog(" -> selector_name -> PR_ID");
	   snprintf($$, STRING_LENGTH, "%s" , $1);
   }
;

direct_name:
   PR_ID_VARIABLE {
	    sinLog(" -> direct_name -> PR_ID_VARIABLE");
		memcpy(&$$, $1,sizeof(struct Simbol));
		$$.atributs.continguts.info_variable.literal = FALSE;
		
		if (((strcmp($1->nom, "true") == 0) || (strcmp($1->nom, "false") == 0))
			&& ($1->atributs.continguts.info_variable.inicialitzat == TRUE)) {
			
			// Ha ficat la paraula reservada "true" o "false", que es un valor literal
			$$.atributs.continguts.info_variable.literal = TRUE;
			
			if ($1->atributs.continguts.info_variable.valor.boolean == TRUE) {
				
				// Codi 3 adreces
				// E.listaverdad = crealista(sigtecuad);
				$$.llista_cert = (Cua*)crear_cua();
				int *pos = (int*)malloc(sizeof(int));
				*pos = seguent_quadruple;
				encuar($$.llista_cert, pos);
				
			} else {
				// Codi 3 adreces
				// E.listafalso = crealista(sigtecuad);
				$$.llista_fals = (Cua*)crear_cua();
				int *pos = (int*)malloc(sizeof(int));
				*pos = seguent_quadruple;
				encuar($$.llista_fals, pos);
			}
			
			// Quan es fica un literal "true", les sentencies generades es guarden al seu simbol
		    // per poder diferenciar d'una condicio a una inicialitzacio/crida
			$$.sentencies = (Cua*)crear_cua();
			emite($$.sentencies, GOTO, NULL, NULL, NULL);
			//seguent_quadruple++;
			// Es sumara una linea al fusionar $$.sentencies amb el codi principal
		}
		
   }
;

prefix:
   name {
	    sinLog(" -> prefix -> name");
		memcpy(&$$, &$1,sizeof(struct Simbol));
   }
;

indexed_component:
   prefix '(' expression ')' {
	   
	   sinLog(" -> indexed_component -> prefix '(' expression ')'");
	   
	   // Es suposa que en prefix hi ha una VARIABLE
	   if ($1.atributs.tipus != VARIABLE) {
		   sinError("L'index ha de ser una variable.");
	   }
	   
	   // Es suposa que en prefix hi ha una VARIABLE del tipus ARRAY
	   if ($1.atributs.continguts.info_variable.tipus != TIPUS_DERIVAT) {
		   sinError("No es possible tractar amb un ARRAY '%s' donat que no es d'aquest tipus.", $1.atributs.continguts.info_variable.nom_tipus);
	   }
	   
	   // Busquem el simbol del tipus de la variable
	   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1.atributs.continguts.info_variable.nom_tipus);
	   if ((symTipus == NULL) || (symTipus->atributs.tipus != ARRAY_TYPE)) {
		   sinError("No es possible tractar amb un ARRAY '%s' donat que no es d'aquest tipus.", $1.atributs.continguts.info_variable.nom_tipus);
	   }
	   
	   if ($3.atributs.tipus != VARIABLE) {
		   sinError("L'index ha de ser una variable.");
	   }
	   
	   // Comprovem que hi esta dintre del rang
	   if ($3.atributs.continguts.info_variable.literal) {
		   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   if (($3.atributs.continguts.info_variable.valor.integer < (int)symTipus->atributs.continguts.info_array.rang.min)
				   || ($3.atributs.continguts.info_variable.valor.integer > (int)symTipus->atributs.continguts.info_array.rang.max)) {
				   sinError("Index fora de rang.");
			   }
			   
			   // Retornem la VARIABLE amb l'offset literal
			   memcpy(&$$, &$1, sizeof(struct Simbol));
			   $$.atributs.continguts.info_variable.offset_literal = TRUE;
			   if (symTipus->atributs.continguts.info_array.tipus != TIPUS_DERIVAT) {
				   $$.atributs.continguts.info_variable.offset = ($3.atributs.continguts.info_variable.valor.integer - (int)symTipus->atributs.continguts.info_array.rang.min) * obtenir_tamany_tipus_basic($3.atributs.continguts.info_variable.tipus);
			   } else {
				   $$.atributs.continguts.info_variable.offset = ($3.atributs.continguts.info_variable.valor.integer - (int)symTipus->atributs.continguts.info_array.rang.min) * calcular_tamany_tipus(taulaSimbolActual, symTipus->atributs.continguts.info_array.nom_tipus);
			   }
		   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   if (($3.atributs.continguts.info_variable.valor.character < (int)symTipus->atributs.continguts.info_array.rang.min)
				   || ($3.atributs.continguts.info_variable.valor.character > (int)symTipus->atributs.continguts.info_array.rang.max)) {
				   sinError("Index fora de rang.");
			   }
			   
			   // Retornem la VARIABLE amb l'offset literal
			   memcpy(&$$, &$1, sizeof(struct Simbol));
			   $$.atributs.continguts.info_variable.offset_literal = TRUE;
			   if (symTipus->atributs.continguts.info_array.tipus != TIPUS_DERIVAT) {
				   $$.atributs.continguts.info_variable.offset = ($3.atributs.continguts.info_variable.valor.integer - (int)symTipus->atributs.continguts.info_array.rang.min) * obtenir_tamany_tipus_basic($3.atributs.continguts.info_variable.tipus);
			   } else {
				   $$.atributs.continguts.info_variable.offset = ($3.atributs.continguts.info_variable.valor.integer - (int)symTipus->atributs.continguts.info_array.rang.min) * calcular_tamany_tipus(taulaSimbolActual, symTipus->atributs.continguts.info_array.nom_tipus);
			   }
		   } else {
			   sinError("El tipus de l'index d'un array nomes pot ser 'integer' o 'character'.");
		   }
		   
		   
	   } else {
		   /*$$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.literal = FALSE;*/
		   sinLog("\t** No son literals.");
		   
		   // Retornem la VARIABLE amb l'offset no literal
		   memcpy(&$$, &$1, sizeof(struct Simbol));
		   $$.atributs.continguts.info_variable.offset_literal = FALSE;
		   sprintf($$.atributs.continguts.info_variable.nom_offset, "%s", $3.nom);
	   }
	   
	   
	   Simbol *symTipusReferenciat;
	   if ((symTipus->atributs.continguts.info_array.tipus == TIPUS_DERIVAT)
			   || (symTipus->atributs.continguts.info_array.tipus == TIPUS_INTEGER_DERIVAT)) {
		   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, symTipus->atributs.continguts.info_array.nom_tipus);
	   
	   } else {
		   switch(symTipus->atributs.continguts.info_array.tipus) {
			   case TIPUS_INTEGER:
				   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, "integer");
				   break;
			   case TIPUS_FLOAT:
				   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, "float");
				   break;
			   case TIPUS_CHARACTER:
				   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, "character");
				   break;
			   case TIPUS_STRING:
				   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, "string");
				   break;
			   case TIPUS_BOOLEAN:
				   symTipusReferenciat = busca_totes_taules(taulaSimbolActual, "boolean");
				   break;
			   default:
				   sinError("??");
				   break;
		   }
	   }
	   
	   $$.atributs.continguts.info_variable.offset_simbol = symTipusReferenciat;
	   
	   /*// Retornem el tipus (ARRAY) de la variable
	   memcpy(&$$, symTipus, sizeof(struct Simbol));*/
	   
	   
   }
;


expression:   
   relation {
	   sinLog(" -> expression -> relation");
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   | expression PR_AND mark relation {
	   if (($1.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN) 
		   || ($4.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN)) {
	   		sinError("L'operador AND ha de trobar-se entre tipus booleans");
	   }
	   
	   // Codi 3 adreces
	   if ($1.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $1.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
		   $3++;	// La marca tambe es deplaça
		   incrementar_elements($4.llista_cert); // Els punts certs/fals de la segona expressio, tambe es deplaçen
		   incrementar_elements($4.llista_fals);
	   }
	   if ($4.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $4.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
	   }
	   
	   if (($1.atributs.continguts.info_variable.literal) && ($4.atributs.continguts.info_variable.literal)) {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = TRUE;
		   if (($1.atributs.continguts.info_variable.valor.boolean == TRUE) && ($4.atributs.continguts.info_variable.valor.boolean == TRUE)) {
			   $$.atributs.continguts.info_variable.valor.boolean = TRUE;

		   } else {
			   $$.atributs.continguts.info_variable.valor.boolean = FALSE;

		   }

	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
	   }
	   // Back patching
	   completa_goto(codi, $1.llista_cert, $3);
	   $$.llista_cert = $4.llista_cert;
	   $$.llista_fals = (Cua*)fusionar_cues($1.llista_fals, $4.llista_fals);
	   $$.sentencies = NULL;
   }
   | expression PR_OR mark relation {
	   if (($1.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN) 
		   || ($4.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN)) {
	   		sinError("L'operador AND ha de trobar-se entre tipus booleans");
	   }
	   
	   if ($1.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $1.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
		   $3++;	// La marca tambe es deplaça
		   incrementar_elements($4.llista_cert); // Els punts certs/fals de la segona expressio, tambe es deplaçen
		   incrementar_elements($4.llista_fals);
	   }
	   if ($4.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $4.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
	   }
	   
	   if (($1.atributs.continguts.info_variable.literal) && ($4.atributs.continguts.info_variable.literal)) {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = TRUE;
		   if (($1.atributs.continguts.info_variable.valor.boolean == TRUE) || ($4.atributs.continguts.info_variable.valor.boolean == TRUE)) {
			   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
			   
		   } else {
			   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
			   
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
	   }
	   // Back patching
	   completa_goto(codi, $1.llista_fals, $3);
	   $$.llista_cert = (Cua*)fusionar_cues($1.llista_cert, $4.llista_cert);
	   $$.llista_fals = $4.llista_fals;
	   $$.sentencies = NULL;
   }
   | expression PR_AND PR_THEN mark relation {
	   // e_left and then e_right: s'avalua e_left i, si es certa, s'avalua e_right
	   if (($1.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN) 
		   || ($5.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN)) {
	   		sinError("L'operador AND ha de trobar-se entre tipus booleans");
	   }
	   
	   // Codi 3 adreces
	   if ($1.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $1.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
		   $4++;	// La marca tambe es deplaça
		   incrementar_elements($5.llista_cert); // Els punts certs/fals de la segona expressio, tambe es deplaçen
		   incrementar_elements($5.llista_fals);
	   }
	   if ($5.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $5.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
	   }
	   
	   if (($1.atributs.continguts.info_variable.literal) && ($5.atributs.continguts.info_variable.literal)) {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = TRUE;
		   if ($1.atributs.continguts.info_variable.valor.boolean == TRUE) {
			   if ($5.atributs.continguts.info_variable.valor.boolean == TRUE) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
			   }
		   } else {
			   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
			   
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;

	   }
	   // Back patching
	   // NOTE: Farem que el "bool AND bool" i "bool AND THEN bool" funcionin igual
	   completa_goto(codi, $1.llista_cert, $4);
	   $$.llista_cert = $5.llista_cert;
	   $$.llista_fals = (Cua*)fusionar_cues($1.llista_fals, $5.llista_fals);
	   $$.sentencies = NULL;
   }
   | expression PR_OR PR_ELSE mark relation {
	   // e_left or else e_right: s'avalua e_left i, si es falsa, s'avalua e_right
	   if (($1.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN) 
		   || ($5.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN)) {
	   		sinError("L'operador AND ha de trobar-se entre tipus booleans");
	   }
	   
	   // Codi 3 adreces
	   if ($1.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $1.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
		   $4++;	// La marca tambe es deplaça
		   incrementar_elements($5.llista_cert); // Els punts certs/fals de la segona expressio, tambe es deplaçen
		   incrementar_elements($5.llista_fals);
	   }
	   if ($5.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $5.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
	   }
	   
	   if (($1.atributs.continguts.info_variable.literal) && ($5.atributs.continguts.info_variable.literal)) {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = TRUE;
		   if ($1.atributs.continguts.info_variable.valor.boolean == FALSE) {
			   if ($5.atributs.continguts.info_variable.valor.boolean == TRUE) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
			   }
		   } else {
			   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
			   
		   }
	   } else {
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
	   }
	   // Back patching
	   // NOTE: Farem que el "bool OR bool" i "bool OR ELSE bool" funcionin igual
	   completa_goto(codi, $1.llista_fals, $4);
	   $$.llista_cert = (Cua*)fusionar_cues($1.llista_cert, $5.llista_cert);
	   $$.llista_fals = $5.llista_fals;
	   $$.sentencies = NULL;
   }
;

mark:
	{
		$$ = seguent_quadruple;
	}
;

n_mark:
	{
		$$.llista_seguents = (Cua*)crear_cua();
		int *pos = (int*)malloc(sizeof(int));
		*pos = seguent_quadruple;
		encuar($$.llista_seguents, pos);
	   
		emite(codi, GOTO, NULL, NULL, NULL);
		seguent_quadruple++;
	}
;

relation:
   simple_expression {
	    sinLog(" -> relation -> simple_expression");
		memcpy(&$$, &$1,sizeof(struct Simbol));
   }
   | simple_expression '=' simple_expression {
	   sinLog(" -> relation -> simple_expression '=' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT) || ($1.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer == $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat == $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character == $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (strcmp($1.atributs.continguts.info_variable.valor.string, $3.atributs.continguts.info_variable.valor.string) == 0) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
				   
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   
		   
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
			   char string1[STRING_LENGTH];
			   char string2[STRING_LENGTH];
			   
			   // A string1 i string2 fiquem els noms de les 2 variables string a comparar
			   if (($1.atributs.continguts.info_variable.literal == FALSE) && ($3.atributs.continguts.info_variable.literal == FALSE)) {
				   sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
			   } else if ($1.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(string1, "%s", $1.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string2, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }

			   } else if ($3.atributs.continguts.info_variable.literal == FALSE) {
				   //sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string1, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($1.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $1.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   sinError("No es pot donar aquest cas.");
			   }
			   
			   // Generem la comparacio de 2 variables (string1, string2)
			   int i;
			   
			   // $t0 := string1(0); -- tamany
			   Simbol *sym_tmp1 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp1);
			   sym_tmp1->atributs.tipus = VARIABLE;
			   sym_tmp1->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_tmp1->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp1->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp1->nom, string1, arg3);
			   seguent_quadruple++;
			   
			   // $t1 := string2(0); -- tamany
			   Simbol *sym_tmp2 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp2);
			   sym_tmp2->atributs.tipus = VARIABLE;
			   sym_tmp2->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_tmp2->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp2->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp2->nom, string2, arg3);
			   seguent_quadruple++;
			   
			   // Back patching
			   $$.llista_fals = (Cua*)crear_cua();
			   int *pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_fals, pos);
			   
			   // if size1 /= size2 then goto [false]
			   emite(codi, IF_NE, sym_tmp1->nom, sym_tmp2->nom, NULL);
			   seguent_quadruple++;
			   
			   // Back patching
			   $$.llista_cert = (Cua*)crear_cua();
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_cert, pos);
			   
			   // if size1 = 0 then goto [true]
			   sprintf(arg2, "0");
			   emite(codi, IF_EQ, sym_tmp1->nom, arg2, NULL);
			   seguent_quadruple++;
			   
			   // $t2 := string[$t0];
			   Simbol *sym_tmp3 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp3);
			   sym_tmp3->atributs.tipus = VARIABLE;
			   sym_tmp3->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_tmp3->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp3->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp3->nom, string1, sym_tmp1->nom);
			   seguent_quadruple++;
			   
			   // $t3 := string[$t0];
			   Simbol *sym_tmp4 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp4);
			   sym_tmp4->atributs.tipus = VARIABLE;
			   sym_tmp4->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_tmp4->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp4->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp4->nom, string2, sym_tmp1->nom);
			   seguent_quadruple++;
			   
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_fals, pos);
			   
			   // if $2 /= $t3 then goto [false]
			   emite(codi, IF_NE, sym_tmp3->nom, sym_tmp4->nom, NULL);
			   seguent_quadruple++;
			   
			   sprintf(arg3, "1");
			   emite(codi, SUBI, sym_tmp1->nom, sym_tmp1->nom, arg3);
			   seguent_quadruple++;
			   
			   sprintf(arg1, "%i",seguent_quadruple - 5);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
		   } else {
			   
			   // Back patching
			   $$.llista_cert = (Cua*)crear_cua();
			   int *pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_cert, pos);
		   
			   // Passem a string els parametres de la comparacio
			   if ($1.atributs.continguts.info_variable.literal) {
				   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg1, "'%c'", $1.atributs.continguts.info_variable.valor.character);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Ja tractat anteriorment
				   }
			   } else {
				   sprintf(arg1, "%s", $1.nom);
			   }
			   
			   if ($3.atributs.continguts.info_variable.literal) {
				   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Ja tractat anteriorment
				   }
			   } else {
				   sprintf(arg2, "%s", $3.nom);
			   }
			   
			   //$$.sentencies = (Cua*)crear_cua();
			   emite(codi, IF_EQ, arg1, arg2, NULL);
			   seguent_quadruple++;
			   
			   $$.llista_fals = (Cua*)crear_cua();
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_fals, pos);
			   
			   //$$.sentencies = (Cua*)crear_cua();
			   emite(codi, GOTO, NULL, NULL, NULL);
			   seguent_quadruple++;
		   }
	   }

   }
   | simple_expression '<' simple_expression {
	   sinLog(" -> relation -> simple_expression '<' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer < $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat < $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character < $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Comparacio no valida pel tipus usat.");
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Back patching
		   $$.llista_cert = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_cert, pos);
			
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   
		   // Passem a string els parametres de la comparacio
		   if ($1.atributs.continguts.info_variable.literal) {
			   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg1, "%s", $1.nom);
		   }
		   
		   if ($3.atributs.continguts.info_variable.literal) {
			   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg2, "%s", $3.nom);
		   }
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   emite(codi, IF_LTF, arg1, arg2, NULL);
		   } else {
			   emite(codi, IF_LTI, arg1, arg2, NULL);
		   }
		   seguent_quadruple++;
		   
		   $$.llista_fals = (Cua*)crear_cua();
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
	   
	   
   }
   | simple_expression '>' simple_expression {
	   sinLog(" -> relation -> simple_expression '>' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer > $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat > $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character > $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Comparacio no valida pel tipus usat.");
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Back patching
		   $$.llista_cert = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_cert, pos);
			
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   
		   // Passem a string els parametres de la comparacio
		   if ($1.atributs.continguts.info_variable.literal) {
			   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg1, "%s", $1.nom);
		   }
		   
		   if ($3.atributs.continguts.info_variable.literal) {
			   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg2, "%s", $3.nom);
		   }
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   emite(codi, IF_GTF, arg1, arg2, NULL);
		   } else {
			   emite(codi, IF_GTI, arg1, arg2, NULL);
		   }
		   seguent_quadruple++;
		   
		   $$.llista_fals = (Cua*)crear_cua();
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
	   
   }
   | simple_expression PR_DIFFERENT simple_expression {
	   sinLog(" -> relation -> simple_expression '!=' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT) || ($1.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer == $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat == $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character == $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (strcmp($1.atributs.continguts.info_variable.valor.string, $3.atributs.continguts.info_variable.valor.string) == 0) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
				   
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   
		   
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
			   char string1[STRING_LENGTH];
			   char string2[STRING_LENGTH];
			   
			   // A string1 i string2 fiquem els noms de les 2 variables string a comparar
			   if (($1.atributs.continguts.info_variable.literal == FALSE) && ($3.atributs.continguts.info_variable.literal == FALSE)) {
				   sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
			   } else if ($1.atributs.continguts.info_variable.literal == FALSE) {
				   sprintf(string1, "%s", $1.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string2, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }

			   } else if ($3.atributs.continguts.info_variable.literal == FALSE) {
				   //sprintf(string1, "%s", $1.nom);
				   sprintf(string2, "%s", $3.nom);
				   
				   // Passem l'string literal a una variable temporal
				   Simbol *sym_tmp_string = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
				   sprintf(string1, "%s", sym_tmp_string->nom);
				   encuar(cua_actual_var_tmp, sym_tmp_string);
				   sym_tmp_string->atributs.tipus = VARIABLE;
				   sym_tmp_string->atributs.continguts.info_variable.tipus = TIPUS_STRING;
				   sym_tmp_string->atributs.continguts.info_variable.literal = FALSE;
				   sym_tmp_string->atributs.continguts.info_variable.constant = FALSE;
				   
				   int caracters = strlen($1.atributs.continguts.info_variable.valor.string);
				   int i;
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $1.atributs.continguts.info_variable.valor.string[i]);
					   
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp_string->nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   sinError("No es pot donar aquest cas.");
			   }
			   
			   // Generem la comparacio de 2 variables (string1, string2)
			   int i;
			   
			   // $t0 := string1(0); -- tamany
			   Simbol *sym_tmp1 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp1);
			   sym_tmp1->atributs.tipus = VARIABLE;
			   sym_tmp1->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_tmp1->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp1->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp1->nom, string1, arg3);
			   seguent_quadruple++;
			   
			   // $t1 := string2(0); -- tamany
			   Simbol *sym_tmp2 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp2);
			   sym_tmp2->atributs.tipus = VARIABLE;
			   sym_tmp2->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
			   sym_tmp2->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp2->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp2->nom, string2, arg3);
			   seguent_quadruple++;
			   
			   // Back patching
			   $$.llista_cert = (Cua*)crear_cua();
			   int *pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_cert, pos);
			   
			   // if size1 /= size2 then goto [false]
			   emite(codi, IF_NE, sym_tmp1->nom, sym_tmp2->nom, NULL);
			   seguent_quadruple++;
			   
			   // Back patching
			   $$.llista_fals = (Cua*)crear_cua();
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_fals, pos);
			   
			   // if size1 = 0 then goto [true]
			   sprintf(arg2, "0");
			   emite(codi, IF_EQ, sym_tmp1->nom, arg2, NULL);
			   seguent_quadruple++;
			   
			   // $t2 := string[$t0];
			   Simbol *sym_tmp3 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp3);
			   sym_tmp3->atributs.tipus = VARIABLE;
			   sym_tmp3->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_tmp3->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp3->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp3->nom, string1, sym_tmp1->nom);
			   seguent_quadruple++;
			   
			   // $t3 := string[$t0];
			   Simbol *sym_tmp4 = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp4);
			   sym_tmp4->atributs.tipus = VARIABLE;
			   sym_tmp4->atributs.continguts.info_variable.tipus = TIPUS_CHARACTER;
			   sym_tmp4->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp4->atributs.continguts.info_variable.constant = FALSE;
			   sprintf(arg3, "0");
			   emite(codi, ASSIGN_FROM_TABLE, sym_tmp4->nom, string2, sym_tmp1->nom);
			   seguent_quadruple++;
			   
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_cert, pos);
			   
			   // if $2 /= $t3 then goto [false]
			   emite(codi, IF_NE, sym_tmp3->nom, sym_tmp4->nom, NULL);
			   seguent_quadruple++;
			   
			   sprintf(arg3, "1");
			   emite(codi, SUBI, sym_tmp1->nom, sym_tmp1->nom, arg3);
			   seguent_quadruple++;
			   
			   sprintf(arg1, "%i",seguent_quadruple - 5);
			   emite(codi, GOTO, arg1, NULL, NULL);
			   seguent_quadruple++;
		   } else {
			   
			   // Back patching
			   $$.llista_cert = (Cua*)crear_cua();
			   int *pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_cert, pos);
		   
			   // Passem a string els parametres de la comparacio
			   if ($1.atributs.continguts.info_variable.literal) {
				   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg1, "'%c'", $1.atributs.continguts.info_variable.valor.character);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
				   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Ja tractat anteriorment
				   }
			   } else {
				   sprintf(arg1, "%s", $1.nom);
			   }
			   
			   if ($3.atributs.continguts.info_variable.literal) {
				   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Ja tractat anteriorment
				   }
			   } else {
				   sprintf(arg2, "%s", $3.nom);
			   }
			   
			   //$$.sentencies = (Cua*)crear_cua();
			   emite(codi, IF_NE, arg1, arg2, NULL);
			   seguent_quadruple++;
			   
			   $$.llista_fals = (Cua*)crear_cua();
			   pos = (int*)malloc(sizeof(int));
			   *pos = seguent_quadruple;
			   encuar($$.llista_fals, pos);
			   
			   //$$.sentencies = (Cua*)crear_cua();
			   emite(codi, GOTO, NULL, NULL, NULL);
			   seguent_quadruple++;
		   }
	   }
	   
   }
   | simple_expression PR_MINEQU simple_expression {
	   sinLog(" -> relation -> simple_expression '<=' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer <= $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat <= $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character <= $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Comparacio no valida pel tipus usat.");
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Back patching
		   $$.llista_cert = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_cert, pos);
			
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   
		   // Passem a string els parametres de la comparacio
		   if ($1.atributs.continguts.info_variable.literal) {
			   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg1, "%s", $1.nom);
		   }
		   
		   if ($3.atributs.continguts.info_variable.literal) {
			   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg2, "%s", $3.nom);
		   }
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   emite(codi, IF_LEF, arg1, arg2, NULL);
		   } else {
			   emite(codi, IF_LEI, arg1, arg2, NULL);
		   }
		   seguent_quadruple++;
		   
		   $$.llista_fals = (Cua*)crear_cua();
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
	   
   }
   | simple_expression PR_MAJEQU simple_expression {
	   sinLog(" -> relation -> simple_expression '>=' simple_expression");
	   if ($1.atributs.tipus != $3.atributs.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus) {
		   sinError("Les expressions han de ser del mateix tipus.");
	   }
	   if (($1.atributs.continguts.info_variable.literal) && ($3.atributs.continguts.info_variable.literal)) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.integer >= $3.atributs.continguts.info_variable.valor.integer) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.vfloat >= $3.atributs.continguts.info_variable.valor.vfloat) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if ($1.atributs.continguts.info_variable.valor.character >= $3.atributs.continguts.info_variable.valor.character) {
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Comparacio no valida pel tipus usat.");
		   }
	   } else {
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Back patching
		   $$.llista_cert = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_cert, pos);
			
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   
		   // Passem a string els parametres de la comparacio
		   if ($1.atributs.continguts.info_variable.literal) {
			   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg1, "%i", $1.atributs.continguts.info_variable.valor.integer);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg1, "%g", $1.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg1, "%s", $1.nom);
		   }
		   
		   if ($3.atributs.continguts.info_variable.literal) {
			   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
			   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   sinError("Comparacio no valida pel tipus usat.");
			   } else {
				   sinError("Comparacio no valida pel tipus usat.");
			   }
		   } else {
			   sprintf(arg2, "%s", $3.nom);
		   }
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   emite(codi, IF_GEF, arg1, arg2, NULL);
		   } else {
			   emite(codi, IF_GEI, arg1, arg2, NULL);
		   }
		   seguent_quadruple++;
		   
		   $$.llista_fals = (Cua*)crear_cua();
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   //$$.sentencies = (Cua*)crear_cua();
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
	   
	   
   }
   
   | simple_expression PR_IN range {
	   sinLog(" -> relation -> simple_expression PR_IN range");
	   if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_array.rang.tipus) {
		   sinError("El rang no es del mateix tipus que l'element comparat.");
	   }
	   if ($1.atributs.continguts.info_variable.literal) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.integer >= (int)$3.atributs.continguts.info_array.rang.min) 
			   		&& ($1.atributs.continguts.info_variable.valor.integer <= (int)$3.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.vfloat >= $3.atributs.continguts.info_array.rang.min) 
			   		&& ($1.atributs.continguts.info_variable.valor.vfloat <= $3.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.character >= (int)$3.atributs.continguts.info_array.rang.min) 
			   		&& ($1.atributs.continguts.info_variable.valor.character <= (int)$3.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Nomes es poden avaluar en rangs els tipus 'integer', 'float' i 'character'.");
		   }
	   } else {
		   
		   // La operacio de comparacio variara segons si es enter o float
		   Tipus_Operacio operacio_IF_GT;
		   Tipus_Operacio operacio_IF_LT;
		   char rang_min[STRING_LENGTH];
		   char rang_max[STRING_LENGTH];
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   operacio_IF_GT = IF_GTI;
			   operacio_IF_LT = IF_LTI;
			   sprintf(rang_min, "%i", (int)$3.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%i", (int)$3.atributs.continguts.info_array.rang.max);
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) {
			   operacio_IF_GT = IF_GTI;
			   operacio_IF_LT = IF_LTI;
			   sprintf(rang_min, "%i", (int)$3.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%i", (int)$3.atributs.continguts.info_array.rang.max);
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   operacio_IF_GT = IF_GTF;
			   operacio_IF_LT = IF_LTF;
			   sprintf(rang_min, "%g", $3.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%g", $3.atributs.continguts.info_array.rang.max);
		   } else {
				sinError("Nomes es poden avaluar en rangs els tipus 'integer', 'float' i 'character'.");
		   }
			
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Codi 3 adreces
		   // NOTE: Traduim el FOR com si fos un WHILE (variable > rang.min and variable < rang.max) { variable++; ... }
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   // If variable > rang_min goto (next if)
		   sprintf(arg1, "%s", $1.nom);
		   //sprintf(arg2, "%i", (int)$3.atributs.continguts.info_array.rang.min);
		   sprintf(arg3, "%i", seguent_quadruple+2);
		   emite(codi, operacio_IF_GT, arg1, rang_min, arg3);
		   seguent_quadruple++;
		   
		   // If NOT(variable > rang_min) GOTO (despres sentencies, encara desconegut)
		   $$.llista_fals = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
		   
		   // If variable < rang_max goto (sentencies)
		   sprintf(arg1, "%s", $1.nom);
		   //sprintf(arg2, "%i", (int)$3.atributs.continguts.info_array.rang.max);
		   sprintf(arg3, "%i", seguent_quadruple+2);
		   emite(codi, operacio_IF_LT, arg1, rang_max, arg3);
		   seguent_quadruple++;
		   
		   // If NOT(variable < rang_max) GOTO (despres sentencies, encara desconegut)
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
	   

   }
   | simple_expression PR_NOT PR_IN range {
	   sinLog(" -> relation -> simple_expression PR_NOT PR_IN range");
	   if ($1.atributs.continguts.info_variable.tipus != $4.atributs.continguts.info_array.rang.tipus) {
		   sinError("El rang no es del mateix tipus que l'element comparat.");
	   }
	   
	   if ($1.atributs.continguts.info_variable.literal) {
		   if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.integer < (int)$4.atributs.continguts.info_array.rang.min) 
			   		|| ($1.atributs.continguts.info_variable.valor.integer > (int)$4.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.vfloat < $4.atributs.continguts.info_array.rang.min) 
			   		|| ($1.atributs.continguts.info_variable.valor.vfloat > $4.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   $$.atributs.tipus = VARIABLE;
			   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
			   $$.atributs.continguts.info_variable.literal = TRUE;
			   if (($1.atributs.continguts.info_variable.valor.character < (int)$4.atributs.continguts.info_array.rang.min) 
			   		|| ($1.atributs.continguts.info_variable.valor.character > (int)$4.atributs.continguts.info_array.rang.max)){
				   $$.atributs.continguts.info_variable.valor.boolean = TRUE;
				   
				   // Codi 3 adreces
				   // E.listaverdad = crealista(sigtecuad);
				   $$.llista_cert = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_cert, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   } else {
				   $$.atributs.continguts.info_variable.valor.boolean = FALSE;
				   
				   // Codi 3 adreces
				   // E.listafalso = crealista(sigtecuad);
				   $$.llista_fals = (Cua*)crear_cua();
				   int *pos = (int*)malloc(sizeof(int));
				   *pos = seguent_quadruple;
				   encuar($$.llista_fals, pos);
				   
				   emite(codi, GOTO, NULL, NULL, NULL);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("Nomes es poden avaluar en rangs els tipus 'integer', 'float' i 'character'.");
		   }
	   } else {
		   
		   // La operacio de comparacio variara segons si es enter o float
		   Tipus_Operacio operacio_IF_GT;
		   Tipus_Operacio operacio_IF_LT;
		   char rang_min[STRING_LENGTH];
		   char rang_max[STRING_LENGTH];
		   if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   operacio_IF_GT = IF_GTI;
			   operacio_IF_LT = IF_LTI;
			   sprintf(rang_min, "%i", (int)$4.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%i", (int)$4.atributs.continguts.info_array.rang.max);
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) {
			   operacio_IF_GT = IF_GTI;
			   operacio_IF_LT = IF_LTI;
			   sprintf(rang_min, "%i", (int)$4.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%i", (int)$4.atributs.continguts.info_array.rang.max);
		   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   operacio_IF_GT = IF_GTF;
			   operacio_IF_LT = IF_LTF;
			   sprintf(rang_min, "%g", $4.atributs.continguts.info_array.rang.min);
			   sprintf(rang_max, "%g", $4.atributs.continguts.info_array.rang.max);
		   } else {
				sinError("Nomes es poden avaluar en rangs els tipus 'integer', 'float' i 'character'.");
		   }
		   
		   // Algun dels 2 o els 2 no son literals
		   $$.atributs.tipus = VARIABLE;
		   $$.atributs.continguts.info_variable.tipus = TIPUS_BOOLEAN;
		   $$.atributs.continguts.info_variable.literal = FALSE;
		   
		   // Codi 3 adreces
		   // NOTE: Traduim el FOR com si fos un WHILE (variable > rang.min and variable < rang.max) { variable++; ... }
		   char arg1[STRING_LENGTH];
		   char arg2[STRING_LENGTH];
		   char arg3[STRING_LENGTH];
		   // If variable < rang_min goto (next if)
		   sprintf(arg1, "%s", $1.nom);
		   //sprintf(arg2, "%i", (int)$4.atributs.continguts.info_array.rang.min);
		   sprintf(arg3, "%i", seguent_quadruple+2);
		   emite(codi, operacio_IF_LT, arg1, rang_min, arg3);
		   seguent_quadruple++;
		   
		   // If NOT(variable < rang_min) GOTO (despres sentencies, encara desconegut)
		   $$.llista_fals = (Cua*)crear_cua();
		   int *pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
		   
		   // If variable > rang_max goto (sentencies)
		   sprintf(arg1, "%s", $1.nom);
		   //sprintf(arg2, "%i", (int)$4.atributs.continguts.info_array.rang.max);
		   sprintf(arg3, "%i", seguent_quadruple+2);
		   emite(codi, operacio_IF_GT, arg1, rang_max, arg3);
		   seguent_quadruple++;
		   
		   // If NOT(variable > rang_max) GOTO (despres sentencies, encara desconegut)
		   pos = (int*)malloc(sizeof(int));
		   *pos = seguent_quadruple;
		   encuar($$.llista_fals, pos);
		   
		   emite(codi, GOTO, NULL, NULL, NULL);
		   seguent_quadruple++;
	   }
   }
;

range:
   simple_expression PR_2_POINTS simple_expression {
	    
	   if ($1.atributs.continguts.info_variable.literal == FALSE || $3.atributs.continguts.info_variable.literal == FALSE) {
			sinError("Les expressions de rang han de ser literals");
	   }
		sinLog(" -> range -> simple_expression T_2_POINTS simple_expression");
		if ($1.atributs.continguts.info_variable.tipus != $3.atributs.continguts.info_variable.tipus){
			sinError("Els tipus de limit de rang tenen tipus diferents");
		}
		/*if (strcmp($1.name_type,$3.name_type)!=0){
			sinError("Limit range types have different types %s != %s",$1.name_type,$3.name_type);
		}*/
		
		
		Rang *rang = malloc(sizeof(Rang));
		rang->tipus = $1.atributs.continguts.info_variable.tipus;
		switch (rang->tipus){
			case (TIPUS_FLOAT):
				if ($1.atributs.continguts.info_variable.valor.vfloat > $3.atributs.continguts.info_variable.valor.vfloat){
					sinError("La segona expressio ha de ser major o igual que la primera expressio");
				}
				rang->max = $3.atributs.continguts.info_variable.valor.vfloat;
				rang->min = $1.atributs.continguts.info_variable.valor.vfloat;
				break;
			case (TIPUS_INTEGER):
			case (TIPUS_INTEGER_DERIVAT):
				if ($1.atributs.continguts.info_variable.valor.integer > $3.atributs.continguts.info_variable.valor.integer){
					sinError("La segona expressio ha de ser major o igual que la primera expressio");
				}
				rang->max = $3.atributs.continguts.info_variable.valor.integer;
				rang->min = $1.atributs.continguts.info_variable.valor.integer;
				break;
			case(TIPUS_CHARACTER):
				if ($1.atributs.continguts.info_variable.valor.character > $3.atributs.continguts.info_variable.valor.character){
					sinError("La segona expressio ha de ser major o igual que la primera expressio");
				}
				rang->max = (int)$3.atributs.continguts.info_variable.valor.character;
				rang->min = (int)$1.atributs.continguts.info_variable.valor.character;
				break;
			default:
				sinError("El tipus no pot ser un rang");
		}
		// Passem el rang cap adalt
		sprintf($$.nom, "");	// No es encara un simbol de la TS
		$$.atributs.tipus = ARRAY_TYPE;
		$$.atributs.continguts.info_array.rang.min = rang->min;
		$$.atributs.continguts.info_array.rang.max = rang->max;
		$$.atributs.continguts.info_array.rang.tipus = rang->tipus;
   }
;


enumeration_type_definition:
	// Potser sera necessari introduir codi intermedi despres del '(' per inicialitzar
	// una taula de simbols del enumerador
	'(' {
			cua_tmp = (Cua *)crear_cua();
			contador = 0;
	}
	enumeration_literal_specification enumeration_type_definition_element {
		sinLog(" -> enumeration_type_definition -> '(' aux_enumeration_type_definition ')'");
		$$.atributs.tipus = ENUMERATED_TYPE;
		$$.atributs.continguts.info_enumerated.elements = cua_tmp;
	}
;

enumeration_type_definition_element: /* Gramatica auxiliar de 'enumeration_type_definition' */
   ',' enumeration_literal_specification enumeration_type_definition_element { 
	   sinLog(" -> enumeration_type_definition_element -> seguent element");
   }
   | ')' { sinLog(" -> enumeration_type_definition_element -> tanca parentesis )"); }
;

enumeration_literal_specification:
	// Tractament de casos de identificadors ja definits?
   defining_identifier {
	   	sinLog(" -> enumeration_literal_specification -> defining_identifier");
		sinLog("Nou tipus enumerat %s insertat", $1);
		Simbol *element = (Simbol*)malloc(sizeof(struct Simbol));
		element->atributs.tipus = VARIABLE;
		element->atributs.continguts.info_variable.valor.taula = NULL;
		element->atributs.continguts.info_variable.inicialitzat = TRUE;
		element->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
		element->atributs.continguts.info_variable.constant = TRUE;
		sprintf(element->nom, $1);
		element->atributs.continguts.info_variable.valor.integer = contador;
		contador++;
		// cua_tmp ha sigut anteriorment inicialitzada en nodes pares
		encuar(cua_tmp, element);
   } 
   | defining_character_literal {
	    // IDEM defining_identifier
	   	sinLog(" -> enumeration_literal_specification -> defining_identifier");
		sinLog("Nou tipus enumerat %s insertat", $1);
		Simbol *element = (Simbol*)malloc(sizeof(struct Simbol));
		element->atributs.tipus = VARIABLE;
		element->atributs.continguts.info_variable.valor.taula = NULL;
		element->atributs.continguts.info_variable.inicialitzat = TRUE;
		element->atributs.continguts.info_variable.tipus = TIPUS_INTEGER;
		element->atributs.continguts.info_variable.constant = TRUE;
		sprintf(element->nom, $1);
		element->atributs.continguts.info_variable.valor.integer = contador;
		contador++;
		// cua_tmp ha sigut anteriorment inicialitzada en nodes pares
		encuar(cua_tmp, element);
   }
   | defined_identifier {} // NOTE: Control de duplicats
;

defining_character_literal:
   PR_CHARACTER_LITERAL {
	   sinLog(" -> defining_character_literal -> PR_CHARACTER_LITERAL");
	   sprintf($$, "%c", $1);
   }
;

defining_identifier:
   PR_ID {
   	sinLog(" -> defining_identifier -> PR_ID");
	// Pasem el nom de l'identificador cap adalt
	snprintf($$, STRING_LENGTH, "%s" , $1);
   }
;

// NOTE: Creat per controlar errors de identificadors ya declarats
defined_identifier:
	PR_ID_VARIABLE {
		sinError("%s ja esta definit anteriorment com una variable", $1->nom);
	}
	| PR_ID_PROCEDURE {
		sinError("%s ja esta definit anteriorment com un procediment", $1->nom);
	}
	| PR_ID_FUNCTION {
		sinError("%s ja esta definit anteriorment com una funcio", $1->nom);
	}
	| PR_ID_TYPE {
		sinError("%s ja esta definit anteriorment com un tipus", $1->nom);
	}
;

defining_program_unit_name:
   defining_identifier {
	    sinLog(" -> defining program unit name -> defining identifier");
		// Pasem el nom de l'identificador cap adalt
		snprintf($$, STRING_LENGTH, "%s" , $1);
   }
;

designator:
   PR_ID_PROCEDURE {
	   sinLog(" -> designator -> procedure designator");
	   memcpy(&$$, $1, sizeof(struct Simbol));
   }
   | PR_ID_FUNCTION {
	   sinLog(" -> designator -> procedure designator");
	   memcpy(&$$, $1, sizeof(struct Simbol));
   }
   | PR_ID {
	   sinError("Designator '%s' no es correspon amb el nom de la funcio/procediment encaixats.", $1);
   }
;

declarative_part:
   {}
   //|declarative_item {} // Redundant
   | declarative_item declarative_part {}
;

declarative_item:
   basic_declarative_item { sinLog(" -> declarative_item -> basic_declarative_item");}
   | body {sinLog(" -> declarative_item -> body");}
;

body:
   proper_body { sinLog(" -> body -> proper_body");}
;

proper_body:
   subprogram_body {sinLog(" -> proper_body -> subprogram_body");}
;

subprogram_body:
	subprogram_body_element declarative_part PR_BEGIN sequence_of_statements_after_begin PR_END ';' {
	   sinLog("-> subprogram_body -> subprogram_body_element PR_IS declarative_part PR_BEGIN sequence_of_statements PR_END ';'");
	   taulaSimbolActual = taulaSimbolActual->anterior;
	   
	   desenpilar(pila_cues_var_locals);
	   cua_actual_var_locals = (Cua*)cim(pila_cues_var_locals);
	   desenpilar(pila_cues_var_tmp);
	   cua_actual_var_tmp = (Cua*)cim(pila_cues_var_tmp);
		   
	   emite(codi, HALT, NULL, NULL, NULL);
	   seguent_quadruple++;
	   
	   // A $1 tenim una copia del simbol, hem de buscar l'original per canviar l'atribut 'complet'
	   Simbol *sym = obtenir_simbol(taulaSimbolActual, $1.nom);
	   
	   // Indicar que la funcio/procediment ja esta complet
	   if (sym != NULL) {
		   if (sym->atributs.tipus == PROCEDURE) {
			   int id = $1.atributs.continguts.info_procedure.id; // Identificador de sobrecarrega
			   
			   if (id > 0) {
				   // Busquem la sobrecarrega que correspon
				   reiniciar_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   int i;
				   for (i = 1; i < id; i++) {
						seguent_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   }
				   Info_Procedure *info_procedure = (Info_Procedure *)obtenir_element_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   info_procedure->complet = TRUE;
				   info_procedure->sentencies = codi;  // Codi 3 adreces generat
				   codi = NULL;
			   } else {
				   // Principal
				   sym->atributs.continguts.info_procedure.complet = TRUE;
				   sym->sentencies = codi;  // Codi 3 adreces generat
				   sym->atributs.continguts.info_procedure.sentencies = codi;
				   codi = NULL;
			   }
			   
		   } else if (sym->atributs.tipus == FUNCTION) {
			   int id = $1.atributs.continguts.info_function.id; // Identificador de sobrecarrega
			   
			   if (id > 0) {
				   // Busquem la sobrecarrega que correspon
				   reiniciar_PDI(sym->atributs.continguts.info_function.sobrecarregues);
				   int i;
				   for (i = 1; i < id; i++) {
						seguent_PDI(sym->atributs.continguts.info_function.sobrecarregues);
				   }
				   
				   Info_Function *info_function = (Info_Function *)obtenir_element_PDI(sym->atributs.continguts.info_function.sobrecarregues);

				   info_function->complet = TRUE;
				   info_function->sentencies = codi;
				   codi = NULL;
			   } else {
				   // Principal
				   sym->atributs.continguts.info_function.complet = TRUE;
				   sym->sentencies = codi;
				   sym->atributs.continguts.info_function.sentencies = codi;
				   codi = NULL;
			   }
		   } else {
			   sinError("S'esperava PROCEDURE o FUNCTION.");
		   }
	   } else {
		   sinError("Simbol no trobat.");
	   }
   }
   | subprogram_body_element declarative_part PR_BEGIN sequence_of_statements_after_begin PR_END designator ';' { 
	   sinLog("-> subprogram_body -> subprogram_body_element PR_IS declarative_part PR_BEGIN sequence_of_statements PR_END ';'");

	   // Comprovem designator
	   if (strcmp($1.nom, $6.nom) != 0) {
		   sinError("Designator '%s' no es correspon amb el nom de la funcio/procediment %s.", $6.nom, $1.nom);
	   }
	   
	   taulaSimbolActual = taulaSimbolActual->anterior;
	   
	   desenpilar(pila_cues_var_locals);
	   cua_actual_var_locals = (Cua*)cim(pila_cues_var_locals);
	   desenpilar(pila_cues_var_tmp);
	   cua_actual_var_tmp = (Cua*)cim(pila_cues_var_tmp);
	   
	   emite(codi, HALT, NULL, NULL, NULL);
	   seguent_quadruple++;
	   
	   // A $1 tenim una copia del simbol, hem de buscar l'original per canviar l'atribut 'complet'
	   Simbol *sym = obtenir_simbol(taulaSimbolActual, $1.nom);
	   
	   // Indicar que la funcio/procediment ja esta complet
	   if (sym != NULL) {
		   if (sym->atributs.tipus == PROCEDURE) {
			   int id = $1.atributs.continguts.info_procedure.id; // Identificador de sobrecarrega
			   
			   if (id > 0) {
				   // Busquem la sobrecarrega que correspon
				   reiniciar_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   int i;
				   for (i = 0; i < id - 1; i++) {
						seguent_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   }
				   Info_Procedure *info_procedure = (Info_Procedure *)obtenir_element_PDI(sym->atributs.continguts.info_procedure.sobrecarregues);
				   info_procedure->complet = TRUE;
				   info_procedure->sentencies = codi;  // Codi 3 adreces generat
				   codi = NULL;
			   } else {
				   // Principal
				   sym->atributs.continguts.info_procedure.complet = TRUE;
				   sym->sentencies = codi;  // Codi 3 adreces generat
				   sym->atributs.continguts.info_procedure.sentencies = codi;
				   codi = NULL;
			   }
			   
		   } else if (sym->atributs.tipus == FUNCTION) {
			   int id = $1.atributs.continguts.info_function.id; // Identificador de sobrecarrega
			   
			   if (id > 0) {
				   // Busquem la sobrecarrega que correspon
				   reiniciar_PDI(sym->atributs.continguts.info_function.sobrecarregues);
				   int i;
				   for (i = 0; i < id - 1; i++) {
						seguent_PDI(sym->atributs.continguts.info_function.sobrecarregues);
				   }
				   Info_Function *info_function = (Info_Function *)obtenir_element_PDI(sym->atributs.continguts.info_function.sobrecarregues);
				   info_function->complet = TRUE;
				   info_function->sentencies = codi;
				   codi = NULL;
			   } else {
				   // Principal
				   sym->atributs.continguts.info_function.complet = TRUE;
				   sym->sentencies = codi;
				   sym->atributs.continguts.info_function.sentencies = codi;
				   codi = NULL;
			   }
		   } else {
			   sinError("S'esperava PROCEDURE o FUNCTION.");
		   }
	   } else {
		   sinError("Simbol no trobat.");
	   }
   }
;

subprogram_body_element:
	subprogram_specification PR_IS {
		if ($1.atributs.tipus == PROCEDURE) {
		   taulaSimbolActual = $1.atributs.continguts.info_procedure.taula;
		   
		   enpilar(pila_cues_var_locals, infoProcedureActual->variables_locals);
		   cua_actual_var_locals = infoProcedureActual->variables_locals;
		   enpilar(pila_cues_var_tmp, infoProcedureActual->variables_temporals);
		   cua_actual_var_tmp = infoProcedureActual->variables_temporals;
		   
		} else if ($1.atributs.tipus == FUNCTION) {
			taulaSimbolActual = $1.atributs.continguts.info_function.taula;
			
			enpilar(pila_cues_var_locals, infoFunctionActual->variables_locals);
			cua_actual_var_locals = infoFunctionActual->variables_locals;
			enpilar(pila_cues_var_tmp, infoFunctionActual->variables_temporals);
			cua_actual_var_tmp = infoFunctionActual->variables_temporals;
		} else {
		   sinError("S'esperava PROCEDURE o FUNCTION. %i", $1.atributs.tipus);
		}
		memcpy(&$$, &$1, sizeof(struct Simbol));
	}
;

basic_declarative_item:
   basic_declaration { sinLog(" -> basic_declarative_item -> basic_declaration"); }
;

sequence_of_statements_after_begin:
	{
		sinLog(" -> sequence_of_statements_after_begin");
		codi = (Cua*)crear_cua();
		seguent_quadruple = 1;
		num_temporal = 0;
		emite(codi, START, NULL, NULL, NULL);  // El nom del programa s'omplira al fer el print del codi 3 adreces
		seguent_quadruple++;
	}
	sequence_of_statements mark {
		//memcpy(&$$, &$1, sizeof(struct Simbol));
		completa_goto(codi, $2.llista_seguents, $3);
	}
;

sequence_of_statements:
	statement mark sequence_of_statements_element {
		sinLog(" -> sequence_of_statements");
		if (strcmp($3.nom, "$$END_SEQUENCE$$") == 0) {
			$$.llista_seguents = $1.llista_seguents;
		} else {
			completa_goto(codi, $1.llista_seguents, $2);
			$$.llista_seguents = $3.llista_seguents;
		}
	}
;

sequence_of_statements_element: /* Gramatica auxiliar de 'sequence_of_statements'*/
   statement {
	   memcpy(&$$, &$1, sizeof(struct Simbol));  // $$.llista_seguents = $1.llista_seguents;
   }
   | statement mark sequence_of_statements {
	   completa_goto(codi, $1.llista_seguents, $2);
	   $$.llista_seguents = $3.llista_seguents;
   }
   | {
	   sprintf($$.nom, "$$END_SEQUENCE$$");
   }
;

statement:
   simple_statement {
	   $$.sentencies = NULL;
	   $$.llista_seguents = NULL;
	   $$.llista_cert = NULL;
	   $$.llista_fals = NULL;
   }
   | compound_statement {
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

simple_statement:
   null_statement {}
   | assignment_statement {}
   | procedure_call_statement {}
   | return_statement {}
;

null_statement:
   PR_NULL ';' { sinLog(" -> null_statement -> PR_NULL ';'"); }
;

assignment_statement: /* Comprovar que el id es una var i que esta declarada*/
   name PR_EQU expression ';' {
	   // TODO: Assignar el valor
	   
	   // A $1 rebem VARIABLE, si es un record hi trobarem la variable corresponent al element referenciat (e.g. date.month)
	   
	   if ($1.atributs.continguts.info_variable.constant == TRUE) {
		   sinError("No es pot assignar un valor a una variable definida com a constant.");
	   }
	   
	   if ((($1.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT))
	   		|| (($1.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT))){
		   // Enumerat o ARRAY de enumerats
		   
		   // Comprovem si es array de enumerats
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1.atributs.continguts.info_variable.nom_tipus);
		   if ((symTipus->atributs.tipus == ARRAY_TYPE) || (symTipus->atributs.tipus == RECORD_TYPE)) {
			   
			   Simbol *symTipusReferenciat = $1.atributs.continguts.info_variable.offset_simbol;
			   if (symTipusReferenciat == NULL) {
				   if (strcmp($1.atributs.continguts.info_variable.nom_tipus, $3.atributs.continguts.info_variable.nom_tipus) == 0) {
					   sinError("No es permeten aquest tipus d'assignacions.");
				   } else {
					   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (20)");
				   }
			   } else if (symTipusReferenciat->atributs.tipus == ENUMERATED_TYPE) { 
				   // Array de enumerats
					   
				   if ($3.atributs.continguts.info_variable.literal == TRUE) {
					   // Posicio array enumerat := valor enumerat
					   // Comprovar que l'enumerat assignat pertany al tipus de la variable enumerada
					   Cua *elements = symTipusReferenciat->atributs.continguts.info_enumerated.elements;
					   Boolean trobat = FALSE;
						   
					   reiniciar_PDI(elements);
					   Simbol *element = (Simbol *)obtenir_element_PDI(elements);
					   while(element != NULL) {
							// Comprovar parametres
							if (strcmp(element->nom, $3.nom) == 0) {
								trobat = TRUE;
								break;
							}
							
							seguent_PDI(elements);
							element = (Simbol *)obtenir_element_PDI(elements);
					   }
					   if (trobat == FALSE) {
						   sinError("La variable no es del mateix tipus ENUMERAT que el valor que s'intenta assignar.");
					   }
				   } else {
					   // Posicio array enumerat := variable tipus enumerat
					   if (strcmp($3.atributs.continguts.info_variable.nom_tipus, symTipusReferenciat->nom) == 0) {
						   // OK
						   char arg1[STRING_LENGTH];
						   char arg2[STRING_LENGTH];
						   char arg3[STRING_LENGTH];
						   
						   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
							   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
						   } else {
							   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
						   }
						   
						   // Usem "constat" en lloc de "literal" pq la sintaxis, al identificar una variable (regla direct_name), assigna literal a FALSE sempre
						   //if ($3.atributs.continguts.info_variable.literal == TRUE) {
						   if ($3.atributs.continguts.info_variable.constant == TRUE) {
							   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
						   } else {
							   sprintf(arg3, "%s", $3.nom);
						   }
						   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
						   seguent_quadruple++;
					   } else {
						   sinError("La variable no es del mateix tipus ENUMERAT que el valor que s'intenta assignar.");
					   }
				   }
			   } else if ((symTipusReferenciat->atributs.tipus == RECORD_TYPE) || (symTipusReferenciat->atributs.tipus == ARRAY_TYPE)) {
				   // symTipusReferenciat->atributs.tipus == RECORD_TYPE o ARRAY_TYPE
				   
				   if ($3.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) {
					   
					   Simbol *symTipusDreta = busca_totes_taules(taulaSimbolActual, $3.atributs.continguts.info_variable.nom_tipus);
					   
					   if ((symTipusDreta->atributs.tipus == ARRAY_TYPE) || (symTipusDreta->atributs.tipus == RECORD_TYPE)) {
						   Simbol *symTipusDretaReferenciat = $3.atributs.continguts.info_variable.offset_simbol;
						   
						   if (symTipusDretaReferenciat == NULL) {
							   sinError("Aquest tipus d'assginament no esta permes.");
						   } if (strcmp(symTipusReferenciat->nom, symTipusDretaReferenciat->nom) == 0) {
							   // OK
							   // ASSIGN_FROM ASSIGN_TO
							   
							   // En els dos llocs hi ha un tipus derivat
							   char arg1[STRING_LENGTH];
							   char arg2[STRING_LENGTH];
							   char arg3[STRING_LENGTH];
					   
							   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
							   encuar(cua_actual_var_tmp, sym_tmp);
							   sym_tmp->atributs.tipus = VARIABLE;
							   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
							   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
							   
							   // Tipus de la variable temporal
							   if (symTipusDretaReferenciat->atributs.tipus == TYPE) {
								   if (symTipusDretaReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
									   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
								   sprintf(sym_tmp->atributs.continguts.info_variable.nom_tipus, "%s", symTipusDretaReferenciat->nom);
								   } else {
									   sym_tmp->atributs.continguts.info_variable.tipus = symTipusDretaReferenciat->atributs.continguts.info_type.tipus;
								   }
							   } else {
								   // ENUMERATED, ARRAY, RECORD
								   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
								   sprintf(sym_tmp->atributs.continguts.info_variable.nom_tipus, "%s", symTipusDretaReferenciat->nom);
							   }
			
							   if ($3.atributs.continguts.info_variable.offset_literal == TRUE) {
								   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.offset);
							   } else {
								   sprintf(arg3, "%s", $3.atributs.continguts.info_variable.nom_offset);
							   }
							   emite(codi, ASSIGN_FROM_TABLE, sym_tmp->nom, $3.nom, arg3);
							   seguent_quadruple++;
							   
							   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
								   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
							   } else {
								   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
							   }
							   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, sym_tmp->nom);
							   seguent_quadruple++;
						   } else {
							   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (1)");
						   }
					   } else {
						   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (2)");
					   }
				   } else {
					   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (3)");
				   }
			   } else if (symTipusReferenciat->atributs.tipus == TYPE) {
				   
				   if ($3.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) {
					   
					   Simbol *symTipusDreta = busca_totes_taules(taulaSimbolActual, $3.atributs.continguts.info_variable.nom_tipus);
					   
					   if ((symTipusDreta->atributs.tipus == ARRAY_TYPE) || (symTipusDreta->atributs.tipus == RECORD_TYPE)) {
						   Simbol *symTipusDretaReferenciat = $3.atributs.continguts.info_variable.offset_simbol;
						   
						   if (strcmp(symTipusReferenciat->nom, symTipusDretaReferenciat->nom) == 0) {
							   // OK
							   // ASSIGN_FROM ASSIGN_TO
							   
							   // En els dos llocs hi ha un tipus derivat
							   
							   char arg1[STRING_LENGTH];
							   char arg2[STRING_LENGTH];
							   char arg3[STRING_LENGTH];
					   
							   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
							   encuar(cua_actual_var_tmp, sym_tmp);
							   sym_tmp->atributs.tipus = VARIABLE;
							   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
							   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
							   
							   // Tipus de la variable temporal
							   if (symTipusDretaReferenciat->atributs.tipus == TYPE) {
								   if (symTipusDretaReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) {
									   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
								   sprintf(sym_tmp->atributs.continguts.info_variable.nom_tipus, "%s", symTipusDretaReferenciat->nom);
								   } else {
									   sym_tmp->atributs.continguts.info_variable.tipus = symTipusDretaReferenciat->atributs.continguts.info_type.tipus;
								   }
							   } else {
								   // ENUMERATED, ARRAY, RECORD
								   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_DERIVAT;
								   sprintf(sym_tmp->atributs.continguts.info_variable.nom_tipus, "%s", symTipusDretaReferenciat->nom);
							   }
			
							   if ($3.atributs.continguts.info_variable.offset_literal == TRUE) {
								   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.offset);
							   } else {
								   sprintf(arg3, "%s", $3.atributs.continguts.info_variable.nom_offset);
							   }
							   emite(codi, ASSIGN_FROM_TABLE, sym_tmp->nom, $3.nom, arg3);
							   seguent_quadruple++;
							   
							   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
								   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
							   } else {
								   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
							   }
							   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, sym_tmp->nom);
							   seguent_quadruple++;
						   } else {
							   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (10)");
						   }
					   } else {
						   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (11)");
					   }
						   
				   } else {
					   if (((symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT)
								|| (symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER))
						  && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)){
							// OK
							// ASSIGN_TO (TABLA(1) := INTEGER_DERIVAT)
							
							char arg1[STRING_LENGTH];
							char arg2[STRING_LENGTH];
							char arg3[STRING_LENGTH];
							   
							if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
							   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
							} else {
							   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
							}
							emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, $3.nom);
							seguent_quadruple++;
					   }
				   }
				   
			   } else {
				   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (4)");
			   }
		   // sino... llavors es ENUMERATED_TYPE
		   } else {
			   if (strcmp($1.atributs.continguts.info_variable.nom_tipus, $3.atributs.continguts.info_variable.nom_tipus) != 0) {
				   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (5)");
			   }
			   // enumerat
			   
			   char arg2[STRING_LENGTH];
			   // Usem "constat" en lloc de "literal" pq la sintaxis, al identificar una variable (regla direct_name), assigna literal a FALSE sempre
			   //if ($3.atributs.continguts.info_variable.literal == TRUE) {
			   if ($3.atributs.continguts.info_variable.constant == TRUE) {
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   } else {
				   sprintf(arg2, "%s", $3.nom);
			   }
			   
			   emite(codi, ASSIGN, $1.nom, arg2, NULL);
			   seguent_quadruple++;
		   }
	   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
		   // Integer derivat
		   if (strcmp($1.atributs.continguts.info_variable.nom_tipus, $3.atributs.continguts.info_variable.nom_tipus) != 0) {
				sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (6)");
		   }
		   
		   // Codi 3 adreces
		   if ($3.atributs.continguts.info_variable.literal == TRUE) {
			   char arg2[STRING_LENGTH];
			   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   emite(codi, ASSIGN, $1.nom, arg2, NULL);
			   seguent_quadruple++;
		   } else {
			   if (strcmp($1.nom, $3.nom) != 0) {
				   // Tot menys assignaments estupids com 'variable := variable'
				   emite(codi, ASSIGN, $1.nom, $3.nom, NULL);
				   seguent_quadruple++;
			   }
		   }
		   
	   } else if (($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
		   // Integer derivat := Integer
		   
		   // Comprovacio limits de rang pels tipus integer derivats
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1.atributs.continguts.info_variable.nom_tipus);
		   if (symTipus->atributs.tipus == TYPE) {
			   if (($3.atributs.continguts.info_variable.valor.integer < (int)symTipus->atributs.continguts.info_type.rang.min)
					|| ($3.atributs.continguts.info_variable.valor.integer > (int)symTipus->atributs.continguts.info_type.rang.max)) {
							sinError("Valor fora de rang, ha d'estar entre %i i %i.", (int)symTipus->atributs.continguts.info_type.rang.min, (int)symTipus->atributs.continguts.info_type.rang.max);
				}
		   } else {
			   sinError("No es pot assignar un valor a un element d'un enumerat.");
		   }
		   
		   // Codi 3 adreces
		   if ($3.atributs.continguts.info_variable.literal == TRUE) {
			   char arg2[STRING_LENGTH];
			   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
			   emite(codi, ASSIGN, $1.nom, arg2, NULL);
			   seguent_quadruple++;
		   } else {
			   if (strcmp($1.nom, $3.nom) != 0) {
				   // Tot menys assignaments estupids com 'variable := variable'
				   emite(codi, ASSIGN, $1.nom, $3.nom, NULL);
				   seguent_quadruple++;
			   }
		   }
	   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) {
		   // Assignacio a una posicio d'un ARRAY
		   
		   // Comprovacio limits de rang pels tipus integer derivats
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1.atributs.continguts.info_variable.nom_tipus);
		   if ((symTipus->atributs.tipus == ARRAY_TYPE) || (symTipus->atributs.tipus == RECORD_TYPE)){

			   Simbol *symTipusReferenciat = $1.atributs.continguts.info_variable.offset_simbol;
			   
			   
			   // Array d'integer derivats (amb operand dret tambe integer_derivat) ja ha sigut tractat en un cas anterior (ha de ser un cas d'error)
			   /*if ((symTipus->atributs.continguts.info_array.tipus == TIPUS_INTEGER_DERIVAT) && ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
				   // Array d'integer derivat := integer derivat
				   if (strcmp(symTipus->atributs.continguts.info_array.nom_tipus, $3.atributs.continguts.info_variable.nom_tipus) != 0) {
					   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
				   }
			   } */
			   
			   // Array d'enumerat, records i array ja ha sigut tractat en un cas anterior (ha de ser un cas d'error)
			   /*if (symTipusReferenciat->atributs.tipus == ENUMERATED_TYPE) {
			   } else if ((symTipusReferenciat->atributs.tipus == RECORD_TYPE) || (symTipusReferenciat->atributs.tipus == ARRAY_TYPE)) {
			   }*/
			   
			   if (symTipusReferenciat->atributs.tipus != TYPE) {
				   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (7)");
			   }
			   
			   if ((symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) 
				   		&& ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
				   // Array d'integer derivat := Integer
				   
				   // Comprovacio limits de rang pels tipus integer derivats
				   if (($3.atributs.continguts.info_variable.valor.integer < (int)symTipusReferenciat->atributs.continguts.info_type.rang.min)
						|| ($3.atributs.continguts.info_variable.valor.integer > (int)symTipusReferenciat->atributs.continguts.info_type.rang.max)) {
								sinError("Valor fora de rang, ha d'estar entre %i i %i.", (int)symTipusReferenciat->atributs.continguts.info_type.rang.min, (int)symTipusReferenciat->atributs.continguts.info_type.rang.max);
				   }
				   
				   // ASSIGN_TO es fa despres del conjunt d'IFs
				   
			   } else if (symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_DERIVAT) {
				   // ARRAY, RECORD
				   //sinError("No poden haver arrays d'arrays ni arrays de records");
				   sinError("No hauria d'entrar aqui.");
				   
				   // TODO: Comprovar si s'esta assignant el mateix tipus derivat
				   /*if ($3.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) {
					   if (strcmp(symTipus->atributs.continguts.info_array.nom_tipus, $3.atributs.continguts.info_variable.nom_tipus) != 0) {
						   //sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
					   }
				   } else {
					   
					   // O comprovar si s'esta assignant un tipus a un offset concret.
					   if ($1.atributs.continguts.info_variable.offset_simbol != NULL) {
						   
						   Simbol *offset_simbol = $1.atributs.continguts.info_variable.offset_simbol;
						   // TODO: comprovar tipus basics
						   if (offset_simbol->atributs.tipus == TYPE) {
							   if (offset_simbol->atributs.continguts.info_type.tipus != $3.atributs.continguts.info_variable.tipus) {
								   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
							   }
						   } else if ((offset_simbol->atributs.tipus == ARRAY_TYPE) || (offset_simbol->atributs.tipus == ENUMERATED_TYPE) || (offset_simbol->atributs.tipus == RECORD_TYPE)) {
							   if (strcmp(offset_simbol->nom, $3.atributs.continguts.info_variable.nom_tipus) != 0) {
								   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
							   }
						   }
					   } else {
						   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
					   }
				   }*/
			   } else if (symTipusReferenciat->atributs.continguts.info_type.tipus == $3.atributs.continguts.info_variable.tipus) {
				   // OK: Integer, Boolean, Float, String, Character
				   // ASSIGN_TO es fa despres del conjunt d'IFs

			   } else {
				   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (8)");
			   }
			   
			   
			   // Codi 3 adreces
			   
			   if ($3.atributs.continguts.info_variable.literal == TRUE) {
				   char arg2[STRING_LENGTH];
				   char arg3[STRING_LENGTH];

				   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
					   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
				   } else {
					   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
				   }

				   
				   if (($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($3.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg3, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
					   int i;
					   
					   // A la primera posicio va el tamany
					   sprintf(arg3, "%i", caracters - 1);
					   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
					   seguent_quadruple++;
					   for (i = 1; i < caracters; i++) {
						   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
							   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset + i);
						   } else {
							   sprintf(arg2, "%s + i", $1.atributs.continguts.info_variable.nom_offset, i);
						   }
						   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
						   
						   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
						   seguent_quadruple++;
					   }
				   }
				   
				   if ($3.atributs.continguts.info_variable.tipus != TIPUS_STRING) {
					   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   char arg2[STRING_LENGTH];
				   
				   if ($1.atributs.continguts.info_variable.offset_literal == TRUE) {
					   sprintf(arg2, "%i", $1.atributs.continguts.info_variable.offset);
				   } else {
					   sprintf(arg2, "%s", $1.atributs.continguts.info_variable.nom_offset);
				   }

				   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, $3.nom);
				   seguent_quadruple++;
			   }
		   } else {
			   sinError("No s'espera altra cosa que un ARRAY_TYPE o RECORD_TYPE aqui.");
		   }
		    
	   } else if ($1.atributs.continguts.info_variable.tipus == $3.atributs.continguts.info_variable.tipus) {
		   // OK: Integer, Boolean, Float, String, Character

		   // Codi 3 adreces
		   if ($3.atributs.continguts.info_variable.literal == TRUE) {
			   if ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.integer);
				   emite(codi, ASSIGN, $1.nom, arg2, NULL);
				   seguent_quadruple++;
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%g", $3.atributs.continguts.info_variable.valor.vfloat);
				   emite(codi, ASSIGN, $1.nom, arg2, NULL);
				   seguent_quadruple++;
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "'%c'", $3.atributs.continguts.info_variable.valor.character);
				   emite(codi, ASSIGN, $1.nom, arg2, NULL);
				   seguent_quadruple++;
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
				   char arg2[STRING_LENGTH];
				   sprintf(arg2, "%i", $3.atributs.continguts.info_variable.valor.boolean);
				   emite(codi, ASSIGN, $1.nom, arg2, NULL);
				   seguent_quadruple++;
			   } else if ($1.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
				   int caracters = strlen($3.atributs.continguts.info_variable.valor.string);
				   int i;
				   char arg2[STRING_LENGTH];
				   char arg3[STRING_LENGTH];
				   
				   // A la primera posicio va el tamany
				   sprintf(arg2, "0");
				   sprintf(arg3, "%i", caracters - 1);
				   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
				   seguent_quadruple++;
				   for (i = 1; i < caracters; i++) {
					   
					   sprintf(arg2, "%i", i);
					   sprintf(arg3, "'%c'", $3.atributs.continguts.info_variable.valor.string[i]);
					   emite(codi, ASSIGN_TO_TABLE, $1.nom, arg2, arg3);
					   seguent_quadruple++;
				   }
			   } else {
				   sinError("Tipus no soportat.");
			   }
		   } else {
			   if (strcmp($1.nom, $3.nom) != 0) {
				   // Tot menys assignaments estupids com 'variable := variable'
				   emite(codi, ASSIGN, $1.nom, $3.nom, NULL);
				   seguent_quadruple++;
			   }
		   }
	   } else if ($3.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT) {
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $3.atributs.continguts.info_variable.nom_tipus);
		   
		   if ((symTipus->atributs.tipus == ARRAY_TYPE) || (symTipus->atributs.tipus == RECORD_TYPE)) {

			   	Simbol *symTipusReferenciat = $3.atributs.continguts.info_variable.offset_simbol;
			   
			   	if ((symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER) 
				   		&& ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					// ASSIGN_TO es fa despres del conjunt d'IFs
				} else if ((symTipusReferenciat->atributs.continguts.info_type.tipus == TIPUS_INTEGER_DERIVAT) 
				   		&& ($1.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
				   // ASSIGN_TO es fa despres del conjunt d'IFs
				} else if (symTipusReferenciat->atributs.continguts.info_type.tipus == $1.atributs.continguts.info_variable.tipus) {
					// OK: Integer, Boolean, Float, String, Character
				    // ASSIGN_TO es fa despres del conjunt d'IFs
				} else {
					sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (12)");
				}
				
				// Codi 3 adreces
				char arg3[STRING_LENGTH];
			   
				if ($3.atributs.continguts.info_variable.offset_literal == TRUE) {
				   sprintf(arg3, "%i", $3.atributs.continguts.info_variable.offset);
				} else {
				   sprintf(arg3, "%s", $3.atributs.continguts.info_variable.nom_offset);
				}

				emite(codi, ASSIGN_FROM_TABLE, $1.nom, $3.nom, arg3);
				seguent_quadruple++;
		   } else {
			   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (13)");
		   }
	   } else {
		   sinError("La variable no es del mateix tipus que el valor que s'intenta assignar. (9)");
	   }

   }
;

procedure_call_statement:
   PR_ID_PROCEDURE {
	    Boolean trobat = FALSE;
		Info_Procedure *info_procedure;
		
		info_procedure = &($1->atributs.continguts.info_procedure);
		if (info_procedure->num_parametres == 0) {
			trobat = TRUE;
		}
		
		if (trobat == FALSE) {
			if ($1->atributs.continguts.info_procedure.sobrecarregues == NULL) {
				sinError("El procediment '%s' requereix '%i' parametre(s).", $1->nom, $1->atributs.continguts.info_procedure.num_parametres);
			} else {
				// Recorrer sobrecarregues
				reiniciar_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				info_procedure = (Info_Procedure *)obtenir_element_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				while(info_procedure != NULL) {
					// Comprovar parametres
					if (info_procedure->num_parametres == 0) {
						trobat = TRUE;
						break;
					}
					
					seguent_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
					info_procedure = (Info_Procedure *)obtenir_element_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				}
			}
		}
		
		if (trobat) {
			// NOTE: Es pot cridar un procediment amb cos no definit si la signatura si ha sigut definida
			// Al final del compilador ja es comprova que tots tenen cos definit.
			/*if (info_procedure->complet == FALSE) {
				sinError("El procediment '%s' no te encara el cos definit.", $1->nom);
			}*/
			
			// TODO: Quin simbol retornem?
			/*$$.atributs.tipus = VARIABLE;
			$$.atributs.continguts.info_variable.tipus = info_procedure->tipus_retorn;
			$$.atributs.continguts.info_variable.literal = FALSE;
			$$.atributs.continguts.info_variable.inicialitzat = FALSE;
		
			if (($1->atributs.continguts.info_procedure.tipus_retorn == TIPUS_DERIVAT)
				|| ($1->atributs.continguts.info_procedure.tipus_retorn == TIPUS_INTEGER_DERIVAT)) {
				memcpy($$.atributs.continguts.info_variable.nom_tipus, info_procedure->nom_tipus_retorn, STRING_LENGTH);
			}*/
			
			// Codi 3 adreces
			char arg1[STRING_LENGTH];
			char arg2[STRING_LENGTH];
			sprintf(arg1, "%s", info_procedure->nom_unic);
			sprintf(arg2, "0");
			emite(codi, CALL_PROC, arg1, arg2, NULL);
			seguent_quadruple++;
			
		} else {
			sinError("El procediment '%s' no te cap sobrecarrega que permeti '0' parametres.", $1->nom);
		}
   } // PROCEDURE
   
   | PR_ID_PROCEDURE actual_parameter_part ';' {
	    Boolean trobat = FALSE;
		Info_Procedure *info_procedure;
		
		info_procedure = &($1->atributs.continguts.info_procedure);
		// Mateix nombre de parametres?

		if (info_procedure->num_parametres == contador) {
			// Comprovar tipus parametres
			if (crida_parametres_diferents(cua_tmp, info_procedure->parametres, info_procedure->taula) == FALSE) {
				// Coincideixen el tipus dels parametres amb la definicio
				trobat = TRUE;
			}
		}
		
		if (trobat == FALSE) {
			if ($1->atributs.continguts.info_procedure.sobrecarregues == NULL) {
				sinError("El procediment '%s' requereix '%i' parametre(s) i no s'han trobat cap tipus coincident.", $1->nom, $1->atributs.continguts.info_procedure.num_parametres);
			} else {
				// Recorrer sobrecarregues
				reiniciar_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				info_procedure = (Info_Procedure *)obtenir_element_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				while(info_procedure != NULL) {
					// Mateix nombre de parametres?
					if (info_procedure->num_parametres == contador) {
						// Comprovar tipus parametres
						if (crida_parametres_diferents(cua_tmp, info_procedure->parametres, info_procedure->taula) == FALSE) {
							// Coincideixen el tipus dels parametres amb la definicio
							trobat = TRUE;
							break;
						}
					}
					
					seguent_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
					info_procedure = (Info_Procedure *)obtenir_element_PDI($1->atributs.continguts.info_procedure.sobrecarregues);
				}
			}
		}
		
		if (trobat) {
			// NOTE: Es pot cridar un procediment amb cos no definit si la signatura si ha sigut definida
			// Al final del compilador ja es comprova que tots tenen cos definit.
			/*if (info_procedure->complet == FALSE) {
				sinError("El procediment '%s' no te encara el cos definit.", $1->nom);
			}*/
			
			
			
			// TODO: Quin simbol retornem?
			/*$$.atributs.tipus = VARIABLE;
			$$.atributs.continguts.info_variable.tipus = info_procedure->tipus_retorn;
			$$.atributs.continguts.info_variable.literal = FALSE;
			$$.atributs.continguts.info_variable.inicialitzat = FALSE;
		
			if (($1->atributs.continguts.info_procedure.tipus_retorn == TIPUS_DERIVAT)
				|| ($1->atributs.continguts.info_procedure.tipus_retorn == TIPUS_INTEGER_DERIVAT)) {
				memcpy($$.atributs.continguts.info_variable.nom_tipus, info_procedure->nom_tipus_retorn, STRING_LENGTH);
			}*/
			
			char arg1[STRING_LENGTH];
			char arg2[STRING_LENGTH];
			
			// PARAMS
			reiniciar_PDI(cua_tmp);
			Simbol *sym_param = (Simbol *)obtenir_element_PDI(cua_tmp);
			Simbol *sym_param_especial = sym_param; // Usat per la funcio especial PUT
			while (sym_param!=NULL) {
				if (sym_param->atributs.continguts.info_variable.literal == TRUE) {
					if ((sym_param->atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || (sym_param->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
					   sprintf(arg1, "%i", sym_param->atributs.continguts.info_variable.valor.integer);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					   sprintf(arg1, "%g", sym_param->atributs.continguts.info_variable.valor.vfloat);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					   sprintf(arg1, "'%c'", sym_param->atributs.continguts.info_variable.valor.character);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_BOOLEAN) {
					   sprintf(arg1, "%i", sym_param->atributs.continguts.info_variable.valor.boolean);
				   } else if (sym_param->atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					   // Generem un temporal per guardar l'string i fem el PARAM amb aquest
					   int caracters = strlen(sym_param->atributs.continguts.info_variable.valor.string);
					   int i;
					   
					   char arg2[STRING_LENGTH];
					   char arg3[STRING_LENGTH];
				   
					   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
					   encuar(cua_actual_var_tmp, sym_tmp);
					   sym_tmp->atributs.tipus = VARIABLE;
					   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_STRING;
					   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
					   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
					   
					   // A la primera posicio va el tamany
					   sprintf(arg2, "0");
					   sprintf(arg3, "%i", caracters - 1);
					   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
					   seguent_quadruple++;
					   for (i = 1; i < caracters; i++) {
						   sprintf(arg2, "%i", i);
						   sprintf(arg3, "'%c'", sym_param->atributs.continguts.info_variable.valor.string[i]);
						   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
						   seguent_quadruple++;
					   }
					   
					   sprintf(arg1, "%s", sym_tmp->nom);
				   }
				} else {
					sprintf(arg1, "%s", sym_param->nom);
				}
				
				emite(codi, PARAM, arg1, NULL, NULL);
				seguent_quadruple++;
				
				seguent_PDI(cua_tmp);
				sym_param_especial = sym_param;
				sym_param = (Simbol *)obtenir_element_PDI(cua_tmp);
			}
			
			if (strcmp($1->nom, "put") == 0) {
				
				if (sym_param_especial->atributs.continguts.info_variable.tipus == TIPUS_INTEGER) {
					emite(codi, PUTI, NULL, NULL, NULL);
					seguent_quadruple++;
				} else if (sym_param_especial->atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
					emite(codi, PUTF, NULL, NULL, NULL);
					seguent_quadruple++;
				} else if (sym_param_especial->atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
					emite(codi, PUTC, NULL, NULL, NULL);
					seguent_quadruple++;
				} else if (sym_param_especial->atributs.continguts.info_variable.tipus == TIPUS_STRING) {
					// Crida normal
					sprintf(arg1, "put");
					sprintf(arg2, "%i", info_procedure->num_parametres);
					emite(codi, CALL_PROC, arg1, arg2, NULL);
					seguent_quadruple++;
				} else {
					sinError("No s'hauria de sobrecarregar PUT, apart de les sobrecarregues ja establertes.");
				}
			}  else {
				// Crida normal
				sprintf(arg1, "%s", info_procedure->nom_unic);
				sprintf(arg2, "%i", info_procedure->num_parametres);
				emite(codi, CALL_PROC, arg1, arg2, NULL);
				seguent_quadruple++;
			}
			
			cua_tmp = NULL;
		} else {
			sinError("El procediment '%s' no te cap sobrecarrega que permeti '%i' parametre(s) amb aquests tipus.", $1->nom, contador);
		}
   }
;

actual_parameter_part:
  '(' {
		cua_tmp = (Cua *)crear_cua();
		contador = 0;
   }
   actual_parameter_part_element ')' {}
;

return_statement:
   PR_RETURN ';' {
	   Simbol *symRutina = busca_totes_taules(taulaSimbolActual, taulaSimbolActual->nom);
	   if (symRutina->atributs.tipus == FUNCTION) {
		   sinError("La funcio ha de retornar un valor.");
	   }
	   
	   emite(codi, RETURN_PROC, NULL, NULL, NULL);
	   seguent_quadruple++;
   }
   | PR_RETURN expression ';' {
	   Simbol *symRutina = busca_totes_taules(taulaSimbolActual, taulaSimbolActual->nom);
	   if (symRutina->atributs.tipus != FUNCTION) {
		   sinError("No es possible ficar un RETURN amb valor a un procediment.");
	   }
	   
	   // S'ha especificat un return al cos de la funcio
	   // No es pot saber en quina sobrecarrega estem, per tant millor utilitzar una variable global
	   // que sempre apunta al info_function actual.
	   //symRutina->atributs.continguts.info_function.return_indicat = TRUE;
	   infoFunctionActual->return_indicat = TRUE;
	   
	   if ((infoFunctionActual->tipus_retorn == TIPUS_DERIVAT) && ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
		   // Enumerat := Element enumerat
			if (strcmp(infoFunctionActual->nom_tipus_retorn, $2.atributs.continguts.info_variable.nom_tipus) != 0) {
				sinError("La variable no es del mateix tipus que el valor que s'intenta assignar.");
		   }
	   } else if ((infoFunctionActual->tipus_retorn == TIPUS_DERIVAT) && ($2.atributs.continguts.info_variable.tipus == TIPUS_DERIVAT)) {
		   // Array, enumerat, record
		   if (strcmp(infoFunctionActual->nom_tipus_retorn, $2.atributs.continguts.info_variable.nom_tipus) != 0) {
				sinError("Error al valor de retorn, el tipus no coincideix.");
		   }
	   } else if ((infoFunctionActual->tipus_retorn == TIPUS_INTEGER_DERIVAT) && ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
		   // Integer derivat
		   if (strcmp(infoFunctionActual->nom_tipus_retorn, $2.atributs.continguts.info_variable.nom_tipus) != 0) {
				sinError("Error al valor de retorn, el tipus no coincideix.");
		   }
	   } else if ((infoFunctionActual->tipus_retorn == TIPUS_INTEGER_DERIVAT) && ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER)) {
		   // Integer derivat := Integer
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, infoFunctionActual->nom_tipus_retorn);
		   if (symTipus->atributs.tipus == TYPE) {
			   if ($2.atributs.continguts.info_variable.literal == TRUE) {
				   if (($2.atributs.continguts.info_variable.valor.integer < (int)symTipus->atributs.continguts.info_type.rang.min)
						|| ($2.atributs.continguts.info_variable.valor.integer > (int)symTipus->atributs.continguts.info_type.rang.max)) {
								sinError("Valor de retorn fora de rang, ha d'estar entre %i i %i.", (int)symTipus->atributs.continguts.info_type.rang.min, (int)symTipus->atributs.continguts.info_type.rang.max);
					}
			   } else {
				   printf("Return amb variable INTEGER, pero s'espera INTEGER DERIVAT. Perill d'overflow a la linia %i columna %i.\n", lineNumber, colNumber);
			   }
		   } else {
			   sinError("Error al valor de retorn, el tipus no coincideix.");
		   }
	   } else if (infoFunctionActual->tipus_retorn == $2.atributs.continguts.info_variable.tipus) {
		   // OK: Integer, Boolean, Float, String, Character
	   } else {
		   sinError("Error al valor de retorn, el tipus no coincideix.");
	   }
	   
	   
	   char arg1[STRING_LENGTH];
	   
	   // Passem a string els parametres de la comparacio
	   if ($2.atributs.continguts.info_variable.literal == TRUE) {
		   if (($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER) || ($2.atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
			   sprintf(arg1, "%i", $2.atributs.continguts.info_variable.valor.integer);
		   } else if ($2.atributs.continguts.info_variable.tipus == TIPUS_FLOAT) {
			   sprintf(arg1, "%g", $2.atributs.continguts.info_variable.valor.vfloat);
		   } else if ($2.atributs.continguts.info_variable.tipus == TIPUS_CHARACTER) {
			   sprintf(arg1, "'%c'", $2.atributs.continguts.info_variable.valor.character);
		   } else if ($2.atributs.continguts.info_variable.tipus == TIPUS_STRING) {
			   // Generem un temporal per guardar l'string i retornem aquest
			   int caracters = strlen($2.atributs.continguts.info_variable.valor.string);
			   int i;
			   
			   char arg2[STRING_LENGTH];
			   char arg3[STRING_LENGTH];
		   
			   Simbol *sym_tmp = crear_i_inserir_en_taula(taulaSimbolActual, nou_temp());
			   encuar(cua_actual_var_tmp, sym_tmp);
			   sym_tmp->atributs.tipus = VARIABLE;
			   sym_tmp->atributs.continguts.info_variable.tipus = TIPUS_STRING;
			   sym_tmp->atributs.continguts.info_variable.literal = FALSE;
			   sym_tmp->atributs.continguts.info_variable.constant = FALSE;
			   
			   // A la primera posicio va el tamany
			   sprintf(arg2, "0");
			   sprintf(arg3, "%i", caracters - 1);
			   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
			   seguent_quadruple++;
			   for (i = 1; i < caracters; i++) {
				   sprintf(arg2, "%i", i);
				   sprintf(arg3, "'%c'", $2.atributs.continguts.info_variable.valor.string[i]);
				   emite(codi, ASSIGN_TO_TABLE, sym_tmp->nom, arg2, arg3);
				   seguent_quadruple++;
			   }
			   
			   sprintf(arg1, "%s", sym_tmp->nom);
		   }
	   } else {
		   sprintf(arg1, "%s", $2.nom);
	   }
	   
	   emite(codi, RETURN_FUNC, arg1, NULL, NULL);
	   seguent_quadruple++;
   }
;

compound_statement:
   if_statement {
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   | loop_statement {
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
;

if_statement:
   PR_IF condition PR_THEN mark sequence_of_statements PR_END PR_IF ';' {
	   sinLog(" -> if_statement -> PR_IF condition PR_THEN sequence_of_statements PR_END PR_IF ';'");
	   completa_goto(codi, $2.llista_cert, $4);
	   $$.llista_seguents = (Cua*)fusionar_cues($2.llista_fals, $5.llista_seguents);
   }
   | PR_IF condition PR_THEN mark sequence_of_statements n_mark PR_ELSE mark sequence_of_statements PR_END PR_IF ';' {
	   sinLog(" -> if_statement -> PR_IF condition PR_THEN sequence_of_statements PR_ELSE sequence_of_statements PR_END PR_IF ';'");
	   completa_goto(codi, $2.llista_cert, $4);
	   completa_goto(codi, $2.llista_fals, $8);
	   $$.llista_seguents = (Cua*)fusionar_cues($5.llista_seguents, (Cua*)fusionar_cues($6.llista_seguents, $9.llista_seguents));
   }
;

condition:
   expression {
	   sinLog(" -> condition -> expression");
	   
	   // El tipus ha de ser boolea
	   if ($1.atributs.tipus != VARIABLE) {
		   sinError("Condicio no valida.");
	   }
	   if ($1.atributs.continguts.info_variable.tipus != TIPUS_BOOLEAN) {
		   sinError("Condicio no valida.");
	   }
	   
	   // Codi 3 adreces
	   // Quan es fica un literal "true", les sentencies generades es guarden al seu simbol
	   // per poder diferenciar d'una condicio a una inicialitzacio/crida
	   if ($1.sentencies != NULL) {
		   codi = (Cua*)fusionar_cues(codi, $1.sentencies);
		   seguent_quadruple++; // $?.sentencies nomes conte 1 sentencia
	   }
	   
	   memcpy(&$$, &$1, sizeof(struct Simbol));
   }
   /* boolean_expression {}*/
;

loop_statement:
   mark iteration_scheme PR_LOOP mark sequence_of_statements PR_END PR_LOOP ';' {
	   sinLog(" -> loop_statement -> mark iteration_scheme PR_LOOP mark sequence_of_statements PR_END PR_LOOP ';'");
	   completa_goto(codi, $5.llista_seguents, $1);
	   completa_goto(codi, $2.llista_cert, $4);
	   $$.llista_seguents = $2.llista_fals;
	   
	   char arg1[STRING_LENGTH];
	   sprintf(arg1, "%i", $1);
	   emite(codi, GOTO, arg1, NULL, NULL);
	   seguent_quadruple++;
   }
;

iteration_scheme:
   PR_WHILE condition {
	   sinLog(" -> iteration_scheme -> PR_WHILE condition");
	   memcpy(&($$), &($2), sizeof(struct Simbol));
   }
   | PR_FOR loop_parameter_specification {
	   sinLog(" -> iteration_scheme -> PR_FOR loop_parameter_specification");
	   memcpy(&($$), &($2), sizeof(struct Simbol));
   }
;

loop_parameter_specification:
   PR_ID_VARIABLE PR_IN discrete_subtype_definition {
	   // TODO: Comprovar tipus de la variable amb el range
	   
	   if ($3.atributs.continguts.info_array.rang.tipus != TIPUS_INTEGER) {
		   sinError("No es permeten FOR amb valors diferents d'integer.");
	   }
	   
	   if ($1->atributs.continguts.info_variable.tipus == TIPUS_INTEGER) {
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, "integer");
		   
		   if (($3.atributs.continguts.info_array.rang.min < (int)symTipus->atributs.continguts.info_type.rang.min)
				   || ($3.atributs.continguts.info_array.rang.max > (int)symTipus->atributs.continguts.info_type.rang.max)) {
		   		sinError("La variable no pot contenir valors definits al rang del FOR.");
		   }
	   } else if ($1->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT) {
		   Simbol *symTipus = busca_totes_taules(taulaSimbolActual, $1->atributs.continguts.info_variable.nom_tipus);
		   
		   if (symTipus->atributs.tipus == TYPE) {
			   if (($3.atributs.continguts.info_array.rang.min < (int)symTipus->atributs.continguts.info_type.rang.min)
					   || ($3.atributs.continguts.info_array.rang.max > (int)symTipus->atributs.continguts.info_type.rang.max)) {
					sinError("La variable no pot contenir valors definits al rang del FOR.");
			   }
		   } else {
			   // Posible element d'enumerador
			   sinError("La variable d'un FOR ha de ser de tipus 'integer' o derivat.");
		   }
	   } else {
		   sinError("La variable d'un FOR ha de ser de tipus 'integer' o derivat.");
	   }
	   
	   if ($1->atributs.continguts.info_variable.constant == TRUE) {
		   sinError("No es pot utilitzar una variable definida com a constant en un FOR.");
	   }
	   
	   // Codi 3 adreces
	   // NOTE: Traduim el FOR com si fos un WHILE (variable > rang.min and variable < rang.max) { variable++; ... }
	   char arg1[STRING_LENGTH];
	   char arg2[STRING_LENGTH];
	   char arg3[STRING_LENGTH];
	   // If variable > rang_min goto (next if)
	   sprintf(arg1, "%s", $1->nom);
	   sprintf(arg2, "%i", (int)$3.atributs.continguts.info_array.rang.min);
	   sprintf(arg3, "%i", seguent_quadruple+2);
	   emite(codi, IF_GTI, arg1, arg2, arg3);
	   seguent_quadruple++;
	   
	   // If NOT(variable > rang_min) GOTO (despres sentencies, encara desconegut)
	   $$.llista_fals = (Cua*)crear_cua();
	   int *pos = (int*)malloc(sizeof(int));
	   *pos = seguent_quadruple;
	   encuar($$.llista_fals, pos);
	   
	   emite(codi, GOTO, NULL, NULL, NULL);
	   seguent_quadruple++;
	   
	   // If variable < rang_max goto (sentencies)
	   sprintf(arg1, "%s", $1->nom);
	   sprintf(arg2, "%i", (int)$3.atributs.continguts.info_array.rang.max);
	   sprintf(arg3, "%i", seguent_quadruple+2);
	   emite(codi, IF_LTI, arg1, arg2, arg3);
	   seguent_quadruple++;
	   
	   // If NOT(variable < rang_max) GOTO (despres sentencies, encara desconegut)
	   pos = (int*)malloc(sizeof(int));
	   *pos = seguent_quadruple;
	   encuar($$.llista_fals, pos);
	   
	   emite(codi, GOTO, NULL, NULL, NULL);
	   seguent_quadruple++;
	   
	   // variable ++ (increment)
	   sprintf(arg1, "%s", $1->nom);
	   sprintf(arg2, "%s", $1->nom);
	   sprintf(arg3, "1");
	   emite(codi, ADDI, arg1, arg2, arg3);
	   seguent_quadruple++;
   }
;


program:
   program_element PR_IS declarative_part PR_BEGIN sequence_of_statements_after_begin PR_END ';' {
	   sinLog(" -> program -> program_element PR_IS declarative_part PR_BEGIN sequence_of_statements PR_END ';'");
	   // NOTE: Comentar a l'informe que es comprova que existeixin tots els body de funcions i procediments
	   taulaSimbolActual = taulaSimbolActual->anterior;
	   
	   
	   
	   Simbol *sym = obtenir_simbol(taulaSimbolActual, $1.nom);
	   emite(codi, HALT, NULL, NULL, NULL);
	   seguent_quadruple++;
	   sym->sentencies = codi;  // Codi 3 adreces generat
	   codi = NULL;
	   
	   desenpilar(pila_cues_var_locals);
	   cua_actual_var_locals = (Cua*)cim(pila_cues_var_locals);
	   desenpilar(pila_cues_var_tmp);
	   cua_actual_var_tmp = (Cua*)cim(pila_cues_var_tmp);
	   
	   // Comprovar que tota funcio/procediment tingui body
	   comprovacio_complets(taulaSimbolActual);
   }
   | program_element PR_IS declarative_part PR_BEGIN sequence_of_statements_after_begin PR_END designator ';' {
	   sinLog(" -> program -> program_element PR_IS declarative_part PR_BEGIN sequence_of_statements PR_END designator ';'");
	   taulaSimbolActual = taulaSimbolActual->anterior;
	   
	   Simbol *sym = obtenir_simbol(taulaSimbolActual, $1.nom);
	   emite(codi, HALT, NULL, NULL, NULL);
	   seguent_quadruple++;
	   sym->sentencies = codi;  // Codi 3 adreces generat
	   codi = NULL;
	   
	   if (strcmp($1.nom, $7.nom) != 0) {
		   sinError("Designator '%s' no es correspon amb el nom de la funcio/procediment %s.", $7.nom, $1.nom);
	   }
	   
	   desenpilar(pila_cues_var_locals);
	   cua_actual_var_locals = (Cua*)cim(pila_cues_var_locals);
	   desenpilar(pila_cues_var_tmp);
	   cua_actual_var_tmp = (Cua*)cim(pila_cues_var_tmp);
	   
	   // Comprovar que tota funcio/procediment tingui body
	   comprovacio_complets(taulaSimbolActual);
   }
;

program_element:
	PR_PROCEDURE defining_program_unit_name {
		sinLog(" -> program_element -> PR_PROCEDURE defining_program_unit_name");
		contador_noms_unics = 0;
		
		Simbol *sym = crear_i_inserir_en_taula(taulaSimbolActual, $2);
		sym->atributs.tipus = PROCEDURE;
		sprintf(sym->atributs.continguts.info_procedure.nom_unic, "%s_%i", $2, contador_noms_unics++);
		sym->atributs.continguts.info_procedure.id = 0;
		sym->atributs.continguts.info_procedure.num_sobrecarregues = 0;
		sym->atributs.continguts.info_procedure.complet = TRUE;
		sym->atributs.continguts.info_procedure.num_parametres = 0;
		sym->atributs.continguts.info_procedure.parametres = NULL;
		sym->atributs.continguts.info_procedure.sobrecarregues = NULL;
		Taula_Simbols *ts = (Taula_Simbols *)crear_taula_simbols($2, taulaSimbolActual);
		sym->atributs.continguts.info_procedure.taula = ts;
		taulaSimbolActual = ts;
		sym->atributs.continguts.info_procedure.variables_locals = (Cua*)crear_cua();
		sym->atributs.continguts.info_procedure.variables_temporals = (Cua*)crear_cua();
		
		pila_cues_var_locals = (Pila*)crear_pila();
		pila_cues_var_tmp = (Pila*)crear_pila();
		enpilar(pila_cues_var_locals, sym->atributs.continguts.info_procedure.variables_locals);
		cua_actual_var_locals = sym->atributs.continguts.info_procedure.variables_locals;
		enpilar(pila_cues_var_tmp, sym->atributs.continguts.info_procedure.variables_temporals);
		cua_actual_var_tmp = sym->atributs.continguts.info_procedure.variables_temporals;
		
		assignar_programa(taulaSimbolActual,ts);
		
		infoProcedureActual = &(sym->atributs.continguts.info_procedure);
		codi = NULL;
		
		memcpy(&$$, sym, sizeof(struct Simbol));
	}
;

%%

// Cerca un simbol a una cua de simbols, si no existeix retorna NULL, sino retorna el simbol
Simbol *cercar_simbol_a_cua(Cua *cua, char *nom) {
	reiniciar_PDI(cua);
	Simbol *element = (Simbol *)obtenir_element_PDI(cua);
	while(element != NULL) {
		if (strcmp(element->nom, nom) == 0) {
			return element;
		}		
		seguent_PDI(cua);
		element = (Simbol *)obtenir_element_PDI(cua);
	}
	return NULL;
}

void init_sintactic (char *file)
{
    char temp[100];
    strcpy(temp,file);
    strcat(temp,".sin.log");
    logsSintactic=fopen(temp,"w");
    if (logsSintactic==NULL)
        fprintf(stderr,"The sintactical analizer log file has not been created\n");
}

void end_sintactic(void)
{
    fclose(logsSintactic);
}


yyerror(char *s) {
    printf ("%s a la linia %i columna %i\n", s, lineNumber, colNumber);
    exit(0);
}


void sinError(const char *format, ...){
    va_list arg;
    char linea[4096];
    va_start(arg,format);
    sprintf(linea,"Error a la linia %i columna %i: ", lineNumber, colNumber);
    vsprintf(&linea[strlen(linea)],format,arg);
    sinLog("%s",linea);
    fprintf(stderr,"%s\n",linea);
    va_end(arg);
    exit(0);
}

void sinLog(const char *format, ...){
    va_list arg;
    char linea[4096];
    va_start(arg,format);
    if (logsSintactic!=NULL){
        vsprintf(linea,format,arg);
        fprintf(logsSintactic,"%s",linea);
        fprintf(logsSintactic,"\n");
    }
    va_end(arg);
}

// Duplica taula de simbols nomes si no conte simbols "dificils" de duplicar
// util per variables declarades RECORD, dupliquem la TS dels elements del RECORD que
// en teoria nomes han de ser del tipus VARIABLE i ARRAY_TYPE (arrays anonims).
// Si no pot duplicar retorna NULL
Taula_Simbols *intentar_duplicar_taula_simbols (Taula_Simbols *ts){
	int i;
	Simbol *simbol_dup, *simbol;
	Taula_Simbols *taula_dup = (Taula_Simbols *)crear_taula_simbols("DUPLICAT", NULL);
	
	for (i = 0; i < TAM_TAULA_SIMBOLS; i++) {
		if (ts->taula[i] != NULL) {
			Cua *colisions = ts -> taula[i];
			reiniciar_PDI(colisions);
			simbol = (Simbol *)obtenir_element_PDI(colisions);
			while(simbol != NULL) {
				if (simbol->atributs.tipus == VARIABLE) {
					simbol_dup = (Simbol *)malloc(sizeof(struct Simbol));
					memcpy(simbol_dup, simbol, sizeof(struct Simbol));
					
					if (simbol->atributs.continguts.info_variable.valor.taula != NULL) {
						// Variable d'un tipus RECORD i per tant te tambe subvariable
						simbol_dup->atributs.continguts.info_variable.valor.taula = intentar_duplicar_taula_simbols(simbol->atributs.continguts.info_variable.valor.taula);
					}
					inserir_simbol_en_taula(taula_dup, simbol_dup);
				} else if (simbol->atributs.tipus == TYPE) {
					simbol_dup = (Simbol *)malloc(sizeof(struct Simbol));
					memcpy(simbol_dup, simbol, sizeof(struct Simbol));
					inserir_simbol_en_taula(taula_dup, simbol_dup);
				} else if (simbol->atributs.tipus == RECORD_TYPE) {
					// NOTE: Conte taula de simbols, no permetem la duplicacio
					return NULL;
				} else if (simbol->atributs.tipus == ARRAY_TYPE) {
					simbol_dup = (Simbol *)malloc(sizeof(struct Simbol));
					memcpy(simbol_dup, simbol, sizeof(struct Simbol));
					inserir_simbol_en_taula(taula_dup, simbol_dup);
				} else if (simbol->atributs.tipus == ENUMERATED_TYPE) {
					// NOTE: Conte cua d'elements, no permetem la duplicacio
					return NULL;
				} else if (simbol->atributs.tipus == PROCEDURE) {
					// NOTE: Conte taula de simbols, no permetem la duplicacio
					return NULL;
				} else if (simbol->atributs.tipus == FUNCTION) {
					// NOTE: Conte taula de simbols, no permetem la duplicacio
					return NULL;
				}
				
				seguent_PDI(colisions);
				simbol = (Simbol *)obtenir_element_PDI(colisions);
			}
			return taula_dup;
		}
	}
}

// Espera rebre 2 cues d'strings de mateix tamany, 2 taules amb els simbols dels strings
// Retorna TRUE si les dues cues/taules son iguals, FALSE en cas contrari
// Util per comparar parametres i veure si el procediment/funcio esta sobrecarregat
Boolean parametres_diferents(Cua *strings_parametres1, Cua *strings_parametres2, Taula_Simbols *parametres1, Taula_Simbols *parametres2) {
	reiniciar_PDI(strings_parametres1);
	reiniciar_PDI(strings_parametres2);
	char *parametre1 = (char *)obtenir_element_PDI(strings_parametres1);
	char *parametre2 = (char *)obtenir_element_PDI(strings_parametres2);
	Boolean algun_nom_diferent = FALSE;
	Boolean tipus_diferent = FALSE;
	
	while (parametre1 != NULL) {
	   if (strcmp(parametre1, parametre2) != 0) {
		   algun_nom_diferent = TRUE;
	   }
	   
	   Simbol *sym_parametre1 = (Simbol *)obtenir_simbol(parametres1, parametre1);
	   Simbol *sym_parametre2 = (Simbol *)obtenir_simbol(parametres2, parametre2);
	   if (sym_parametre1->atributs.continguts.info_variable.tipus != sym_parametre2->atributs.continguts.info_variable.tipus) {
		   tipus_diferent = TRUE;
		   break;
	   }
	   
	   if ((sym_parametre1->atributs.continguts.info_variable.tipus == TIPUS_DERIVAT)
	   		|| (sym_parametre1->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
		   if (strcmp(sym_parametre1->atributs.continguts.info_variable.nom_tipus, sym_parametre2->atributs.continguts.info_variable.nom_tipus) != 0) {
			   tipus_diferent = TRUE;
			   break;
		   }
	   }
	   
	   seguent_PDI(strings_parametres1);
	   seguent_PDI(strings_parametres2);
	   parametre1 = (char *)obtenir_element_PDI(strings_parametres1);
	   parametre2 = (char *)obtenir_element_PDI(strings_parametres2);
	}
	

	if (tipus_diferent == TRUE) {
		// Tipus diferents, no importa si hi ha algun nom que coincideix
		return TRUE;
	} else if (algun_nom_diferent == TRUE) {
		// Tipus iguals amb noms diferents. Lo important son els tipus, per tant considerem que NO son diferents..
		sinError("No es pot definir una funcio/procediment amb els mateixos tipus de parametres d'una altra ja definida, pero amb noms diferents.");
		//return TRUE;
	} else {
		// Tipus diferents i noms diferents
		return FALSE;
	}
		
}

// Espera rebre 1 cua de simbols VARIABLE, 1 cua d'strings de mateix tamany, 1 taula amb els simbols dels strings
// Retorna TRUE si les VARIABLEs son del mateix tipus que els parametres, FALSE en cas contrari
// Util per comparar parametres de la funcio/procediment i la seva crida
Boolean crida_parametres_diferents(Cua *valors, Cua *strings_parametres, Taula_Simbols *parametres) {
	reiniciar_PDI(valors);
	reiniciar_PDI(strings_parametres);
	Simbol *sym_valor = (Simbol *)obtenir_element_PDI(valors);
	char *parametre = (char *)obtenir_element_PDI(strings_parametres);
	Boolean diferent = FALSE;
	while (parametre != NULL) {

	   Simbol *sym_parametre = (Simbol *)obtenir_simbol(parametres, parametre);
	   if (sym_valor->atributs.continguts.info_variable.tipus != sym_parametre->atributs.continguts.info_variable.tipus) {
		   diferent = TRUE;
		   
		   break;
	   }
	   
	   if ((sym_parametre->atributs.continguts.info_variable.tipus == TIPUS_DERIVAT)
	   			|| (sym_parametre->atributs.continguts.info_variable.tipus == TIPUS_INTEGER_DERIVAT)) {
		   
	   		if (strcmp(sym_valor->atributs.continguts.info_variable.nom_tipus, sym_parametre->atributs.continguts.info_variable.nom_tipus) != 0) {
			   diferent = TRUE;
			   break;
			}
			
	   }
	   
	   seguent_PDI(valors);
	   seguent_PDI(strings_parametres);
	   sym_valor = (Simbol *)obtenir_element_PDI(valors);
	   parametre = (char *)obtenir_element_PDI(strings_parametres);
	}
	
	return diferent;
}

// Comprova de forma recursiva, que tots els procediments/funcions tinguin body definit
// es a dir, siguin 'complets'.
void comprovacio_complets(Taula_Simbols *ts) {
	int i;
	
	for (i = 0; i < TAM_TAULA_SIMBOLS; i++) {
		if (ts->taula[i] != NULL) {
			Cua *colisions = ts -> taula[i];
			reiniciar_PDI(colisions);
			Simbol *simbol = (Simbol *)obtenir_element_PDI(colisions);
			while(simbol != NULL) {
				// EL SIMBOL ES UN PROCEDIMENT
				if (simbol->atributs.tipus == PROCEDURE) {
					if (simbol->atributs.continguts.info_procedure.complet == FALSE) {
						sinError("El procediment '%s' no te un cos definit.", simbol->nom);
					}
					// Comprovar procediments/funcions encaixades
					comprovacio_complets(simbol->atributs.continguts.info_procedure.taula);
					
					
					// Comprovar que totes les sobrecarregues tenen body
					if (simbol->atributs.continguts.info_procedure.num_sobrecarregues>0) {
						
						Cua *sobrecarregues = simbol->atributs.continguts.info_procedure.sobrecarregues;
						reiniciar_PDI(sobrecarregues);
						Info_Procedure *info_procedure = (Info_Procedure *)obtenir_element_PDI(sobrecarregues);
						while (info_procedure != NULL) {

							if (info_procedure->complet == FALSE) {
								sinError("El procediment sobrecarregat '%s' (%i) no te un cos definit.", simbol->nom, info_procedure->id);
							}
							// Comprovar procediments/funcions encaixades
							comprovacio_complets(info_procedure->taula);
							
							seguent_PDI(sobrecarregues);
							info_procedure = (Info_Procedure *)obtenir_element_PDI(sobrecarregues);
						}
					}
				}
	
				// EL SIMBOL ES UNA FUNCIO
				if (simbol->atributs.tipus == FUNCTION){
					if (simbol->atributs.continguts.info_function.complet == FALSE) {
						sinError("La funcio '%s' no te un cos definit.", simbol->nom);
					}
					if (simbol->atributs.continguts.info_function.return_indicat == FALSE) {
						sinError("La funcio '%s' no retorna cap valor en cap moment (falta RETURN).", simbol->nom);
					}
					// Comprovar procediments/funcions encaixades
					comprovacio_complets(simbol->atributs.continguts.info_function.taula);
					
					// Comprovar que totes les sobrecarregues tenen body
					if (simbol->atributs.continguts.info_function.num_sobrecarregues>0) {
						Cua *sobrecarregues = simbol->atributs.continguts.info_function.sobrecarregues;
						reiniciar_PDI(sobrecarregues);
						Info_Function *info_function = (Info_Function *)obtenir_element_PDI(sobrecarregues);
						while (info_function != NULL) {
							if (info_function->complet == FALSE) {
								sinError("La funcio sobrecarregada '%s' (%i) no te un cos definit.", simbol->nom, info_function->id);
							}
							if (info_function->return_indicat == FALSE) {
								sinError("La funcio sobrecarregada '%s' (%i) no retorna cap valor en cap moment (falta RETURN).", simbol->nom, info_function->id);
							}
							// Comprovar procediments/funcions encaixades
							comprovacio_complets(info_function->taula);
							
							seguent_PDI(sobrecarregues);
							info_function = (Info_Function *)obtenir_element_PDI(sobrecarregues);
						}
					}
				}
	
				seguent_PDI(colisions);
				simbol = (Simbol *)obtenir_element_PDI(colisions);
			}
		}
	}
}

void completa_goto(Cua *codi, Cua *posicions, int valor) {
	if (codi != NULL) {
		if (posicions != NULL) {
			reiniciar_PDI(posicions);
			int *pos = (int*)obtenir_element_PDI(posicions);
			while (pos != NULL) {
				Tipus_Sentencia *sentencia = (Tipus_Sentencia*)obtenir_element_pos(codi, *pos);
				if (sentencia->operacio == GOTO) {
					sprintf(sentencia->arg1, "%i", valor);
					// PRINTFs
					//printf("** Completa linea %i amb 'GOTO %i'\n", *pos, valor);
				} else {
					// IF
					sprintf(sentencia->arg3, "%i", valor);
					// PRINTFs
					//printf("** Completa linea %i amb 'IF ... GOTO %i'\n", *pos, valor);
				}
				
				seguent_PDI(posicions);
				pos = (int*)obtenir_element_PDI(posicions);
			}
		}
	}
}

void emite(Cua *sentencies, Tipus_Operacio operacio, char *arg1, char *arg2, char *arg3) {
	if (sentencies == NULL) {
		return;
	}
	Tipus_Sentencia *sentencia = (Tipus_Sentencia*)malloc(sizeof(Tipus_Sentencia));
	sentencia->esborrat = FALSE;
	sentencia->operacio = operacio;
	if (arg1 != NULL) {
		sprintf(sentencia->arg1, "%s", arg1);
	}
	if (arg2 != NULL) {
		sprintf(sentencia->arg2, "%s", arg2);
	}
	if (arg3 != NULL) {
		sprintf(sentencia->arg3, "%s", arg3);
	}
	encuar(sentencies, sentencia);
	
	// PRINTFs
	/*if (arg3 != NULL) {
		printf("%i: %i %s %s %s\n", seguent_quadruple, operacio, arg1, arg2, arg3);
	} else if (arg2 != NULL) {
		printf("%i: %i %s %s\n", seguent_quadruple, operacio, arg1, arg2);
	} else if (arg1 != NULL) {
		printf("%i: %i %s\n", seguent_quadruple, operacio, arg1);
	} else {
		printf("%i: %i\n", seguent_quadruple, operacio);
	}*/
}

char *nou_temp() {
	char *tmp = (char*)malloc(STRING_LENGTH);
	// num_temporal es inicialitzat a 0 a cada inici de procediment/funcio
	sprintf(tmp, "$t%i", num_temporal);
	num_temporal++;
	return tmp;
}

/// Procediment que emmagatzema la TS del programa principal de MiniAda per poder tractar amb
/// major coherencia les impressions de codi de tres adreces i dels registres d'activacio,
/// partint de la taula de simbols del programa i descartant la taula de simbols Main que conte
/// funcions predefinides.
void assignar_programa(Taula_Simbols *taula_actual, Taula_Simbols *principal)
{
    Taula_Simbols *punter_taula = taula_actual;
    while(punter_taula -> anterior != NULL) {
        punter_taula = punter_taula -> anterior;
    }
    punter_taula -> programa_principal = principal;
}


// Marca com esborrats les sentencies del 'codi' de les posicions indicades a la cua de posicions
/*void marcar_sentencies_esborrades(Cua *codi, Cua* posicions) {
	if (codi != NULL) {
		if (posicions != NULL) {
			reiniciar_PDI(posicions);
			int *pos = (int*)obtenir_element_PDI(posicions);
			while (pos != NULL) {
				Tipus_Sentencia *sentencia = (Tipus_Sentencia *)obtenir_element_pos(codi, *pos);
				sentencia->esborrat = TRUE;
				//seguent_quadruple--; // Chapuza
				//printf("- %i\n", *pos);
				
				seguent_PDI(posicions);
				pos = (int*)obtenir_element_PDI(posicions);
			}
		}
	}
	
}*/

void incrementar_elements(Cua * posicions) {
	if (posicions != NULL) {
		reiniciar_PDI(posicions);
		int *pos = (int*)obtenir_element_PDI(posicions);
		while (pos!=NULL) {
			//printf(" %i", *pos);
			*pos = *pos + 1;
			//printf(" -> %i\n", *pos);
			seguent_PDI(posicions);
			pos = (int*)obtenir_element_PDI(posicions);
		}
	}
}



